<?php

namespace Netzmacher\Caddy\Data;

use TYPO3\CMS\Extensionmanager\Utility\EmConfUtility;
use TYPO3\CMS\Extensionmanager\Utility\ListUtility;

if( !defined( 'TYPO3_MODE' ) )
{
	die( 'Access denied.' );
}

/* * *************************************************************
 *  Copyright notice
 *
 *  (c) 2017-2020 -  Dirk Wildt <http://wildt.at.die-netzmacher.de>
 *  All rights reserved
 *
 *  This script is part of the TYPO3 project. The TYPO3 project is
 *  free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  The GNU General Public License can be found at
 *  http://www.gnu.org/copyleft/gpl.html.
 *
 *  This script is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  This copyright notice MUST APPEAR in all copies of the script!
 * ************************************************************* */

/**
 * Class provides methods for managing default powermail data
 *
 * @author      Dirk Wildt <http://wildt.at.die-netzmacher.de>
 * @package     TYPO3
 * @subpackage  caddy
 * @internal		#i0149/#t1251
 * @version     7.4.4
 * @since       7.4.4
 */
class DataPowermailDefaults
{

	private $_extKey;
	private $_pid;

	/**
	 * PowermailDefaultValues(): Set the Powermail POST data array
	 *
	 * @return      void
	 * @version     7.4.4
	 * @since       7.4.4
	 */
	public function PowermailDefaultValues()
	{	
		if( !$this->_requirements() )
		{
			return;
		}
		
		$sesArray = $GLOBALS[ 'TSFE' ]->fe_user->getKey( 'ses', $this->extKey . '_' . $this->pid );
		$_POST[ 'tx_powermail_pi1' ] = $sesArray[ 'tx_powermail_pi1' ];
		//var_dump(__METHOD__, __LINE__, $_POST, $sesArray);
	}

	/**
	 * PowermailPostToCaddySession(): Stores Powermail POST data array in the Caddy session
	 *
	 * @return      void
	 * @version     7.4.4
	 * @since       7.4.4
	 */
	public function PowermailPostToCaddySession()
	{
		if( !$this->_requirements() )
		{
			return;
		}

		$tx_powermail_pi1 = \TYPO3\CMS\Core\Utility\GeneralUtility::_POST( 'tx_powermail_pi1' );
		//var_dump( __METHOD__, __LINE__, $tx_powermail_pi1 );

		if( !isset( $tx_powermail_pi1[ 'field' ] ) )
		{
			unset( $tx_powermail_pi1 );
			return;
		}

		$fields = $tx_powermail_pi1[ 'field' ];
		//var_dump( __METHOD__, __LINE__, $fields );

		$sesArray = $GLOBALS[ 'TSFE' ]->fe_user->getKey( 'ses', $this->extKey . '_' . $this->pid );
		$sesArray[ 'tx_powermail_pi1' ][ 'field' ] = $fields;
		$GLOBALS[ 'TSFE' ]->fe_user->setKey( 'ses', $this->extKey . '_' . $this->pid, $sesArray );
		$GLOBALS[ 'TSFE' ]->storeSessionData();

		unset( $tx_powermail_pi1 );
	}

	/**
	 * _requirements(): 
	 *
	 * @return      boolean
	 * @version     7.4.4
	 * @since       7.4.4
	 */
	private function _requirements()
	{
		if( !$this->_requirementsPmVersGreaterThan2x() )
		{
			return false;
		}
		return true;
	}

	/**
	 * _requirementsPmVersGreaterThan2x(): 
	 *
	 * @return      boolean
	 * @version     7.4.4
	 * @since       7.4.4
	 */
	private function _requirementsPmVersGreaterThan2x()
	{
//		$EmConfUtility = \TYPO3\CMS\Core\Utility\GeneralUtility::makeInstance( EmConfUtility::class );
//		$ListUtitlity = \TYPO3\CMS\Core\Utility\GeneralUtility::makeInstance( ListUtility::class );
//		$ListUtitlity->injectEmConfUtility($EmConfUtility);
//		$test = $ListUtitlity->enrichExtensionsWithEmConfAndTerInformation( array( 'powermail' => array('version' => '7.x.x' )) );
//		var_dump( __METHOD__, __LINE__, $test );
//		return false;
		return true;
	}

	/**
	 * setExtKey(): 
	 *
	 * @param string	$extKey
	 * @return      void
	 * @version     7.4.4
	 * @since       7.4.4
	 */
	public function setExtKey( $extKey )
	{
		if( empty( $extKey ) )
		{
			$prompt = 'FATAL ERROR: ' . __METHOD__ . " (#" . __LINE__ . "): extKey is empty.";
			die( $prompt );
		}
		$this->_extKey = $extKey;
	}

	/**
	 * setPid(): 
	 *
	 * @param integer	$pid
	 * @return      void
	 * @version     7.4.4
	 * @since       7.4.4
	 */
	public function setPid( $pid )
	{
		if( empty( $pid ) )
		{
			$prompt = 'FATAL ERROR: ' . __METHOD__ . " (#" . __LINE__ . "): pid is empty.";
			die( $prompt );
		}
		$this->_pid = $pid;
	}

}
