<?php

namespace Netzmacher\Caddy\Lib\Caddy;

/* * *************************************************************
 *  Copyright notice
 *
 *  (c) 2014-2020 - Dirk Wildt <http://wildt.at.die-netzmacher.de>
 *  All rights reserved
 *
 *  This script is part of the TYPO3 project. The TYPO3 project is
 *  free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  The GNU General Public License can be found at
 *  http://www.gnu.org/copyleft/gpl.html.
 *
 *  This script is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  This copyright notice MUST APPEAR in all copies of the script!
 * ************************************************************* */

/**
 * plugin 'Cart to powermail' for the 'caddy' extension.
 *
 * @author	Dirk Wildt <http://wildt.at.die-netzmacher.de>
 * @package	TYPO3
 * @subpackage	tx_caddy
 * @version	7.2.12
 * @since       4.0.7
 */
class Stockmanager
{

	private $_conf = null; // Array    : TypoScript configuration of the parent object
	private $_initInstances = null; // Boolean  : true, if instances are init
	private $_quantityIndefinitely = 9999999;
	private $_objDrs = null; // Object   : Caddy DRS
	private $_objParent = null; // Object   : parent object
	private $_objSession = null; // Object   : Caddy session
	public $extKey = 'caddy';
	public $prefixId = 'tx_caddy_pi1';
	public $scriptRelPath = 'Resources/Private/Language/Plugins/pi1/class.tx_caddy_pi1.php';


	/**
	 * _init( )
	 *
	 * @return	void
	 * @access private
	 * @internal   #54628
	 * @version    4.0.7
	 * @since      4.0.7
	 */
	private function _init()
	{
		$this->_initInstances();
	}

	/*	 * *********************************************
	 *
	 * Init
	 *
	 * ******************************************** */

	/**
	 * _clearPageCache( )
	 *
	 * @return	void
	 * @access private
	 * @internal   #i0196
	 * @version    8.8.0
	 * @since      8.8.0
	 */
	private function _clearPageCache()
	{
		$GLOBALS['TYPO3_DB']->exec_TRUNCATEquery('cf_cache_pages');
	}

	/**
	 * _initInstances( )
	 *
	 * @return	void
	 * @access private
	 * @internal   #54628
	 * @version    4.0.7
	 * @since      4.0.7
	 */
	private function _initInstances()
	{
		if( !( $this->_initInstances === null ) )
		{
			return;
		}

		$this->_initInstancesSession();

		$this->_initInstances = true;
	}

	/**
	 * _initInstancesSession( )
	 *
	 * @return	void
	 * @access private
	 * @internal   #54628
	 * @version    4.0.7
	 * @since      4.0.7
	 */
	private function _initInstancesSession()
	{
		// #i0226, 180312, dwildt, ~
		$this->_objSession = \TYPO3\CMS\Core\Utility\GeneralUtility::makeInstance( \Netzmacher\Caddy\Lib\Session::class );
		$this->_objSession = \TYPO3\CMS\Core\Utility\GeneralUtility::makeInstance( 'Netzmacher\Caddy\Lib\Session' );
		$this->_objSession->setParentObject( $this );
	}

	/*	 * *********************************************
	 *
	 * State
	 *
	 * ******************************************** */

	/**
	 * _stateIsDisabled( ) : Returns a prompt, if inventory control is disabled
	 *
	 * @return	boolean		        $_stateIsDisabled  : true;
	 * @access private
	 * @version    4.0.7
	 * @since      4.0.7
	 */
	private function _stateIsDisabled( $item )
	{
		switch( true )
		{
			case( $this->_stateIsDisabledByCaddy() ):
				return true;
			case( $this->_stateIsDisabledByShop() ):
				return true;
			case( $this->_stateIsDisabledByItem( $item ) ):
				return true;
			default:
				return false;
		}
	}

	/**
	 * _stateIsDisabledByCaddy( ) : Returns true, if inventory control is disabled by Caddy
	 *
	 * @return	boolean
	 * @access private
	 * @version    4.0.7
	 * @since      4.0.7
	 */
	private function _stateIsDisabledByCaddy()
	{
		if( !$this->_objDrs->drsStockmanager )
		{
			// #i0130, 170131, dwildt 1-/+
			//return !$disabled;
			return $disabled;
		}

		switch( $disabled )
		{
			case( false ):
				$prompt = 'Inventory control is enabled by Caddy.';
				\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[INFO/STOCKMANAGER] ' . $prompt, $this->extKey, 0 );
				break;
			case( true ):
			default:
				$prompt = 'Inventory control is disabled by Caddy.';
				\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[INFO/STOCKMANAGER] ' . $prompt, $this->extKey, 0 );
				break;
		}
		$prompt = 'See Constant Editor > Category [CADDY - INVENTORY CONTROL]';
		\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[HELP/STOCKMANAGER] ' . $prompt, $this->extKey, 1 );

		return $disabled;
	}

	/**
	 * _stateIsDisabledByItem( ) : Returns true, if inventory control is disabled by teh current item
	 *
	 * @return	boolean
	 * @access private
	 * @version    4.0.7
	 * @since      4.0.7
	 */
	private function _stateIsDisabledByItem( $item )
	{

		$disabled = !$item[ 'stockmanagement' ];

		// RETURN : no DRS
		if( !$this->_objDrs->drsStockmanager )
		{
			return $disabled;
		}
		// RETURN : no DRS
		// DRS
		switch( $disabled )
		{
			case( false ):
				$prompt = 'Inventory control is enabled by the item.';
				\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[INFO/STOCKMANAGER] ' . $prompt, $this->extKey, 0 );
				break;
			case( true ):
			default:
				$prompt = 'Inventory control is disabled by the item.';
				\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[INFO/STOCKMANAGER] ' . $prompt, $this->extKey, 0 );
				break;
		}
		$prompt = 'See the stockmanager field of the item.';
		\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[HELP/STOCKMANAGER] ' . $prompt, $this->extKey, 1 );
		// DRS

		return $disabled;
	}

	/**
	 * _stateIsDisabledByShop( ) : Returns true, if inventory control is disabled by the shop
	 *
	 * @return	boolean
	 * @access private
	 * @version    4.0.7
	 * @since      4.0.7
	 */
	private function _stateIsDisabledByShop()
	{
		$disabled = false;

		$stockquantity = $this->_conf[ 'db.' ][ 'stockquantity' ];
		$stockmanagement = $this->_conf[ 'db.' ][ 'stockmanagement' ];

		switch( true )
		{
			case(!$stockquantity ):
			case(!$stockmanagement ):
				$disabled = true;
				break;
			default:
				$disabled = false;
				break;
		}

		unset( $stockquantity );
		unset( $stockmanagement );

		// RETURN : no DRS
		if( !$this->_objDrs->drsStockmanager )
		{
			return $disabled;
		}
		// RETURN : no DRS
		// DRS
		switch( $disabled )
		{
			case( false ):
				$prompt = 'Inventory control is enabled by the shop.';
				\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[INFO/STOCKMANAGER] ' . $prompt, $this->extKey, 0 );
				break;
			case( true ):
			default:
				$prompt = 'Inventory control isn\'t enabled by the shop.';
				\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[INFO/STOCKMANAGER] ' . $prompt, $this->extKey, 0 );
				break;
		}
		$prompt = 'Inventory control is enabled by the shop, if the TypoScript properties db.stockquantity and db.stockmanagement are set.';
		\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[HELP/STOCKMANAGER] ' . $prompt, $this->extKey, 1 );
		// DRS

		return $disabled;
	}

	/*	 * *********************************************
	 *
	 * Get Methods
	 *
	 * ******************************************** */

	/**
	 * getItemQuantity( )  : Get the quantity of the given item in stock.
	 *
	 * @param	string		$table    : label of the SQL items table
	 * @param	string		$field    : label of the SQL stock quantity field
	 * @param	integer		$uid      : uid of the item
	 * @return	integer		$quantity : quantity in stock after decreasing
	 * @access public
	 * @version    4.0.7
	 * @since      4.0.7
	 */
	public function getItemQuantity( $item )
	{
		// RETURN indefinitely quantity : stockmanagement isn't enabled
		if( $this->_stateIsDisabled( $item ) )
		{
			return $this->_quantityIndefinitely;
		}

		// RETURN : quantity of the item in stock
		$stockquantity = intval( $item[ 'stockquantity' ] );

		if( $stockquantity < 0 )
		{
			$stockquantity = 0;
		}
		return intval( $stockquantity );
	}

	/*	 * *********************************************
	 *
	 * Prompts
	 *
	 * ******************************************** */

	/**
	 * dieQtyError( )  :
	 *
	 * @param	string		$title      : item title
	 * @param	integer		$qtyByOrder : quantity of the item for ordering
	 * @param	integer		$qtyInStock : quantity of the item in the stock
	 * @access public
	 * @version    4.0.7
	 * @since      4.0.7
	 */
	public function dieQtyError( $title, $qtyByOrder, $qtyInStock )
	{
		switch( true )
		{
			case( $qtyByOrder > 1 ):
				$prompt = 'Sorry, but the item "' . $title . '" isn\'t available ' . $qtyByOrder . ' time any longer. '
								. 'Please go back to the order form, decrease the wanted quantity to ' . $qtyInStock . ' and send the form again.'
				;
				break;
			default:
				$prompt = 'Sorry, but the item "' . $title . '" isn\'t available any longer. '
								. 'Please go back to the order form, remove the item and send the form again.'
				;
				break;
		}
		
		$prompt = ''
						. '<div style="background:red;color:white;font-weight:bold;padding:1em;text-align:center;">'
						. $prompt
						. '</div>';

		die( $prompt );
	}

	/*	 * *********************************************
	 *
	 * Items
	 *
	 * ******************************************** */

	/**
	 * itemsUpdate( ) :
	 *
	 * @param	array		$items :
	 * @return	void
	 * @access public
	 * @version    4.0.7
	 * @since      4.0.7
	 */
	public function itemsUpdate( $items )
	{
		foreach( ( array ) $items as $item )
		{
			$this->_itemsUpdateItem( $item );
		}
	}

	/**
	 * _itemsUpdateItem( ) :
	 *
	 * @param	array		$item :
	 * @return	void
	 * @access private
	 * @version    4.0.7
	 * @since      4.0.7
	 */
	private function _itemsUpdateItem( $item )
	{
		if( $this->_stateIsDisabled( $item ) )
		{
			return;
		}

		$table = $this->_conf[ 'db.' ][ 'table' ];
		$stockquantity = $this->_conf[ 'db.' ][ 'stockquantity' ];
		$orderquantity = $item[ 'qty' ];
		$uid = $item[ 'uid' ];

		// Build the query
		$query = 'UPDATE ' . $table . ' '
						. 'SET ' . $stockquantity . ' = ' . $stockquantity . ' - ' . $orderquantity . ' '
						. 'WHERE  uid = ' . $uid;

		// DRS - Development Reporting System
		if( !$this->_objDrs->drsStockmanager )
		{
			\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[INFO/STOCKMANAGER] ' . $query, $this->_objParent->extKey, 0 );
		}
		// DRS - Development Reporting System
		// Execute the query
		$GLOBALS[ 'TYPO3_DB' ]->sql_query( $query );

		// Evaluate the query
		$affected_rows = $GLOBALS[ 'TYPO3_DB' ]->sql_affected_rows();
		$error = $GLOBALS[ 'TYPO3_DB' ]->sql_error();


		///////////////////////////////////////////////////////////////////////////////
		//
      // ERROR: debug report in the frontend

		if( $error )
		{
			die();
		}
		// ERROR: debug report in the frontend

		if( $affected_rows )
		{

//			$TCE = \TYPO3\CMS\Core\Utility\GeneralUtility::makeInstance( 't3lib_TCEmain' );
//			$TCE->admin = 1;
//			$TCE->clear_cacheCmd( 'pages' );
			// #i0196, 180118, dwildt, 3-,1+
//			// #i0134, 170218, dwildt, ~
//			$CacheService = \TYPO3\CMS\Core\Utility\GeneralUtility::makeInstance( TYPO3\CMS\Extbase\Service\CacheService::class );
//  		$CacheService->clearPageCache();
			$this->_clearPageCache();
			
		}

		return $affected_rows;
	}

	/*	 * *********************************************
	 *
	 * Set Methods
	 *
	 * ******************************************** */

	/**
	 * setItemQuantity( )  : Changes the quantity of the given item in stock.
	 *                       If given quantity is minus, quantity in stock will decreased.
	 *
	 * @param	string		$table    : label of the SQL items table
	 * @param	string		$field    : label of the SQL stock quantity field
	 * @param	integer		$uid      : uid of the item
	 * @param	integer		$quantity : the quantity for decreasing
	 * @return	integer		$quantity : quantity in stock after decreasing
	 * @access public
	 * @version    4.0.7
	 * @since      4.0.7
	 */
	public function setItemQuantity( $table, $field, $uid, $quantity )
	{
		if( $this->_stateIsDisabled() )
		{
			return $quantity;
		}
		// #54628, 131229, dwildt, 1+
		$this->_init();

//var_dump( __METHOD__, __LINE__ );

		return $quantity;
	}

	/**
	 * setItemsDecrease( )  : Changes the quantity of the given item in stock.
	 *                       If given quantity is minus, quantity in stock will decreased.
	 *
	 * @param	array		$table    : label of the SQL items table
	 * @return	integer		$quantity : quantity in stock after decreasing
	 * @access public
	 * @version    4.0.7
	 * @since      4.0.7
	 */
	public function setItemsDecrease( $pid )
	{
		if( $this->_stateIsDisabled() )
		{
			$prompt = ':TODO: stateisdisabled( )';
			var_dump( __METHOD__, __LINE__, $prompt );
			die();
		}

		$this->initPidCaddy( $pid );
		$items = $this->_objSession->productsGet( $this->pidCaddy );
// 141010: Returns noting ????
		var_dump( __METHOD__, __LINE__, $items );
		die();


		return;
	}

	/**
	 * setParentObject( )  : Set the parent object
	 *
	 * @param	object		$_objParent: parent object
	 * @return	void
	 * @access public
	 * @version    4.0.7
	 * @since      4.0.7
	 */
	public function setParentObject( $pObj )
	{
		if( !is_object( $pObj ) )
		{
			$prompt = 'ERROR: no object!<br />' . PHP_EOL .
							'Sorry for the trouble.<br />' . PHP_EOL .
							'TYPO3 Caddy<br />' . PHP_EOL .
							__METHOD__ . ' (' . __LINE__ . ')';
			die( $prompt );
		}
		$this->_objParent = $pObj;

		$this->setParentObjectConf();
		$this->setParentObjectDrs();
	}

	/**
	 * setParentObjectConf( )  : Set the TypoScript configuration
	 *
	 * @return	void
	 * @access private
	 * @version    4.0.7
	 * @since      4.0.7
	 */
	private function setParentObjectConf()
	{
		$conf = $this->_objParent->getConf();

		if( !is_array( $conf ) || empty( $conf ) )
		{
//$prompt = 'debug trail: ' . t3lib_utility_Debug::debugTrail( ) . PHP_EOL .
//          'TYPO3 Caddy<br />' . PHP_EOL .
//        __METHOD__ . ' (' . __LINE__ . ')';
//$prompt = str_replace( '//', PHP_EOL . '//', $prompt );
//var_dump( __METHOD__, __LINE__, $prompt );
			$prompt = 'ERROR: no configuration!<br />' . PHP_EOL .
							'Sorry for the trouble.<br />' . PHP_EOL .
							'TYPO3 Caddy<br />' . PHP_EOL .
							__METHOD__ . ' (' . __LINE__ . ')';
			die( $prompt );
		}
		$this->_conf = $conf;
	}

	/**
	 * setParentObjectDrs( )  : Set the DRS object
	 *
	 * @return	void
	 * @access private
	 * @version    4.0.7
	 * @since      4.0.7
	 */
	private function setParentObjectDrs()
	{
		$drs = $this->_objParent->getDrs();

		if( !is_object( $drs ) || empty( $drs ) )
		{
//$prompt = 'debug trail: ' . t3lib_utility_Debug::debugTrail( ) . PHP_EOL .
//          'TYPO3 Caddy<br />' . PHP_EOL .
//        __METHOD__ . ' (' . __LINE__ . ')';
//$prompt = str_replace( '//', PHP_EOL . '//', $prompt );
//var_dump( __METHOD__, __LINE__, $prompt );
			$prompt = 'ERROR: no DRS!<br />' . PHP_EOL .
							'Sorry for the trouble.<br />' . PHP_EOL .
							'TYPO3 Caddy<br />' . PHP_EOL .
							__METHOD__ . ' (' . __LINE__ . ')';
			die( $prompt );
		}
		$this->_objDrs = $drs;
	}

}