<?php

namespace Netzmacher\Caddy\Lib\EPayment\Paymill;

use \Netzmacher\Refresh\Compatibility\Frontend\ContentObject\ContentObjectRenderer;
use TYPO3\CMS\Core\Utility\DebugUtility;

/* * *************************************************************
 *  Copyright notice
 *
 *  (c) 2014-2020 - Dirk Wildt <http://wildt.at.die-netzmacher.de>
 *  All rights reserved
 *
 *  This script is part of the TYPO3 project. The TYPO3 project is
 *  free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  The GNU General Public License can be found at
 *  http://www.gnu.org/copyleft/gpl.html.
 *
 *  This script is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  This copyright notice MUST APPEAR in all copies of the script!
 * ************************************************************* */

/**
 * Get the HTML template with the payment methods
 *
 * @author	Dirk Wildt <http://wildt.at.die-netzmacher.de>
 * @package    TYPO3
 * @subpackage    tx_caddy
 * @internal    #53678
 * @version     8.12.0
 * @since       4.0.6
 */
class Payment extends \Netzmacher\Caddy\AbstractPlugin
{

	public $extKey = 'caddy';
	public $prefixId = 'Payment';
	public $scriptRelPath = 'Classes/Lib/EPayment/Paymill/Payment.php';
	public $conf = null;		// array    : current TypoScript configuration
	private $content = null;		// string   : content. Will returned by main( )
	public $drs = null;		// object   : instance of drs class.
	private $dynamicMarkers = null;		// object   : instance of dynamicMarkers class.
	public $local_cObj = null;
	private $paymentId = null;		// integer  : current payment id. 1: credit card, 2: elv. 3: sepa (elv-iban).
	private $pObj = null;		// object   : parent object
	private $pid = null;		// integer  : pid of the current page
	private $prompts = array();	// array    : prompts

	/*	 * *********************************************
	 *
	 * Main
	 *
	 * ******************************************** */

	/**
	 * main( ):
	 *
	 * @param	integer		$paymentId  : current payment id. 1: credit card, 2: elv. 3: sepa (elv-iban).
	 * @param	integer		$pid        : ...
	 * @return	array		$payment    : true, if payment was successful
	 * @access public
	 * @version     4.0.6
	 * @since       4.0.6
	 */
	public function main( $paymentId = null, $pid = null )
	{
//$sesArray = $GLOBALS['TSFE']->fe_user->getKey( 'ses', $this->extKey . '_' . $pid );
//var_dump( __METHOD__, __LINE__, $sesArray['e-payment'] );
//$prompt = 'debug trail: ' . DebugUtility::debugTrail( );
//$prompt = str_replace( '//' , PHP_EOL . '//' , $prompt );
//var_dump( __METHOD__, __LINE__);

		$arrReturn = array(
				'content' => null,
				'payment' => false
		);

		$this->init( $paymentId, $pid );

		$this->requirementsDie();

		$this->content = $this->template( '###CONTENT###' );
		$this->content = $this->templateSectionRemove( $this->content );
		$this->content = $this->templateSectionActive( $this->content );

		$this->prompts = $this->paymillLib->paymillInit();
		if( $this->prompts )
		{
//var_dump( __METHOD__, __LINE__);
//var_dump( __METHOD__, __LINE__, $this->prompts );
			$this->sessionDataRemove();
			$this->serverPrompt();
			$this->templateMode();
			$this->dynamicMarkers->scriptRelPath = $this->scriptRelPath;
			$this->content = $this->dynamicMarkers->main( $this->content, $this );

			$arrReturn = array(
					'content' => $this->content,
					'payment' => false
			);
//$sesArray = $GLOBALS['TSFE']->fe_user->getKey( 'ses', $this->extKey . '_' . $pid );
//var_dump( __METHOD__, __LINE__, $sesArray['e-payment'] );
			return $arrReturn;
		}

		$this->templateMode();


		if( !$this->requirements() )
		{
			$this->sessionDataRemove();
			$this->dynamicMarkers->scriptRelPath = $this->scriptRelPath;
			$this->content = $this->dynamicMarkers->main( $this->content, $this );
			$arrReturn = array(
					'content' => $this->content,
					'payment' => false
			);
//$sesArray = $GLOBALS['TSFE']->fe_user->getKey( 'ses', $this->extKey . '_' . $pid );
//var_dump( __METHOD__, __LINE__, $sesArray['e-payment'] );
//var_dump( __METHOD__, __LINE__);
			return $arrReturn;
		}

		// Insert payment data into content
		$payment = $this->paymentMethods();
		// Insert prompt into content
		$this->serverPrompt();
		// Insert
		$this->session();

		$this->dynamicMarkers->scriptRelPath = $this->scriptRelPath;
		$this->content = $this->dynamicMarkers->main( $this->content, $this );

		$arrReturn = array(
				'content' => $this->content,
				'payment' => $payment
		);
//$sesArray = $GLOBALS['TSFE']->fe_user->getKey( 'ses', $this->extKey . '_' . $pid );
//var_dump( __METHOD__, __LINE__, $sesArray['e-payment'] );
//var_dump( __METHOD__, __LINE__, $arrReturn );
//var_dump( __METHOD__, __LINE__);
		return $arrReturn;
	}

	/*	 * *********************************************
	 *
	 * Init
	 *
	 * ******************************************** */

	/**
	 * init( ):
	 *
	 * @param	integer		$paymentId  : current payment id. 1: credit card, 2: elv. 3: sepa (elv-iban).
	 * @param	[type]		$pid: ...
	 * @return	void
	 * @access private
	 * @version     4.0.6
	 * @since       4.0.6
	 */
	private function init( $paymentId = null, $pid = null )
	{
		$this->initVars( $paymentId, $pid );
		$this->initInstances();
		$this->initDrs();
	}

	/**
	 * initDrs( )
	 *
	 * @return	void
	 * @access private
	 * @version    4.0.6
	 * @since      4.0.6
	 */
	private function initDrs()
	{
		$this->drs->init();
	}

	/**
	 * initInstances( ):
	 *
	 * @return	void
	 * @access private
	 * @version     4.0.6
	 * @since       4.0.6
	 */
	private function initInstances()
	{
		$this->initInstancesDrs();
		$this->initInstancesDynamicMarkers();
		$this->initInstancesPaymillLib();
	}

	/**
	 * initInstancesDrs( ):
	 *
	 * @return	void
	 * @access private
	 * @version     4.0.6
	 * @since       4.0.6
	 */
	private function initInstancesDrs()
	{
		// #i0226, 180312, dwildt, ~
		$this->drs = \TYPO3\CMS\Core\Utility\GeneralUtility::makeInstance( \Netzmacher\Caddy\Utility\Debug\Drs::class );
		// #i0061, 141129, dwildt, 2-/3+
		//$this->drs->pObj = $this;
		//$this->drs->row = $this->cObj->data;
		$this->drs->setExtConf( $this->arr_extConf );
		$this->drs->setFlexform( $this->flexform );
		$this->drs->setRow( $this->cObj->data );
	}

	/**
	 * initInstancesDynamicMarkers( ):
	 *
	 * @return	void
	 * @access private
	 * @version     4.0.6
	 * @since       4.0.6
	 */
	private function initInstancesDynamicMarkers()
	{
		// #i0226, 180312, dwildt, ~
		$this->dynamicMarkers = \TYPO3\CMS\Core\Utility\GeneralUtility::makeInstance( \Netzmacher\Caddy\Lib\Dynamicmarkers::class );
		$this->dynamicMarkers->scriptRelPath = $this->scriptRelPath;
	}

	/**
	 * initInstancesPaymillLib( ) :  Requires and initiate the class of the current e-payment provider.
	 *                               If no e-payment provider is enabled, nothing will happen.
	 *
	 * @return	void
	 * @access private
	 * @internal   #53678
	 * @version    4.0.6
	 * @since      4.0.6
	 */
	private function initInstancesPaymillLib()
	{
		// #i0226, 180312, dwildt, ~
		$this->paymillLib = \TYPO3\CMS\Core\Utility\GeneralUtility::makeInstance( \Netzmacher\Caddy\Lib\EPayment\Paymill\Main::class );
		$this->paymillLib->setParentObject( $this );
		$this->paymillLib->setPiVars( $this->pObj->piVars );
		$this->paymillLib->setPid( $this->pid );
	}

	/**
	 * initVars( )
	 *
	 * @param	integer		$paymentId  : current payment id. 1: credit card, 2: elv. 3: sepa (elv-iban).
	 * @param	[type]		$pid: ...
	 * @return	void
	 * @access private
	 * @version    4.0.6
	 * @since      4.0.6
	 */
	private function initVars( $paymentId, $pid )
	{
		$this->conf = $this->pObj->conf;
		$this->local_cObj = $GLOBALS[ 'TSFE' ]->cObj;
		$this->paymentId = $paymentId;



		$this->initVarsPid( $pid );
		$this->pi_setPiVarDefaults();
		$this->pi_loadLL();
		//$this->pi_initPIflexForm( );
	}

	/**
	 * initPid( )  : Returns the globlas tsfe id, if the given pid is null
	 *
	 * @param	integer		$pid  : given pid (may be null)
	 * @return	integer		$pid  : id of the page with the caddy plugin
	 * @internal    #54634
	 * @version     4.0.3
	 * @since       4.0.6
	 */
	private function initVarsPid( $pid = null )
	{
		if( $pid !== null )
		{
			$this->pid = $pid;
			return;
		}

		if( ( int ) $this->conf[ 'userFunc.' ][ 'caddyPid' ] )
		{
			$pid = ( int ) $this->conf[ 'userFunc.' ][ 'caddyPid' ];
		}

		if( $pid === null || $pid === '' )
		{
			if( $this->drs->drsError || $this->drsUserfunc )
			{
				$prompt = 'Given pid of the Caddy is empty!';
				\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[ERROR/SESSION] ' . $prompt, $this->extKey, 3 );
			}
			$pid = $GLOBALS[ "TSFE" ]->id;
		}

		$this->pid = $pid;
	}

	/*	 * *********************************************
	 *
	 * Payment Methods
	 *
	 * ******************************************** */

	/**
	 * paymentMethods( ):
	 *
	 * @param	integer		$paymentId  : current payment id. 1: credit card, 2: elv. 3: sepa (elv-iban).
	 * @return	boolean		$paymill  : true, if paymill was successful
	 * @access private
	 * @version     8.9.6
	 * @since       4.0.6
	 */
	private function paymentMethods()
	{
		switch( true )
		{
			case( $this->paymentId == 14 ):
			case( $this->paymentWiPowermail() ):
			case( $this->paymentWoEpayment() ):
			case( $this->paymillLib->getPaymillToken() ):
				break;
			default:
				return;
		}

//var_dump( __METHOD__, __LINE__, $errors );

		switch( $this->paymentId )
		{
			case( 1 ):
				$this->paymentMethodsCashinadvance();
				break;
			case( 2 ):
				$this->paymentMethodsInvoice();
				break;
			case( 3 ):
				$this->paymentMethodsCashondelivery();
				break;
			case( 4 ):
				$this->paymentMethodsCashonpickup();
				break;
			case( 5 ):
				$this->paymentMethodsDebitNote();
				break;
			case( 6 ):
				$this->paymentMethodsCashinadvancePaypal();
				break;
			case( 11 ):
				$this->paymentMethodsCreditcard();
				break;
			case( 12 ):
				$this->paymentMethodsElv();
				break;
			case( 13 ):
				$this->paymentMethodsElvIban();
				break;
			// #i0165/#t2145, 180311, dwildt +
			case( 14 ):
				$this->paymentMethodsPaypal();
				break;
			default:
				$this->paymentMethodsUndefined();
				break;
		}
	}

	/**
	 * paymentMethodsCashinadvance( ):
	 *
	 * @param	integer		$paymentId  : current payment id. 1: credit card, 2: elv. 3: sepa (elv-iban).
	 * @return	void
	 * @access private
	 * @version     4.0.8
	 * @since       4.0.8
	 */
	private function paymentMethodsCashinadvance()
	{
		// notice (grey) = secondary, info (blue) = [empty!], ok (green) = success, error (red) = alert
		$this->prompts[] = 'SERVER_PROMPT_WICLOSE_SUBPART|secondary alert-info|' . $this->pi_getLL( 'prompt-cashinadvance-ok' );
		$this->prompts[] = 'SERVER_PROMPT_WOCLOSE_SUBPART|success alert-success|' . $this->pi_getLL( 'prompt-paywith-cashinadvance' );

		//$marker['###VALUE_CREDITCARD-NUMBER###']  = '**** **** **** ' . $this->paymillLib->getCardNumber( );
		//$this->content = $this->pObj->MarkerBasedTemplateService->substituteMarkerArray( $this->content, $marker );
	}

	/**
	 * paymentMethodsCashinadvancePaypal( ):
	 *
	 * @param	integer		$paymentId  : current payment id. 1: credit card, 2: elv. 3: sepa (elv-iban).
	 * @return	void
	 * @access private
	 * @version     8.9.6
	 * @since       8.9.6
	 */
	private function paymentMethodsCashinadvancePaypal()
	{
		// notice (grey) = secondary, info (blue) = [empty!], ok (green) = success, error (red) = alert
		$this->prompts[] = 'SERVER_PROMPT_WICLOSE_SUBPART|secondary alert-info|' . $this->pi_getLL( 'prompt-cashinadvancepaypal-ok' );
		$this->prompts[] = 'SERVER_PROMPT_WOCLOSE_SUBPART|success alert-success|' . $this->pi_getLL( 'prompt-paywith-cashinadvancepaypal' );

		//$marker['###VALUE_CREDITCARD-NUMBER###']  = '**** **** **** ' . $this->paymillLib->getCardNumber( );
		//$this->content = $this->pObj->MarkerBasedTemplateService->substituteMarkerArray( $this->content, $marker );
	}

	/**
	 * paymentMethodsCashondelivery( ):
	 *
	 * @param	integer		$paymentId  : current payment id. 1: credit card, 2: elv. 3: sepa (elv-iban).
	 * @return	void
	 * @access private
	 * @version     4.0.8
	 * @since       4.0.8
	 */
	private function paymentMethodsCashondelivery()
	{
		// notice (grey) = secondary, info (blue) = [empty!], ok (green) = success, error (red) = alert
		$this->prompts[] = 'SERVER_PROMPT_WICLOSE_SUBPART|secondary alert-info|' . $this->pi_getLL( 'prompt-cashondelivery-ok' );
		$this->prompts[] = 'SERVER_PROMPT_WOCLOSE_SUBPART|success alert-success|' . $this->pi_getLL( 'prompt-paywith-cashondelivery' );

		//$marker['###VALUE_CREDITCARD-NUMBER###']  = '**** **** **** ' . $this->paymillLib->getCardNumber( );
		//$this->content = $this->pObj->MarkerBasedTemplateService->substituteMarkerArray( $this->content, $marker );
	}

	/**
	 * paymentMethodsCashonpickup( ):
	 *
	 * @param	integer		$paymentId  : current payment id. 1: credit card, 2: elv. 3: sepa (elv-iban).
	 * @return	void
	 * @access private
	 * @version     4.0.8
	 * @since       4.0.8
	 */
	private function paymentMethodsCashonpickup()
	{
		// notice (grey) = secondary, info (blue) = [empty!], ok (green) = success, error (red) = alert
		$this->prompts[] = 'SERVER_PROMPT_WICLOSE_SUBPART|secondary alert-info|' . $this->pi_getLL( 'prompt-cashonpickup-ok' );
		$this->prompts[] = 'SERVER_PROMPT_WOCLOSE_SUBPART|success alert-success|' . $this->pi_getLL( 'prompt-paywith-cashonpickup' );

		//$marker['###VALUE_CREDITCARD-NUMBER###']  = '**** **** **** ' . $this->paymillLib->getCardNumber( );
		//$this->content = $this->pObj->MarkerBasedTemplateService->substituteMarkerArray( $this->content, $marker );
	}

	/**
	 * paymentMethodsCreditcard( ):
	 *
	 * @param	integer		$paymentId  : current payment id. 1: credit card, 2: elv. 3: sepa (elv-iban).
	 * @return	void
	 * @access private
	 * @version     4.0.6
	 * @since       4.0.6
	 */
	private function paymentMethodsCreditcard()
	{
		// notice (grey) = secondary, info (blue) = [empty!], ok (green) = success, error (red) = alert
		$this->prompts[] = 'SERVER_PROMPT_WICLOSE_SUBPART|secondary alert-info|' . $this->pi_getLL( 'prompt-creditcard-ok' );
		$this->prompts[] = 'SERVER_PROMPT_WOCLOSE_SUBPART|success alert-success|' . $this->pi_getLL( 'prompt-paywith-creditcard' );
//    var_dump( __FILE__, __LINE__,
//            $this->paymillLib->paymillPaymentResponse
////            $this->paymillLib->paymillPaymentResponse->getCode( ),
////            $this->paymillLib->paymillPaymentResponse->getAccount( ),
////            $this->paymillLib->paymillPaymentResponse->getHolder( )
//            );
		$expiry = $this->paymillLib->getCardMonth() . '/' . $this->paymillLib->getCardYear();
		$marker[ '###VALUE_CREDITCARD-NUMBER###' ] = '**** **** **** ' . $this->paymillLib->getCardNumber();
		$marker[ '###VALUE_CREDITCARD-EXPIRY###' ] = $expiry;
		$marker[ '###VALUE_CREDITCARD-HOLDER###' ] = $this->paymillLib->getCardHolder();
		$marker[ '###VALUE_CREDITCARD-CVC###' ] = '****';

		$this->content = $this->pObj->MarkerBasedTemplateService->substituteMarkerArray( $this->content, $marker );
	}

	/**
	 * paymentMethodsDebitNote( ):
	 *
	 * @param	integer		$paymentId  : current payment id. 1: credit card, 2: elv. 3: sepa (elv-iban).
	 * @return	void
	 * @access private
	 * @version     8.9.6
	 * @since       8.9.6
	 */
	private function paymentMethodsDebitNote()
	{
		// notice (grey) = secondary, info (blue) = [empty!], ok (green) = success, error (red) = alert
		$this->prompts[] = 'SERVER_PROMPT_WICLOSE_SUBPART|secondary alert-info|' . $this->pi_getLL( 'prompt-debitnote-ok' );
		$this->prompts[] = 'SERVER_PROMPT_WOCLOSE_SUBPART|success alert-success|' . $this->pi_getLL( 'prompt-paywith-debitnote' );

		//$marker['###VALUE_CREDITCARD-NUMBER###']  = '**** **** **** ' . $this->paymillLib->getCardNumber( );
		//$this->content = $this->pObj->MarkerBasedTemplateService->substituteMarkerArray( $this->content, $marker );
	}

	/**
	 * paymentMethodsElv( ):
	 *
	 * @param	integer		$paymentId  : current payment id. 1: credit card, 2: elv. 3: sepa (elv-iban).
	 * @return	void
	 * @access private
	 * @version     4.0.6
	 * @since       4.0.6
	 */
	private function paymentMethodsElv()
	{
		// notice (grey) = secondary, info (blue) = [empty!], ok (green) = success, error (red) = alert
		$this->prompts[] = 'SERVER_PROMPT_WICLOSE_SUBPART|secondary alert-info|' . $this->pi_getLL( 'prompt-elv-ok' );
		$this->prompts[] = 'SERVER_PROMPT_WOCLOSE_SUBPART|success alert-success|' . $this->pi_getLL( 'prompt-paywith-elv' );

		$marker[ '###VALUE_ELV-ACCOUNT###' ] = $this->paymillLib->getBankAccount();
		$marker[ '###VALUE_ELV-BANKCODE###' ] = $this->paymillLib->getBankCode();
		$marker[ '###VALUE_ELV-HOLDER###' ] = $this->paymillLib->getBankHolder();

		$this->content = $this->pObj->MarkerBasedTemplateService->substituteMarkerArray( $this->content, $marker );
	}

	/**
	 * paymentMethodsElvIban( ):
	 *
	 * @param	integer		$paymentId  : current payment id. 1: credit card, 2: elv. 3: sepa (elv-iban).
	 * @return	void
	 * @access private
	 * @version     4.0.6
	 * @since       4.0.6
	 */
	private function paymentMethodsElvIban()
	{
		// notice (grey) = secondary, info (blue) = [empty!], ok (green) = success, error (red) = alert
		$this->prompts[] = 'SERVER_PROMPT_WICLOSE_SUBPART|secondary alert-info|' . $this->pi_getLL( 'prompt-elv-iban-ok' );
		$this->prompts[] = 'SERVER_PROMPT_WOCLOSE_SUBPART|success alert-success|' . $this->pi_getLL( 'prompt-paywith-elv-iban' );

		//$iban = '****' . $this->paymillLib->paymillPaymentResponse->getCode( ) . $this->paymillLib->paymillPaymentResponse->getAccount( );
		$iban = '****' . $this->paymillLib->getBankCode() . $this->paymillLib->getBankAccount();
		$marker[ '###VALUE_ELV-IBAN###' ] = $iban;
		$marker[ '###VALUE_ELV-BIC###' ] = '**********';
		$marker[ '###VALUE_ELV-IBAN-HOLDER###' ] = $this->paymillLib->getBankHolder();
//var_dump(__METHOD__, __LINE__, $marker);
		$this->content = $this->pObj->MarkerBasedTemplateService->substituteMarkerArray( $this->content, $marker );
	}

	/**
	 * paymentMethodsInvoice( ):
	 *
	 * @param	integer		$paymentId  : current payment id. 1: credit card, 2: elv. 3: sepa (elv-iban).
	 * @return	void
	 * @access private
	 * @version     4.0.8
	 * @since       4.0.8
	 */
	private function paymentMethodsInvoice()
	{
		// notice (grey) = secondary, info (blue) = [empty!], ok (green) = success, error (red) = alert
		$this->prompts[] = 'SERVER_PROMPT_WICLOSE_SUBPART|secondary alert-info|' . $this->pi_getLL( 'prompt-invoice-ok' );
		$this->prompts[] = 'SERVER_PROMPT_WOCLOSE_SUBPART|success alert-success|' . $this->pi_getLL( 'prompt-paywith-invoice' );

		//$marker['###VALUE_CREDITCARD-NUMBER###']  = '**** **** **** ' . $this->paymillLib->getCardNumber( );
		//$this->content = $this->pObj->MarkerBasedTemplateService->substituteMarkerArray( $this->content, $marker );
	}

	/**
	 * paymentMethodsPaypal( ):
	 *
	 * @param	integer		$paymentId  : current payment id. 1: credit card, 2: elv. 3: sepa (elv-iban).
	 * @return	void
	 * @access private
	 * @version     7.5.2
	 * @since       7.5.2
	 * @internal		#i0165, #t2145
	 */
	private function paymentMethodsPaypal()
	{
		// notice (grey) = secondary, info (blue) = [empty!], ok (green) = success, error (red) = alert
		$this->prompts[] = 'SERVER_PROMPT_WICLOSE_SUBPART|secondary alert-info|' . $this->pi_getLL( 'prompt-paypal-ok' );
		$this->prompts[] = 'SERVER_PROMPT_WOCLOSE_SUBPART|success alert-success|' . $this->pi_getLL( 'prompt-paywith-paypal' );
	}

	/**
	 * paymentMethodsUndefined( ):
	 *
	 * @param	integer		$paymentId  : current payment id. 1: credit card, 2: elv. 3: sepa (elv-iban).
	 * @return	void
	 * @access private
	 * @version     4.0.6
	 * @since       4.0.6
	 */
	private function paymentMethodsUndefined()
	{
		// notice (grey) = secondary, info (blue) = [empty!], ok (green) = success, error (red) = alert
		$this->prompts[] = 'SERVER_PROMPT_WICLOSE_SUBPART|alert alert-danger|' . $this->pi_getLL( 'prompt-paywith-undefined' );
	}

	/**
	 * paymentWiPowermail( ):
	 *
	 * @return	boolean
	 * @access private
	 * @version     4.0.6
	 * @since       4.0.6
	 */
	private function paymentWiPowermail()
	{
		return $this->powermailInAction();
	}

	/**
	 * paymentWoEpayment( ):
	 *
	 * @return	boolean
	 * @access private
	 * @internal #i0050
	 * @version     4.0.9
	 * @since       4.0.9
	 */
	private function paymentWoEpayment()
	{
		return $this->requirementsWoEpayment();
	}

	/*	 * *********************************************
	 *
	 * Powermail
	 *
	 * ******************************************** */

	/**
	 * powermailInAction( ):
	 *
	 * @return	boolean
	 * @access private
	 * @version     4.0.6
	 * @since       4.0.6
	 */
	private function powermailInAction()
	{
		$GP = \TYPO3\CMS\Core\Utility\GeneralUtility::_GET() + \TYPO3\CMS\Core\Utility\GeneralUtility::_POST()
		;

		switch( true )
		{
			case( $GP[ 'tx_powermail_pi1' ][ 'action' ] == 'confirmation' ):
			case( $GP[ 'tx_powermail_pi1' ][ 'action' ] == 'create' ):
			case( $GP[ 'tx_powermail_pi1' ][ 'action' ] == 'form' ):
				return true;
			//break;
			default:
				return false;
			//break;
		}

		unset( $GP );
	}

	/*	 * *********************************************
	 *
	 * Requirements
	 *
	 * ******************************************** */

	/**
	 * requirements( ):
	 *
	 * @return	boolean
	 * @access private
	 * @version     4.0.6
	 * @since       4.0.6
	 */
	private function requirements()
	{
		$requirementsMatched = false;

		switch( true )
		{
			case( $this->requirementsWoEpayment() ):
			case( $this->requirementsWiPaypal() ):
			case( $this->requirementsToken() ):
			case( $this->requirementsWiPowermail() ):
				$requirementsMatched = true;
				break;
			default:
				$requirementsMatched = false;
				break;
		}

		return $requirementsMatched;
	}

	/**
	 * requirementsDie( ) :
	 *
	 * @return	void
	 * @access public
	 * @version     4.0.6
	 * @since       4.0.6
	 */
	private function requirementsDie()
	{
		if( !is_object( $this->pObj ) )
		{
			$prompt = 'ERROR: no parent object!<br />' . PHP_EOL .
							'Sorry for the trouble.<br />' . PHP_EOL .
							'TYPO3 Caddy<br />' . PHP_EOL .
							__METHOD__ . ' (' . __LINE__ . ')';
			die( $prompt );
		}
	}

	/**
	 * requirementsToken( ): Returns false, if token isn't part of the URL params
	 *
	 * @return	boolen
	 * @access private
	 * @version     4.0.6
	 * @since       4.0.6
	 */
	private function requirementsToken()
	{
		if( $this->paymillLib->getPaymillToken() )
		{
			return true;
		}

		$this->prompts[] = 'SERVER_PROMPT_WOCLOSE_SUBPART|alert alert-danger|' . $this->pi_getLL( 'prompt-token-error' );
		$this->prompts[] = 'SERVER_PROMPT_WICLOSE_SUBPART||' . $this->pi_getLL( 'prompt-token-info1' );
		$this->prompts[] = 'SERVER_PROMPT_WICLOSE_SUBPART||' . $this->pi_getLL( 'prompt-token-info2' );

		return false;
	}

	/**
	 * requirementsWoEpayment( ): Returns true, if current payment method isn't an e-payment method
	 *
	 * @return	boolen
	 * @access private
	 * @internal #i0050
	 * @version     4.0.9
	 * @since       4.0.9
	 */
	private function requirementsWoEpayment()
	{
		$conf = $this->pObj->conf;
		$isEpayment = $conf[ 'api.' ][ 'options.' ][ 'payment.' ][ 'options.' ][ $this->paymentId . '.' ][ 'e-payment' ];

		if( !$isEpayment )
		{
			return true;
		}

		return false;
	}

	/**
	 * requirementsWiPaypal( ) :
	 *
	 * @return	boolean
	 * @access private
	 * @internal #t2145
	 * @version    8.9.6
	 * @since      8.9.6
	 */
	private function requirementsWiPaypal()
	{
		switch( $this->paymentId )
		{
			case(14):
				return true;
			default:
				return false;
		}
	}

	/**
	 * requirementsWiPowermail( ) :
	 *
	 * @return	boolean
	 * @access private
	 * @version    4.0.6
	 * @since      4.0.6
	 */
	private function requirementsWiPowermail()
	{
		return $this->powermailInAction();
	}

	/*	 * *********************************************
	 *
	 * Prompts
	 *
	 * ******************************************** */

	/**
	 * serverPrompt( ):  Prompts the items of $prompts below the submit buttons.
	 *                   An item has the format: marker|type|prompt
	 *                   * marker  :
	 *                               * SERVER_PROMPT_WICLOSE_SUBPART
	 *                               * SERVER_PROMPT_WOCLOSE_SUBPART
	 *                   * type    :
	 *                               * secondary (grey)
	 *                               * [empty!]  (blue)
	 *                               * success   (green)
	 *                               * error     (red)
	 *                   Example: SERVER_PROMPT_WOCLOSE_SUBPART|alert alert-danger|message
	 *
	 * @return	void
	 * @access private
	 * @version     7.4.5
	 * @since       4.0.6
	 */
	private function serverPrompt()
	{

		$marker = array();
		$server_prompt = null;

		// #t1369, 170729, dwildt, 3+
		foreach( ( array ) $this->prompts as $prompt )
		{
			// #t1369, 170729, dwildt, 3+
			if( is_bool( $prompt ) )
			{
				continue;
			}
			list( $subpartMarker, $marker[ '###TYPE###' ], $marker[ '###PROMPT###' ] ) = explode( '|', $prompt );
			$subpart = $this->template( '###' . $subpartMarker . '###' );
			$subpart = $this->pObj->MarkerBasedTemplateService->substituteMarkerArray( $subpart, $marker );
			$server_prompt = $server_prompt
							. $subpart;
		}

		unset( $marker );
		$marker[ '###SERVER_PROMPT###' ] = $server_prompt;
		$this->content = $this->pObj->MarkerBasedTemplateService->substituteMarkerArray( $this->content, $marker );
	}

	/*	 * *********************************************
	 *
	 * Session
	 *
	 * ******************************************** */

	/**
	 * session( ):
	 *
	 * @param	integer		$paymentId  : current payment id. 1: credit card, 2: elv. 3: sepa (elv-iban).
	 * @return	void
	 * @access private
	 * @version     4.0.6
	 * @since       4.0.6
	 */
	private function session()
	{

		switch( true )
		{
			case( $this->requirementsWoEpayment() ):
				$this->sessionDataAddWoEpayment();
				break;
			case( $this->paymillLib->getPaymillToken() ):
				$this->sessionDataAddWiToken();
				break;
			case( $this->paymentId == 14 ):
				$this->sessionDataAddWiPaypal();
				break;
			default:
				$this->sessionDataRemove();
				break;
		}
	}

	/**
	 * sessionDataAddWiPaypal( ):
	 *
	 * @param	integer		$paymentId  : current payment id. 1: credit card, 2: elv. 3: sepa (elv-iban).
	 * @return	void
	 * @access private
	 * @internal #t2145
	 * @version     8.9.6
	 * @since       8.9.6
	 */
	private function sessionDataAddWiPaypal()
	{
		$sesArray = $GLOBALS[ 'TSFE' ]->fe_user->getKey( 'ses', $this->extKey . '_' . $this->pid );

		unset( $sesArray[ 'e-payment' ] );

//		$sesArray[ 'e-payment' ][ 'paymill' ][ 'token' ] = $this->paymillLib->getPaymillToken();
//		$sesArray[ 'e-payment' ][ 'paymill' ][ 'client' ][ 'id' ] = $this->paymillLib->getPaymillClientId();
//		$sesArray[ 'e-payment' ][ 'paymill' ][ 'payment' ][ 'id' ] = $this->paymillLib->getPaymillPaymentId();
		$sesArray[ 'e-payment' ][ 'woEpayment' ] = false;

		$GLOBALS[ 'TSFE' ]->fe_user->setKey( 'ses', $this->extKey . '_' . $this->pid, $sesArray );
		// save session
		$GLOBALS[ 'TSFE' ]->storeSessionData();
//    var_dump( __FILE__, __LINE__, $this->extKey . '_' . $this->pid, $sesArray, $GLOBALS['TSFE']->fe_user->getKey( 'ses', $this->extKey . '_' . $this->pid ) );
	}

	/**
	 * sessionDataAddWiToken( ):
	 *
	 * @param	integer		$paymentId  : current payment id. 1: credit card, 2: elv. 3: sepa (elv-iban).
	 * @return	void
	 * @access private
	 * @version     4.0.6
	 * @since       4.0.6
	 */
	private function sessionDataAddWiToken()
	{
		$sesArray = $GLOBALS[ 'TSFE' ]->fe_user->getKey( 'ses', $this->extKey . '_' . $this->pid );

		// #i0050, 140310, dwildt, 1-
		//unset( $sesArray['e-payment']['paymill'] );
		// #i0050, 140310, dwildt, 1+
		unset( $sesArray[ 'e-payment' ] );

		$sesArray[ 'e-payment' ][ 'paymill' ][ 'token' ] = $this->paymillLib->getPaymillToken();
		$sesArray[ 'e-payment' ][ 'paymill' ][ 'client' ][ 'id' ] = $this->paymillLib->getPaymillClientId();
		$sesArray[ 'e-payment' ][ 'paymill' ][ 'payment' ][ 'id' ] = $this->paymillLib->getPaymillPaymentId();
		$sesArray[ 'e-payment' ][ 'woEpayment' ] = false;

		$GLOBALS[ 'TSFE' ]->fe_user->setKey( 'ses', $this->extKey . '_' . $this->pid, $sesArray );
		// save session
		$GLOBALS[ 'TSFE' ]->storeSessionData();
//    var_dump( __FILE__, __LINE__, $this->extKey . '_' . $this->pid, $sesArray, $GLOBALS['TSFE']->fe_user->getKey( 'ses', $this->extKey . '_' . $this->pid ) );
	}

	/**
	 * sessionDataAddWoEpayment( ):
	 *
	 * @param	integer		$paymentId  : current payment id. 1: credit card, 2: elv. 3: sepa (elv-iban).
	 * @return	void
	 * @internal #i0050
	 * @access private
	 * @version     4.0.9
	 * @since       4.0.9
	 */
	private function sessionDataAddWoEpayment()
	{
		$sesArray = $GLOBALS[ 'TSFE' ]->fe_user->getKey( 'ses', $this->extKey . '_' . $this->pid );

		// #i0050, 140310, dwildt, 1-
		//unset( $sesArray['e-payment']['paymill'] );
		// #i0050, 140310, dwildt, 1+
		unset( $sesArray[ 'e-payment' ] );

		$sesArray[ 'e-payment' ][ 'woEpayment' ] = true;

		$GLOBALS[ 'TSFE' ]->fe_user->setKey( 'ses', $this->extKey . '_' . $this->pid, $sesArray );
		// save session
		$GLOBALS[ 'TSFE' ]->storeSessionData();
//    var_dump( __FILE__, __LINE__, $this->extKey . '_' . $this->pid, $sesArray, $GLOBALS['TSFE']->fe_user->getKey( 'ses', $this->extKey . '_' . $this->pid ) );
	}

	/**
	 * sessionDataRemove( ):
	 *
	 * @param	integer		$paymentId  : current payment id. 1: credit card, 2: elv. 3: sepa (elv-iban).
	 * @return	void
	 * @access private
	 * @version     8.10.2
	 * @since       4.0.6
	 */
	private function sessionDataRemove( $force = false )
	{

		switch( true )
		{
			case( $force ):
				// follow the workflow
				break;
			case( $this->requirementsToken() ):
			case( $this->requirementsWiPowermail() ):
				return;
			default:
				// follow the workflow
				break;
		}

		unset( $force );

//$prompt = 'debug trail: ' . DebugUtility::debugTrail( );
//var_dump( __METHOD__, __LINE__, $prompt);

		$sesArray = $GLOBALS[ 'TSFE' ]->fe_user->getKey( 'ses', $this->extKey . '_' . $this->pid );
		// :TODO: Recalculte option costs (option costs without payment)
		// Remove paymill data
		// #i0050, 140310, dwildt, 1-
		//unset( $sesArray['e-payment']['paymill'] );
		// #i0050, 140310, dwildt, 1+
		unset( $sesArray[ 'e-payment' ] );
		// Remove payment id
		unset( $sesArray[ 'payment' ] );
		// Remove options payment
		unset( $sesArray[ 'options' ][ 'payment' ] );

		$GLOBALS[ 'TSFE' ]->fe_user->setKey( 'ses', $this->extKey . '_' . $this->pid, $sesArray );
		// save session
		$GLOBALS[ 'TSFE' ]->storeSessionData();
//var_dump( __FILE__, __LINE__, $GLOBALS['TSFE']->fe_user->getKey( 'ses', $this->extKey . '_' . $this->pid ) );
	}

	/*	 * *********************************************
	 *
	 * Setting methods
	 *
	 * ******************************************** */

	/**
	 * setParentObject( )  : Returns a caddy with HTML form and HTML options among others
	 *
	 * @param	[type]		$$pObj: ...
	 * @return	void
	 * @access public
	 * @version    2.0.0
	 * @since      2.0.0
	 */
	public function setParentObject( $pObj )
	{
		if( !is_object( $pObj ) )
		{
			$prompt = 'ERROR: no parent object!<br />' . PHP_EOL .
							'Sorry for the trouble.<br />' . PHP_EOL .
							'TYPO3 Caddy<br />' . PHP_EOL .
							__METHOD__ . ' (' . __LINE__ . ')';
			die( $prompt );
		}
		$this->pObj = $pObj;

//$prompt = 'debug trail: ' . DebugUtility::debugTrail( ) . PHP_EOL .
//          'TYPO3 Caddy<br />' . PHP_EOL .
//        __METHOD__ . ' (' . __LINE__ . ')';
//echo $prompt;

		if( !is_object( $pObj->drs ) )
		{
			$prompt = 'ERROR: no DRS object!<br />' . PHP_EOL .
							'Sorry for the trouble.<br />' . PHP_EOL .
//                'debug trail: ' . DebugUtility::debugTrail( ) . PHP_EOL .
							'TYPO3 Caddy<br />' . PHP_EOL .
							__METHOD__ . ' (' . __LINE__ . ')';
			die( $prompt );
		}
		$this->drs = $pObj->drs;
	}

	/*	 * *********************************************
	 *
	 * Template
	 *
	 * ******************************************** */

	/**
	 * template( ): Returns the template
	 *
	 * @param	string		$subpart  :
	 * @return	string		$template : HTML template
	 * @access private
	 * @version     8.7.0
	 * @since       4.0.6
	 */
	private function template( $subpart )
	{
		$cObj = $this->pObj->cObj;
		$conf = $this->pObj->conf;
		// #t1817, 171213, dwildt, 1-/2+
		//$template = $cObj->fileResource( $conf[ 'api.' ][ 'e-payment.' ][ 'provider.' ][ 'paymill.' ][ 'files.' ][ 'html' ] );
		$absFileName = \TYPO3\CMS\Core\Utility\GeneralUtility::getFileAbsFileName( $conf[ 'api.' ][ 'e-payment.' ][ 'provider.' ][ 'paymill.' ][ 'files.' ][ 'html' ] );
		$template = file_get_contents( $absFileName );

		// Die if there isn't any HTML template
		if( empty( $template ) )
		{
			$prompt = '
        <div style="border:1em solid red;color:red;padding:1em;text-align:center">
          <h1>
            ERROR: No HTML Template
          </h1>
          <p>
            There isn\'t any HTML template.
          </p>
          <p>
            Please check your template configuration. See
          </p>
          <ul>
            <li>
              TypoScript Constant Editor: [CADDY - E-PAYMENT - PAYMILL FILES]
            </li>
            <li>
              TypoScript Object Browser: plugin.tx_caddy_pi1.api.e-payment.provider.paymill.files.html
            </li>
          </ul>
          <p>
            ' . __METHOD__ . ' (' . __LINE__ . ')
          </p>
          <p>
            Sorry for the trouble.<br 7>
            Caddy - the TYPO3 Shopping Cart
          </p>
        </div>
        ';
			die( $prompt );
		}

		$template = $this->pObj->MarkerBasedTemplateService->getSubpart( $template, $subpart );

		// Die if there isn't any HTML template
		if( empty( $template ) )
		{
			$debugTrail = 'debug trail: ' . DebugUtility::debugTrail();
			$debugTrail = str_replace( '//', '<br />' . PHP_EOL . '//', $debugTrail );
//var_dump( __METHOD__, __LINE__, $prompt);
			$prompt = '
        <div style="border:1em solid red;color:red;padding:1em;text-align:center">
          <h1>
            ERROR: HTML Template without subpart ' . $subpart . '
          </h1>
          <p>
            HTML template doesn\'t contain the subpart ' . $subpart . '
          </p>
          <p>
            Please check your template configuration. See
          </p>
          <ul>
            <li>
              TypoScript Constant Editor: [CADDY - E-PAYMENT - PAYMILL FILES]
            </li>
            <li>
              TypoScript Object Browser: plugin.tx_caddy_pi1.api.e-payment.provider.paymill.files.html
            </li>
          </ul>
          <p>
            ' . __METHOD__ . ' (' . __LINE__ . ')
          </p>
          <p>
            Sorry for the trouble.<br 7>
            Caddy - the TYPO3 Shopping Cart
          </p>
          <p>
            debug trial: ' . $debugTrail . ')
          </p>
        </div>
        ';
			die( $prompt );
		}

		return $template;
	}

	/**
	 * templateMode( ) :
	 *
	 * @return	void
	 * @access private
	 * @version    4.0.6
	 * @since      4.0.6
	 */
	private function templateMode()
	{
		//plugin.tx_caddy_pi1.api.e-payment.provider.paymill.mode
		$mode = $this->conf[ 'api.' ][ 'e-payment.' ][ 'provider.' ][ 'paymill.' ][ 'mode' ];

		switch( $mode )
		{
			case( 'live' ):
				$this->templateModeLive();
				break;
			case( 'test' ):
			default:
				$this->templateModeTest();
				break;
		}
	}

	/**
	 * templateModeLive( ) :
	 *
	 * @return	void
	 * @access private
	 * @version    4.0.6
	 * @since      4.0.6
	 */
	private function templateModeLive()
	{
		$this->content = str_replace( '###TEST_OR_LIVE_MODE###', 'live', $this->content );
		$this->content = $this->pObj->MarkerBasedTemplateService->substituteSubpart( $this->content, '###PAYMILL_TEST_MODE###', null );
	}

	/**
	 * templateModeTest( ) :
	 *
	 * @return	void
	 * @access private
	 * @version    4.0.6
	 * @since      4.0.6
	 */
	private function templateModeTest()
	{
		$this->content = str_replace( '###TEST_OR_LIVE_MODE###', 'test', $this->content );

		$marker = array(
				'###PAYMILL_LIVE_MODE_PROMPT###' => $this->pi_getLL( 'paymill-live-mode' ),
				'###PAYMILL_TEST_MODE_PROMPT###' => $this->pi_getLL( 'paymill-test-mode' )
		);
		$subpart = $this->template( '###PAYMILL_TEST_MODE###' );
		$subpart = $this->pObj->MarkerBasedTemplateService->substituteMarkerArray( $subpart, $marker );

		$this->content = $this->pObj->MarkerBasedTemplateService->substituteSubpart( $this->content, '###PAYMILL_TEST_MODE###', $subpart );

		$this->templateModeTestData();
	}

	/**
	 * templateModeTestData( ) :
	 *
	 * @return	void
	 * @access private
	 * @version    4.0.6
	 * @since      4.0.6
	 */
	private function templateModeTestData()
	{
		if( $this->paymillLib->getPaymillToken() )
		{
//var_dump( __METHOD__, __LINE__ );
			return;
		}

		if( $this->templateModeTestDataWiPowermail() )
		{
//var_dump( __METHOD__, __LINE__ );
			return;
		}
		$this->templateModeTestDataCreditcard();
		$this->templateModeTestDataElv();
		$this->templateModeTestDataElvIban();
//var_dump( __METHOD__, __LINE__ );
	}

	/**
	 * templateModeTestDataCreditcard( ) :
	 *
	 * @return	void
	 * @access private
	 * @version    4.0.6
	 * @since      4.0.6
	 */
	private function templateModeTestDataCreditcard()
	{

		$marker = array(
				'###VALUE_CREDITCARD-NUMBER###' => '4111111111111111', // Visa
				'###VALUE_CREDITCARD-EXPIRY###' => strftime( '%m/%g', strtotime( "+2 years" ) ),
				'###VALUE_CREDITCARD-HOLDER###' => 'Alois Schäfer',
				'###VALUE_CREDITCARD-CVC###' => '123'
		);

		$this->content = $this->pObj->MarkerBasedTemplateService->substituteMarkerArray( $this->content, $marker );
	}

	/**
	 * templateModeTestDataElv( ) :
	 *
	 * @return	void
	 * @access private
	 * @version    6.0.0
	 * @since      4.0.6
	 */
	private function templateModeTestDataElv()
	{
		$marker = array(
				'###VALUE_ELV-ACCOUNT###' => '648489890',
				'###VALUE_ELV-BANKCODE###' => '50010517',
				'###VALUE_ELV-HOLDER###' => 'Chris Hansen'
		);

		$this->content = $this->pObj->MarkerBasedTemplateService->substituteMarkerArray( $this->content, $marker );
	}

	/**
	 * templateModeTestDataElvIban( ) :
	 *
	 * @return	void
	 * @access private
	 * @version    4.0.6
	 * @since      4.0.6
	 */
	private function templateModeTestDataElvIban()
	{
		$marker = array(
				'###VALUE_ELV-IBAN###' => 'DE12500105170648489890',
				'###VALUE_ELV-BIC###' => 'BENEDEPPYYY',
				'###VALUE_ELV-IBAN-HOLDER###' => 'Alex Tabo'
		);

		$this->content = $this->pObj->MarkerBasedTemplateService->substituteMarkerArray( $this->content, $marker );
	}

	/**
	 * templateModeTestDataWiPowermail( ) :
	 *
	 * @return	boolean
	 * @access private
	 * @version    4.0.6
	 * @since      4.0.6
	 */
	private function templateModeTestDataWiPowermail()
	{
		return $this->powermailInAction();
	}

	/**
	 * templateSectionActive( ) : Replace the ###ACTIVE### marker
	 *
	 * @param	string		$template : HTML template
	 * @return	string		$template : HTML template
	 * @access private
	 * @version    8.12.0
	 * @since      4.0.6
	 */
	private function templateSectionActive( $template )
	{
//var_dump( __METHOD__, __LINE__, $template);
		// Set default value;
		$active = 1; // 1: credit card, 2: elv. 3: sepa (elv-iban).
		if( $this->paymentId !== null )
		{
			$active = $this->paymentId;
		}

			// #i0255, 180909, dwildt, 3+
		$method = $this->conf[ '_HTMLMARKER.' ][ 'cssAccordionActive' ];
		$conf = $this->conf[ '_HTMLMARKER.' ][ 'cssAccordionActive.' ];
		$cssAccordionActive = ContentObjectRenderer::cObjGetSingle( $this->pObj->cObj, $method, $conf );

		$markers = array(
				0 => '###CADDY_PAYMENT_SECTION_' . $active . '###',
				1 => '###CADDY_PAYMENT_TYPE_' . $active . '###',
					// #i0255, 180909, dwildt, 1+
				2 => '###CADDY_SECTION_' . $active . '###',
		);

		$values = array(
				0 => 'active',
				1 => 'payment-type-active',
				2 => $cssAccordionActive,
		);

		foreach( $markers as $key => $marker )
		{
			$value = $values[ $key ];
			$template = str_replace( $marker, ' ' . $value . ' ', $template );
			$template = str_replace( '  ' . $value, ' ' . $value, $template );
			$template = str_replace( $value . '  ', $value . ' ', $template );
			$template = str_replace( '" ' . $value, '"' . $value, $template );
			$template = str_replace( $value . ' "', $value . '"', $template );
		}

		return $template;
	}

	/**
	 * templateSectionRemove( ) :  Remove the subpart ###SECTION_X###, if correspond payment method is disabled.
	 *                             X is the key of the payment method.
	 *
	 * @param	string		$template : HTML template
	 * @return	string		$template : HTML template
	 * @access private
	 * @version    4.0.8
	 * @since      4.0.8
	 */
	private function templateSectionRemove( $template )
	{
		$conf = $this->pObj->conf;
		$paymentOptions = $conf[ 'api.' ][ 'options.' ][ 'payment.' ][ 'options.' ];

		foreach( array_keys( ( array ) $paymentOptions ) as $key )
		{
			if( stristr( $key, '.' ) )
			{
				continue;
			}

			$enabled = $paymentOptions[ $key . '.' ][ 'enabled' ];

			if( $enabled )
			{
				continue;
			}

			$section = '###SECTION_' . $key . '###';
			$template = $this->pObj->MarkerBasedTemplateService->substituteSubpart( $template, $section, null );
		}

		return $template;
	}

}