<?php

namespace Netzmacher\Caddy\Lib\EPayment\Paymill;

/* * *************************************************************
 *  Copyright notice
 *
 *  (c) 2018-2020 - Dirk Wildt <http://wildt.at.die-netzmacher.de>
 *  All rights reserved
 *
 *  This script is part of the TYPO3 project. The TYPO3 project is
 *  free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  The GNU General Public License can be found at
 *  http://www.gnu.org/copyleft/gpl.html.
 *
 *  This script is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  This copyright notice MUST APPEAR in all copies of the script!
 * ************************************************************* */

/**
 * @author			Dirk Wildt <http://wildt.at.die-netzmacher.de>
 * @package			TYPO3
 * @subpackage  tx_caddy
 * @version     8.10.2
 * @since       8.10.2
 */
class Paypal
{

	/**
	 * @var object
	 */
	private $_pObj;

	/**
	 * @var boolean
	 */
	static private $_promptToConsole = FALSE;

	/**
	 * @var string
	 */
	static private $_extKey = 'caddy';

	/**
	 * @var string
	 */
	private $_Checksum;

	/**
	 * @param			object	$pObj		Parent object
	 * @return		void
	 * @access		public
	 * @version   8.10.2
	 * @since     8.10.2
	 */
	public function __construct( $pObj )
	{
		$this->_pObj = $pObj;
	}

	/**
	 * The manager controls the workflow process:
	 * 				- If PayPal is payed, page will reloaded with changed params
	 * 				- If current payment method is PayPal, but PayPal isn't payed, page will forwarded to PayPal
	 * 				- If current payment method isn't PayPal
	 * 
	 * 
	 * @return		void
	 * @access		public
	 * @version   8.10.2
	 * @since     8.10.2
	 */
	public function Manager()
	{
		$this->_initDebugging();

		// RETURN: Powermail action isn't create
		if( !SELF::_isPowermailActionCreate() )
		{
			return;
		}

		// If requirements match, page will reloaded with changed params
		if( $this->_GoToShop() )
		{
			return;
		}
		// If requirements match, page will forwarded to PayPal
		$this->_GoToPaypal();
		// If no requirement doesn't match, the workflow of the Browser plugin will continued
	}

	/**
	 * @return		void
	 * @access		private
	 * @version   8.10.2
	 * @since     8.10.2
	 */
	private function _GoToPaypal()
	{
		if( !$this->_Requirements() )
		{
			return;
		}

		$this->_Workflow();
	}

	/**
	 * @return		boolean
	 * @access		private
	 * @version   8.10.2
	 * @since     8.10.2
	 */
	private function _GoToShop()
	{
		// RETURN: Buying isn't payed with PayPal yet
		if( !SELF::_getSessionValuePaypalIsPayed() )
		{
			$data = '_getSessionValuePaypalIsPayed is FALSE';
			\Netzmacher\Caddy\Utility\Debug\Console::log( $data, __METHOD__, __LINE__, SELF::$_promptToConsole );
			return FALSE;
		}

		$data = '_getSessionValuePaypalIsPayed is TRUE';
		\Netzmacher\Caddy\Utility\Debug\Console::log( $data, __METHOD__, __LINE__, SELF::$_promptToConsole );

//		$data = SELF::_getSessionValueEPayment();
//		\Netzmacher\Caddy\Utility\Debug\Console::log( $data, __METHOD__, __LINE__, SELF::$_promptToConsole );

		return $this->_forwardToShop();
	}

	/**
	 * @return		void
	 * @access		private
	 * @version   8.10.3
	 * @since     8.10.3
	 */
	private function _HtmlPaypal()
	{
		echo '
<!DOCTYPE html>
<html lang="de" class="no-js">
<head>
	<meta charset="utf-8">
	<!-- 
		Developed by https://die-netzmacher.de
	-->
	<title>PayPal Kaufabwicklung</title>
	<script type="text/javascript" src="https://bridge.paymill.com/"></script>
	<script type="text/javascript" >
		var PAYMILL_PUBLIC_KEY = \'' . PAYMILL_PUBLIC_KEY . '\';
		paymill.createTransaction({
			checksum: "' . $this->_Checksum . '"
		}, function(error) {
			if (error) {
				// Payment setup failed, handle error and try again.
				console.log(error);
				setTimeout(function() {
						history.go(-1);
				}, 0);
			}
		});
	</script>
</head>
<body style="background:#EEEEEE;padding:1em;">
	<div style="background:#009cde;border-radius:10px;color:white;font-family:sans-serif;padding:1em;text-align:center;">
		<h1>Einen Moment</h1>
		<h2>Sie werden zu Paypal weitergeleitet ...</h2>
		<p style="margin-top: 3em;">
			<a href="javascript:history.back()" style="background:#002f86;border-radius:10px;font-weight:bold;padding: 1em;text-decoration: none;color: white;">Zurück zum Warenkorb</a>
		</p>
	</div>
</body>
'
		;
		exit();
	}

	/**
	 * @return		boolean
	 * @access		private
	 * @version   8.10.2
	 * @since     8.10.2
	 */
	private function _Requirements()
	{
		switch( TRUE )
		{
//			case(SELF::_getSessionValuePaypalIsPayed()):
//				$this->_forwardToShop();
//				return FALSE;
			case(!SELF::_isEPaymentPaypal() ):
				// E-Payment method isn't Paypal
				return FALSE;
			case( $this->_isEPaymentPaypalPayed() ):
				// Paypal is payed, 
				return FALSE;
			default:
				// E-Payment method is Paypal and Paypal isn't payed
				return TRUE;
		}
	}

	/**
	 * @return		boolean
	 * @access		private
	 * @version   8.10.2
	 * @since     8.10.2
	 */
	private function _Workflow()
	{
		//var_dump( __METHOD__, __LINE__, $_SERVER[ 'REQUEST_URI' ] );

		$this->_setChecksum();
		SELF::_setSessionPowermailPOST();
		$this->_HtmlPaypal();

		echo ''
		. '<div style="background:red;border:1em solid red;color:white;font-family:sans-serif;padding:1em;text-align:center;">'
		. '<h1>Error</h1>'
		. '<h2>This code should never reached</h2>'
		. '<p>'
		. '  Something went wrong with Paypal payment :(<br />'
		. '  Please select another payment method.'
		. '</p>'
		. '<p><a href="javascript:history.back()" style="color:white;font-weight:bold;">Go Back!</a></p>'
		. '<p>&nbsp;</p>'
		. '<p>'
		. '  Sorry for the trouble.<br />'
		. '  ' . __METHOD__ . ' (#' . __LINE__ . ')'
		. '</p>'
		. '</div>'
		;
		die();
	}

	/**
	 * @param			string	$url	
	 * @param			string	$method	
	 * @param			string	$line	
	 * return			void
	 * @version   8.10.3
	 * @since     8.10.3
	 */
	private function _dieUrlIsEmpty( $url, $method, $line )
	{
		if( !empty( $url ) )
		{
			return;
		}

		echo ''
		. '<div style="background:red;border:1em solid red;color:white;font-family:sans-serif;padding:1em;text-align:center;">'
		. '<h1>Error</h1>'
		. '<h2>URL is empty</h2>'
		. '<p><a href="javascript:history.back()" style="color:white;font-weight:bold;">Go Back!</a></p>'
		. '<p>&nbsp;</p>'
		. '<p>Sorry for the trouble.</p>'
		. '<p>' . $method . ' (#' . $line . ')</p>'
		. '</div>'
		;

		die();
	}

	/**
	 * @param			string	$url	
	 * @param			string	$method	
	 * @param			string	$line	
	 * return			void
	 * @version   8.10.3
	 * @since     8.10.3
	 */
	static private function _dieWorkflowError( $method, $line )
	{
		echo ''
		. '<div style="background:red;border:1em solid red;color:white;font-family:sans-serif;padding:1em;text-align:center;">'
		. '<h1>Error</h1>'
		. '<h2>Code should never reach this code</h2>'
		. '<p><a href="javascript:history.back()" style="color:white;font-weight:bold;">Go Back!</a></p>'
		. '<p>&nbsp;</p>'
		. '<p>Sorry for the trouble.</p>'
		. '<p>' . $method . ' (#' . $line . ')</p>'
		. '</div>'
		;

		die();
	}

	/**
	 * @access private
	 * @return      void
	 * @version     8.10.2
	 * @since       8.10.2
	 */
	static private function _extConfSerialize()
	{
		if( is_array( $GLOBALS[ 'TYPO3_CONF_VARS' ][ 'EXT' ][ 'extConf' ][ 'caddy' ] ) )
		{
			$GLOBALS[ 'TYPO3_CONF_VARS' ][ 'EXT' ][ 'extConf' ][ 'caddy' ] = serialize( $GLOBALS[ 'TYPO3_CONF_VARS' ][ 'EXT' ][ 'extConf' ][ 'caddy' ] );
		}
	}

	/**
	 * @access private
	 * @return      void
	 * @version     8.10.2
	 * @since       8.10.2
	 */
	static private function _extConfUnserialize()
	{
		if( !is_array( $GLOBALS[ 'TYPO3_CONF_VARS' ][ 'EXT' ][ 'extConf' ][ 'caddy' ] ) )
		{
			$GLOBALS[ 'TYPO3_CONF_VARS' ][ 'EXT' ][ 'extConf' ][ 'caddy' ] = unserialize( $GLOBALS[ 'TYPO3_CONF_VARS' ][ 'EXT' ][ 'extConf' ][ 'caddy' ] );
		}
	}

	/**
	 * @return		boolean
	 * @access		private
	 * @version   8.10.2
	 * @since     8.10.2
	 */
	private function _forwardToShop()
	{
		if( SELF::_forwardToShopWiPowermailPOST() )
		{
			return TRUE;
		}
		SELF::_forwardToShopByReturnUrl();
		SELF::_dieWorkflowError( __METHOD__, __LINE__ );
	}

	/**
	 * @return		void
	 * @access		private
	 * @version   8.10.2
	 * @since     8.10.2
	 */
	private function _forwardToShopByReturnUrl()
	{
		$data = 'Location: ' . SELF::_getSessionValuePaypalReturnUrl();
		\Netzmacher\Caddy\Utility\Debug\Console::log( $data, __METHOD__, __LINE__, SELF::$_promptToConsole );
		header( 'Location: ' . SELF::_getSessionValuePaypalReturnUrl() );
	}

	/**
	 * @return		void
	 * @access		private
	 * @version   8.10.2
	 * @since     8.10.2
	 */
	static private function _forwardToShopWiPowermailPOST()
	{
		if( SELF::_getUrl() != SELF::_getSessionValuePaypalReturnUrl() )
		{
			$data = 'Current URL != Return URL in session';
			\Netzmacher\Caddy\Utility\Debug\Console::log( $data, __METHOD__, __LINE__, SELF::$_promptToConsole );
			return FALSE;
		}

		$data = 'Set $_POST[tx_powermail_pi1]';
		\Netzmacher\Caddy\Utility\Debug\Console::log( $data, __METHOD__, __LINE__, SELF::$_promptToConsole );
		$_POST[ 'tx_powermail_pi1' ] = SELF::_getSessionValuePowermailPOST();
		return TRUE;
	}

	/**
	 * @return		integer
	 * @access		private
	 * @version   8.10.2
	 * @since     8.10.2
	 */
	static private function _getPid()
	{
		return $GLOBALS[ "TSFE" ]->id;
	}

	/**
	 * @return		string
	 * @access		private
	 * @version   8.10.2
	 * @since     8.10.2
	 */
	static private function _getUrl()
	{
		$url = (isset( $_SERVER[ 'HTTPS' ] ) ? "https" : "http") . '://' . $_SERVER[ 'SERVER_NAME' ] . $_SERVER[ 'REQUEST_URI' ];
		return $url;
	}

	/**
	 * @return		integer
	 * @access		private
	 * @version   8.10.2
	 * @since     8.10.2
	 */
	static private function _getSessionData()
	{
		return $GLOBALS[ 'TSFE' ]->fe_user->getKey( 'ses', SELF::$_extKey . '_' . SELF::_getPid() );
	}

	/**
	 * @return		integer
	 * @access		private
	 * @version   8.10.2
	 * @since     8.10.2
	 */
	static private function _getSessionValuePaypalIsPayed()
	{
		$sesArray = SELF::_getSessionData();
		return $sesArray[ 'e-payment' ][ 'paymill' ][ 'paypal' ][ 'isPayed' ];
	}

	/**
	 * @return		integer
	 * @access		private
	 * @version   8.10.2
	 * @since     8.10.2
	 */
	static private function _getSessionValuePaypalReturnUrl()
	{
		$sesArray = SELF::_getSessionData();
		return $sesArray[ 'e-payment' ][ 'paymill' ][ 'paypal' ][ 'returnUrl' ];
	}

	/**
	 * @return		integer
	 * @access		private
	 * @version   8.10.2
	 * @since     8.10.2
	 */
	static private function _getSessionValuePowermailPOST()
	{
		$sesArray = SELF::_getSessionData();
		return $sesArray[ 'tx_powermail_pi1' ][ 'POST' ];
	}

	/**
	 * @return		integer
	 * @access		private
	 * @version   8.10.2
	 * @since     8.10.2
	 */
	static private function _getSessionValueClientEMail()
	{
		$sesArray = SELF::_getSessionData();
		return $sesArray[ 'powermail' ][ 'values' ][ 'customer' ][ 'email' ];
	}

	/**
	 * @return		integer
	 * @access		private
	 * @version   8.10.2
	 * @since     8.10.2
	 */
	static private function _getSessionValueClientId()
	{
		$sesArray = SELF::_getSessionData();
		// 180326, dwildt, 1-/+: Workflow, because sesArray from below is lost while processing
		//$clientId = $sesArray[ 'e-payment' ][ 'paymill' ][ 'client' ][ 'id' ];
		$clientId =	$sesArray[ 'client' ][ 'id' ];


		$data = $clientId;
		\Netzmacher\Caddy\Utility\Debug\Console::log( $data, __METHOD__, __LINE__, SELF::$_promptToConsole );

		return $clientId;
	}

	/**
	 * @return		integer
	 * @access		private
	 * @version   8.10.2
	 * @since     8.10.2
	 */
	static private function _getSessionValueClientName()
	{
		$sesArray = SELF::_getSessionData();
		return $sesArray[ 'invoiceFirstname' ] . ' ' . $sesArray[ 'invoiceLastname' ];
	}

	/**
	 * @return		integer
	 * @access		private
	 * @version   8.10.2
	 * @since     8.10.2
	 */
	static private function _getSessionValueEPayment()
	{
		$sesArray = SELF::_getSessionData();
		return $sesArray[ 'e-payment' ];
	}

	/**
	 * @return		integer
	 * @access		private
	 * @version   8.10.2
	 * @since     8.10.2
	 */
	static private function _getSessionValuePaymentId()
	{
		$sesArray = SELF::_getSessionData();
		return $sesArray[ 'payment' ];
	}

	/**
	 * @return		boolean
	 * @access		private
	 * @version   8.10.2
	 * @since     8.10.2
	 */
	private function _initDebugging()
	{
		SELF::_ExtConfUnserialize();
		SELF::$_promptToConsole = $GLOBALS[ 'TYPO3_CONF_VARS' ][ 'EXT' ][ 'extConf' ][ 'caddy' ][ 'debuggingConsole' ];
		SELF::_ExtConfSerialize();
		if( SELF::$_promptToConsole )
		{
			$data = 'WARNING: Log to console is enabled by the Extension Manager of Caddy. This is a security risk!';
			\Netzmacher\Caddy\Utility\Debug\Console::log( $data, __METHOD__, __LINE__, SELF::$_promptToConsole );
		}
	}

	/**
	 * @return		boolean
	 * @access		private
	 * @version   8.10.2
	 * @since     8.10.2
	 */
	static private function _isEPaymentPaypal()
	{
		switch( SELF::_getSessionValuePaymentId() )
		{
			case(14):
				// 14: Paypal
				$data = 'E-Payment id = 14 (PayPal)';
				\Netzmacher\Caddy\Utility\Debug\Console::log( $data, __METHOD__, __LINE__, SELF::$_promptToConsole );
				return TRUE;
			default:
				$data = 'E-Payment id != 14. This is not PayPal';
				\Netzmacher\Caddy\Utility\Debug\Console::log( $data, __METHOD__, __LINE__, SELF::$_promptToConsole );
				return FALSE;
		}
	}

	/**
	 * @return		boolean
	 * @access		private
	 * @version   8.10.2
	 * @since     8.10.2
	 */
	private function _isEPaymentPaypalPayed()
	{
		// Get the Paymill transaction uid from URL
		$transactionId = \TYPO3\CMS\Core\Utility\GeneralUtility::_GET( 'paymill_trx_id' );

		// RETURN FALSE: There isn't any transaction uid
		if( empty( $transactionId ) )
		{
			$data = 'Is not set: $_GET[ paymill_trx_id ]';
			\Netzmacher\Caddy\Utility\Debug\Console::log( $data, __METHOD__, __LINE__, SELF::$_promptToConsole );
			return FALSE;
		}

		$data = 'Is set: $_GET[ paymill_trx_id ]';
		\Netzmacher\Caddy\Utility\Debug\Console::log( $data, __METHOD__, __LINE__, SELF::$_promptToConsole );
		// Autoload Paymill PHP API
		$_Autoload = \TYPO3\CMS\Core\Utility\GeneralUtility::makeInstance( \Netzmacher\Caddy\Lib\EPayment\Paymill\Autoload::class, $this->_pObj );
		$_Autoload->PaymillPhpAPI();

		// Get response from the known transaction
		$request = new \Paymill\Request( PAYMILL_API_KEY );
		$transaction = new \Paymill\Models\Request\Transaction();
		$transaction->setId( $transactionId );
		$response = $request->getOne( $transaction );

		switch( TRUE )
		{
			case($response->getUpdatedAt() > $response->getCreatedAt() ):
				// RETURN FALSE: transaction was updated
				$data = 'Transaction is updated: return TRUE';
				\Netzmacher\Caddy\Utility\Debug\Console::log( $data, __METHOD__, __LINE__, SELF::$_promptToConsole );
				return TRUE;
			case( empty( $response->getUpdatedAt() )):
				// RETURN FALSE: transaction was never updated
				$data = 'Transaction was never updated: return FALSE';
				\Netzmacher\Caddy\Utility\Debug\Console::log( $data, __METHOD__, __LINE__, SELF::$_promptToConsole );
				return FALSE;
			case($response->getUpdatedAt() == $response->getCreatedAt() ):
			default:
				// follow the workflow
				$data = 'Transaction creation and update has the same timestamp';
				\Netzmacher\Caddy\Utility\Debug\Console::log( $data, __METHOD__, __LINE__, SELF::$_promptToConsole );
				break;
		}

		// evaluate response code
		switch( $response->getResponseCode() )
		{
			case(20000):
				// 20000 marks a successful transaction
				// follow the workflow
				$data = 'Transaction response code is 20000 (succesful transaction).';
				\Netzmacher\Caddy\Utility\Debug\Console::log( $data, __METHOD__, __LINE__, SELF::$_promptToConsole );
				break;
			default:
				return FALSE;
		}

		// Block transaction for for another paypal payment check
		$this->_setPaypalToPayed( $transactionId );
		$this->_forwardToShop();
		SELF::_dieWorkflowError( __METHOD__, __LINE__ );
	}

	/**
	 * @return		boolean
	 * @access		private
	 * @version   8.10.2
	 * @since     8.10.2
	 */
	static private function _isPowermailActionCreate()
	{
		$_paramGet = \TYPO3\CMS\Core\Utility\GeneralUtility::_GET( 'tx_powermail_pi1' );
		if( $_paramGet[ 'action' ] == 'create' )
		{
			$data = '$_GET[tx_powermail_pi1][action] = create';
			\Netzmacher\Caddy\Utility\Debug\Console::log( $data, __METHOD__, __LINE__, SELF::$_promptToConsole );
			return true;
		}
		$data = '$_GET[tx_powermail_pi1][action] != create';
		\Netzmacher\Caddy\Utility\Debug\Console::log( $data, __METHOD__, __LINE__, SELF::$_promptToConsole );
		return false;
	}

	/**
	 * @return		void
	 * @access		private
	 * @version   8.10.3
	 * @since     8.10.3
	 */
	private function _setChecksum()
	{
		$_Autoload = \TYPO3\CMS\Core\Utility\GeneralUtility::makeInstance( \Netzmacher\Caddy\Lib\EPayment\Paymill\Autoload::class, $this->_pObj );
		$_Autoload->PaymillPhpAPI();

		$request = new \Paymill\Request( PAYMILL_API_KEY );
		$_checksum = new \Paymill\Models\Request\Checksum();
		$_checksum
						->setChecksumType( \Paymill\Models\Request\Checksum::TYPE_PAYPAL ) // Checksum type
						->setAmount( SELF::_setChecksumAmount() ) // e.g. "4200" for 42.00 EUR
						->setClient( SELF::_setClientId( $request ) ) // Client ID
						->setCurrency( 'EUR' ) // Alpha-3 country code
						->setReturnUrl( $this->_setChecksumReturnUrl() )
						->setCancelUrl( $this->_setChecksumCancelUrl() )
		;

		$_response = $request->create( $_checksum );
		$this->_Checksum = $_response->getId();

		$data = 'Checksum is set.';
		\Netzmacher\Caddy\Utility\Debug\Console::log( $data, __METHOD__, __LINE__, SELF::$_promptToConsole );
//var_dump(__METHOD__, __LINE__, $_response, $this->_Checksum);
//die();
	}

	/**
	 * @param			object		$request
	 * @return		object		$response
	 * @access		private
	 * @version   8.10.2
	 * @since     8.10.2
	 */
	static private function _setClient( $request )
	{
		$clientId = SELF::_getSessionValueClientId();
		switch( TRUE )
		{
			case($clientId):
				return SELF::_setClientUpdate( $request, $clientId );
			case(!$clientId):
			default:
				return SELF::_setClientCreate( $request );
		}
	}

	/**
	 * @param			object		$request
	 * @return		object		$response
	 * @access		private
	 * @version   8.10.2
	 * @since     8.10.2
	 */
	static private function _setClientCreate( $request )
	{
		$client = new \Paymill\Models\Request\Client();
		$client
						->setEmail( SELF::_getSessionValueClientEMail() )
						->setDescription( SELF::_getSessionValueClientName() )
		;

		$response = $request->create( $client );

		$data = $response;
		\Netzmacher\Caddy\Utility\Debug\Console::log( $data, __METHOD__, __LINE__, SELF::$_promptToConsole );

//		var_dump(__METHOD__, __LINE__, $response, $sesArray);
//		die();
		return $response;
	}

	/**
	 * @param			object		$request
	 * @param			integer		$clientId
	 * @return		object		$response
	 * @access		private
	 * @version   8.10.2
	 * @since     8.10.2
	 */
	static private function _setClientUpdate( $request, $clientId )
	{
		$client = new \Paymill\Models\Request\Client();
		$client->setId( $clientId )
						->setEmail( SELF::_getSessionValueClientEMail() )
						->setDescription( SELF::_getSessionValueClientName() )
		;

		$response = $request->update( $client );

		$data = $response;
		\Netzmacher\Caddy\Utility\Debug\Console::log( $data, __METHOD__, __LINE__, SELF::$_promptToConsole );

		return $response;
	}

	/**
	 * @param			object		$request
	 * @return		integer   $clientId
	 * @access		private
	 * @version   8.10.2
	 * @since     8.10.2
	 */
	static private function _setClientId( $request )
	{
		$response = SELF::_setClient( $request );
		$clientId = $response->getId();
		SELF::_setSessionClientId( $clientId );

		return $clientId;
	}

	/**
	 * @return		integer
	 * @access		private
	 * @version   8.10.2
	 * @since     8.10.2
	 */
	static private function _setChecksumAmount()
	{
		$_sesArray = SELF::_getSessionData();
		$_sumsumgross = $_sesArray[ 'sumsumgross' ];
		$Amount = round( $_sumsumgross * 100, 0 );
		return $Amount;
	}

	/**
	 * @return		integer
	 * @access		private
	 * @version   8.10.2
	 * @since     8.10.2
	 */
	private function _setChecksumCancelUrl()
	{
		$name = 'TEXT';
		$conf = array(
				'typolink.' => array(
						'forceAbsoluteUrl' => ( string ) 1,
						'parameter' => ( string ) SELF::_getPid(),
						'returnLast' => 'url',
				),
		);
		$url = $this->_pObj->local_cObj->cObjGetSingle( $name, $conf );

		SELF::_dieUrlIsEmpty( $url, __METHOD__, __LINE__ );

		return $url;
	}

	/**
	 * @return		integer
	 * @access		private
	 * @version   8.10.2
	 * @since     8.10.2
	 */
	private function _setChecksumReturnUrl()
	{
		$url = SELF::_getUrl();
		SELF::_setSessionPaypalReturnUrl( $url );
		return $url;
	}

	/**
	 * @param			string		$transactionId
	 * @return		void
	 * @access		private
	 * @version   8.10.3
	 * @since     8.10.3
	 */
	private function _setPaypalToPayed( $transactionId )
	{
		$request = new \Paymill\Request( PAYMILL_API_KEY );
		$transaction = new \Paymill\Models\Request\Transaction();
		$transaction
						->setId( $transactionId )
						->setDescription( $this->_setPaypalToPayedDescription() )
		;

		$response = $request->update( $transaction );
		unset( $response );

		SELF::_setSessionPaypalIsPayed();
		SELF::_setSessionEPaymentIsPayed();

		$data = 'Update transaction. Update timestamp is a marker for the workflow.';
		\Netzmacher\Caddy\Utility\Debug\Console::log( $data, __METHOD__, __LINE__, SELF::$_promptToConsole );
	}

	/**
	 * @return		string
	 * @access		private
	 * @version   8.10.3
	 * @since     8.10.3
	 */
	private function _setPaypalToPayedDescription()
	{
		$sesArray = SELF::_getSessionData();
		// #i0235, 180412, dwildt, 2-/+
//		$labelNumberOrder = $GLOBALS[ 'LANG' ]->sL( 'LLL:EXT:caddy/Classes/Lib/EPayment/Paymill/locallang.xml:label-short-number-order' );
//		$labelNumberInvoice = $GLOBALS[ 'LANG' ]->sL( 'LLL:EXT:caddy/Classes/Lib/EPayment/Paymill/locallang.xml:label-short-number-invoice' );
		$labelNumberOrder = 'R-Nr.';
		$labelNumberInvoice = 'Best-Nr.';

		$numberOrder = $labelNumberOrder . ': ' . $sesArray[ 'numberOrderCurrent' ];
		$numberInvoice = $labelNumberInvoice . ': ' . $sesArray[ 'numberInvoiceCurrent' ];

		return $numberOrder . ', ' . $numberInvoice;
	}

	/**
	 * @param			array		$sesArray
	 * @return		void
	 * @access		private
	 * @version   8.10.3
	 * @since     8.10.3
	 */
	static private function _setSessionData( $sesArray )
	{
		$GLOBALS[ 'TSFE' ]->fe_user->setKey( 'ses', SELF::$_extKey . '_' . SELF::_getPid(), $sesArray );
		$GLOBALS[ 'TSFE' ]->storeSessionData();
	}

	/**
	 * @param			integer		$clientId
	 * @return		void
	 * @access		private
	 * @version   8.10.3
	 * @since     8.10.3
	 */
	static private function _setSessionClientId( $clientId )
	{
		$sesArray = SELF::_getSessionData();

		// 180326, dwildt, 1-/+: Workflow, because sesArray from below is lost while processing
		//$sesArray[ 'e-payment' ][ 'paymill' ][ 'client' ][ 'id' ] = $clientId;
		$sesArray[ 'client' ][ 'id' ] = $clientId;
		SELF::_setSessionData( $sesArray );

		$data = 'Client ID: ' . $clientId;
		\Netzmacher\Caddy\Utility\Debug\Console::log( $data, __METHOD__, __LINE__, SELF::$_promptToConsole );
	}

	/**
	 * @return		void
	 * @access		private
	 * @version   8.10.3
	 * @since     8.10.3
	 */
	static private function _setSessionEPaymentIsPayed()
	{
		$sesArray = SELF::_getSessionData();

		$sesArray[ 'e-payment' ][ 'order' ][ 'numberOrderCurrent' ] = $sesArray[ 'numberOrderCurrent' ];
		$sesArray[ 'e-payment' ][ 'order' ][ 'isPayed' ] = TRUE;
		SELF::_setSessionData( $sesArray );

		$data = 'Session: e-payment.order.numberOrderCurrent = ' . $sesArray[ 'numberOrderCurrent' ];
		\Netzmacher\Caddy\Utility\Debug\Console::log( $data, __METHOD__, __LINE__, SELF::$_promptToConsole );
		$data = 'Session: e-payment.order.isPayed = TRUE';
		\Netzmacher\Caddy\Utility\Debug\Console::log( $data, __METHOD__, __LINE__, SELF::$_promptToConsole );
	}

	/**
	 * @return		void
	 * @access		private
	 * @version   8.10.3
	 * @since     8.10.3
	 */
	static private function _setSessionPaypalIsPayed()
	{
		$sesArray = SELF::_getSessionData();

		$sesArray[ 'e-payment' ][ 'paymill' ][ 'paypal' ][ 'isPayed' ] = TRUE;
		SELF::_setSessionData( $sesArray );

		$data = 'Session: isPayed = TRUE';
		\Netzmacher\Caddy\Utility\Debug\Console::log( $data, __METHOD__, __LINE__, SELF::$_promptToConsole );
	}

	/**
	 * @param			string		$returnUrl
	 * @return		void
	 * @access		private
	 * @version   8.10.3
	 * @since     8.10.3
	 */
	static private function _setSessionPaypalReturnUrl( $returnUrl )
	{
		$sesArray = SELF::_getSessionData();

		$sesArray[ 'e-payment' ][ 'paymill' ][ 'paypal' ][ 'returnUrl' ] = $returnUrl;
		SELF::_setSessionData( $sesArray );

//		$sesArray = SELF::_getSessionData();
//		var_dump(__METHOD__, __LINE__, $sesArray);
//		die();
	}

	/**
	 * @return		void
	 * @access		private
	 * @version   8.10.3
	 * @since     8.10.3
	 */
	static private function _setSessionPowermailPOST()
	{
		if( !$_POST[ 'tx_powermail_pi1' ] )
		{
			return;
		}

		$sesArray = SELF::_getSessionData();

		$sesArray[ 'tx_powermail_pi1' ][ 'POST' ] = $_POST[ 'tx_powermail_pi1' ];
		SELF::_setSessionData( $sesArray );

//		$sesArray = SELF::_getSessionData();
//		var_dump(__METHOD__, __LINE__, $sesArray);
//		die();
	}

}
