<?php

namespace Netzmacher\Caddy\Lib\Powermail;

use Netzmacher\Caddy\Utility\Template\CleanupTemplateUtility;
use TYPO3\CMS\Core\Utility\GeneralUtility;

/* * *************************************************************
 *  Copyright notice
 *
 *  (c) 2013-2020 - Dirk Wildt <http://wildt.at.die-netzmacher.de>
 *  All rights reserved
 *
 *  This script is part of the TYPO3 project. The TYPO3 project is
 *  free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  The GNU General Public License can be found at
 *  http://www.gnu.org/copyleft/gpl.html.
 *
 *  This script is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  This copyright notice MUST APPEAR in all copies of the script!
 * ************************************************************* */

/**
 * Class 'pdf' for the 'caddy' extension.
 *
 * @author	Dirk Wildt <http://wildt.at.die-netzmacher.de>
 * @package	TYPO3
 * @subpackage	tx_caddy
 * @version     8.17.0
 * @since       2.0.0
 */
class Attachments extends \Netzmacher\Caddy\AbstractPlugin
{

// path for localisation file
	public $scriptRelPath = 'Resources/Private/Language/Plugins/pi1/class.tx_caddy_pi1.php';
	public $extKey = 'caddy';
	public $drsUserfunc = null;
	public $conf = null;
	private $confPdf = null;
	private $currentAttachment = null; // deliveryorder || invoice || revocation || terms
	public $local_cObj = null;
// [Object]
	private $pObj = null;
	private $pdfTemplateLastPage = 0;
	private $pdfPageCurrent = 1;
// [Object]
	private $tcpdf = null;
	public $tmpl = null;

	/*	 * *********************************************
	 *
	 * Caddy
	 *
	 * ******************************************** */

	/**
	 * caddy( ) : Generates the HTML caddy and write it to tcpdf
	 *
	 * @param string $attachment: delveryorder, invoice
	 * @return	void
	 * @version    6.2.0
	 * @since      2.0.0
	 */
	private function caddy( $attachment = 'deliveryorder' )
	{
		$caddy = $this->caddy->caddy();
		//var_dump( __METHOD__, __LINE__, $caddy );
		$marker = $caddy[ 'marker' ];
		$subparts = $caddy[ 'subparts' ];
		$tmpl = $caddy[ 'tmpl' ];
		unset( $caddy );

		$content = $content . $this->MarkerBasedTemplateService->substituteMarkerArrayCached( $tmpl, $marker, $subparts );

		$content = $this->dynamicMarkers->main( $content, $this ); // Fill dynamic locallang or typoscript markers
		$content = preg_replace( '|###.*?###|i', '', $content ); // Finally clear not filled markers
//    var_dump( __METHOD__, __LINE__, $content );
//    die( );
		// write the HTML content
		$body = $this->confPdf[ $attachment . '.' ][ 'content.' ][ 'caddy.' ][ 'body.' ];
		$this->tcpdfWriteHtmlCell( $body[ 'properties.' ], $content, 'caddy' );
	}

	/*	 * *********************************************
	 *
	 * Delivery Order
	 *
	 * ******************************************** */

	/**
	 * deliveryorder( ) : Generates the PDF delivery order and returns the path to the PDF file
	 *
	 * @return	string		$path : path to the rendered pdf file
	 * @access public
	 * @version    2.0.0
	 * @since      2.0.0
	 */
	public function deliveryorder()
	{
		$destFile = null;
		$destPath = null;

		// Init caddy pdf
		$this->init( 'deliveryorder' );

		// RETURN : any pdf is requested
		if( !$this->deliveryorderInit() )
		{
			return;
		}
		// RETURN : any pdf is requested
		// Get the caddy session
		$sesArray = $GLOBALS[ 'TSFE' ]->fe_user->getKey( 'ses', $this->extKey . '_' . $GLOBALS[ "TSFE" ]->id );

		// Add session data to the local cObj
		$this->local_cObj->start( $sesArray, $this->pObj->conf[ 'db.' ][ 'table' ] );

		// DRS
		if( $this->pObj->drsUserfunc )
		{
			$data = var_export( $this->local_cObj->data, true );
			$prompt = 'cObj->data: ' . $data;
			\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[INFO/COBJ] ' . $prompt, $this->extKey, 0 );
		}
		// DRS
		// Get the path of the destination file
		$destFile = $this->local_cObj->cObjGetSingle
						(
						$this->confPdf[ 'deliveryorder.' ][ 'filename' ], $this->confPdf[ 'deliveryorder.' ][ 'filename.' ]
		);

		$destPath = PATH_site . 'uploads/tx_caddy/' . $destFile;

		if( file_exists( $destPath ) )
		{
			if( $this->pObj->drsUserfunc )
			{
				$prompt = 'uploads/tx_caddy' . '/' . $destFile . ' exists and will overridden!';
				\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[WARN/USERFUNC] ' . $prompt, $this->extKey, 2 );
			}
		}

		$srceFile = $sesArray[ 'sendVendorDeliveryorder' ];
		if( empty( $srceFile ) )
		{
			$srceFile = $sesArray[ 'sendCustomerDeliveryorder' ];
		}
		if( empty( $srceFile ) )
		{
			$prompt = 'Can\'t get source file from session data ' .
							'sendCustomerDeliveryorder/sendVendorDeliveryorder<br />' . PHP_EOL .
							__METHOD__ . '(line ' . __LINE__ . ')';
			die( $prompt );
		}
		$this->tcpdfInit( $srceFile );
		$this->tcpdfAddPage();

		// Write the delivery order address
		$fallBackToInvoiceAddress = true;
		$this->deliveryorderAddress( $fallBackToInvoiceAddress );

		// Write the delivery order date
		$this->deliveryorderDate();

		// Write the delivery order number
		$this->deliveryorderNumbers();

		// Write the delivery order term of credit
		$this->deliveryorderTermOfCredit();

		// Write additional textblocks
		$this->deliveryorderAdditionalTextblocks();

		// Write the caddy
		$this->caddy( 'deliveryorder' );

		// Create the PDF
		$this->tcpdfOutput( $destPath );

		return $destPath;
	}

	/**
	 * deliveryorderAdditionalTextblocks( ) : Write addional text blocks
	 *
	 * @return	void
	 * @access private
	 * @version    2.0.0
	 * @since      2.0.0
	 */
	private function deliveryorderAdditionalTextblocks()
	{
		$additionalTextblocks = $this->confPdf[ 'deliveryorder.' ][ 'content.' ][ 'additionaltextblocks.' ];

// LOOP : additional textblocks
		foreach( array_keys( ( array ) $additionalTextblocks ) as $key )
		{
			if( !stristr( $key, '.' ) )
			{
				continue;
			}

			$this->writeTextblock( $additionalTextblocks[ $key ], 'deliveryorder.additionaltextblocks.' . $key );
		}
// LOOP : additional textblocks
	}

	/**
	 * deliveryorderAddress( ) : Write the delivery address. If delivery address isn't given,
	 *                           take the invoice address
	 *
	 * @param	[type]		$$fallBackToInvoiceAddress: ...
	 * @return	void
	 * @access private
	 * @version    2.0.0
	 * @since      2.0.0
	 */
	private function deliveryorderAddress( $fallBackToInvoiceAddress = false )
	{
// Get the body content
		$body = $this->confPdf[ 'deliveryorder.' ][ 'content.' ][ 'address.' ][ 'deliveryorder.' ][ 'body.' ];
		$htmlContent = $GLOBALS[ 'TSFE' ]->cObj->cObjGetSingle( $body[ 'content' ], $body[ 'content.' ] );
		$content = strip_tags( $htmlContent );
		$content = trim( $content );
		$content = str_replace( ' ', null, $content );
// Get the body content

		switch( true )
		{
			case(!empty( $content ) ):
				$deliveryorderaddress = $this->confPdf[ 'deliveryorder.' ][ 'content.' ][ 'address.' ][ 'deliveryorder.' ];
				$this->writeTextblock( $deliveryorderaddress, 'deliveryorderaddress' );
				break;
			case( empty( $content ) ):
			default:
// FALLBACK : take the invoice address
				if( $fallBackToInvoiceAddress )
				{
					$invoiceaddress = $this->confPdf[ 'deliveryorder.' ][ 'content.' ][ 'address.' ][ 'invoice.' ];
					$this->writeTextblock( $invoiceaddress, 'invoiceAddress' );
				}
// FALLBACK : take the invoice address
				break;
		}
		unset( $htmlContent );

		return;
	}

	/**
	 * deliveryorderDate( ) : Write the current date
	 *
	 * @return	void
	 * @access private
	 * @version    2.0.0
	 * @since      2.0.0
	 */
	private function deliveryorderDate()
	{
		$date = $this->confPdf[ 'deliveryorder.' ][ 'content.' ][ 'date.' ];
		$this->writeTextblock( $date, 'deliveryorderDate' );
	}

	/**
	 * deliveryorderInit( ) :  Return false, if delivery order should not sent to the customer or
	 *                         the vendor
	 *
	 * @return	boolean		$deliveryorderInit: false, if delivery order should not sent
	 * @access private
	 * @version    2.0.0
	 * @since      2.0.0
	 */
	private function deliveryorderInit()
	{
		$deliveryorderInit = null;

		$sesArray = $GLOBALS[ 'TSFE' ]->fe_user->getKey( 'ses', $this->extKey . '_' . $GLOBALS[ "TSFE" ]->id );

		switch( true )
		{
			case(!empty( $sesArray[ 'sendCustomerDeliveryorder' ] ) ):
			case(!empty( $sesArray[ 'sendVendorDeliveryorder' ] ) ):
				$deliveryorderInit = true;
				break;
			default:
				$deliveryorderInit = false;
				break;
		}
		unset( $sesArray );

		if( !$deliveryorderInit )
		{
			if( $this->pObj->drsUserfunc )
			{
				$prompt = __METHOD__ . ' returns null.';
				\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[INFO/USERFUNC] ' . $prompt, $this->extKey, 0 );
			}
		}

		return $deliveryorderInit;
	}

	/**
	 * deliveryorderNumbers( ) : Write the delivery order number and maybe numbers for invoice and order
	 *
	 * @return	void
	 * @access private
	 * @version    2.0.0
	 * @since      2.0.0
	 */
	private function deliveryorderNumbers()
	{
		$numbers = $this->confPdf[ 'deliveryorder.' ][ 'content.' ][ 'numbers.' ];

// LOOP : fields, the elements of a product
		foreach( array_keys( ( array ) $numbers ) as $key )
		{
			if( !stristr( $key, '.' ) )
			{
				continue;
			}

			$this->writeTextblock( $numbers[ $key ], 'deliveryorder.numbers.' . $key );
		}
	}

	/**
	 * deliveryorderTermOfCredit( ) : Write the current date
	 *
	 * @return	void
	 * @access private
	 * @version    2.0.8
	 * @since      2.0.8
	 * @internal: #i0021
	 */
	private function deliveryorderTermOfCredit()
	{
		$termOfCredit = $this->confPdf[ 'deliveryorder.' ][ 'content.' ][ 'termOfCredit.' ];
		$this->writeTextblock( $termOfCredit, 'deliveryorderTermOfCredit' );
	}

	/*	 * *********************************************
	 *
	 * Init
	 *
	 * ******************************************** */

	/**
	 * init( ) : Init some global variables and some classes
	 *           Dies in case
	 *           * of an unproper dir
	 *           * of no products
	 *
	 * @param	$string		$case : deliveryorder || invoice || revocation || terms
	 * @return	void
	 * @access private
	 * @version    2.0.0
	 * @since      2.0.0
	 */
	private function init( $case )
	{
		$this->conf = $GLOBALS[ 'TSFE' ]->tmpl->setup[ 'plugin.' ][ 'tx_caddy_pi1.' ];
		$this->confPdf = $this->conf[ 'pdf.' ];

		$this->cObj = $GLOBALS[ 'TSFE' ]->cObj;
		$this->local_cObj = $GLOBALS[ 'TSFE' ]->cObj;


		$this->pi_loadLL();

		// DIE  : if there is an unproper directory
		$this->initCheckDir();

		// DIE  : if there isn't any product
		$this->initCheckProducts();

		$this->currentAttachment = $case;

		$this->initInstances();
		$this->initTemplate( $case );

		$this->caddy->setParentObject( $this );
		$this->caddy->setContentRow( $this->cObj->data );
	}

	/**
	 * initCheckDir( ) : Checks if uploads/tx_caddy is a directory.
	 *                   And dies, if not,
	 *
	 * @return	void
	 * @access private
	 * @version    2.0.0
	 * @since      2.0.0
	 */
	private function initCheckDir()
	{
// CHECK: Is directory for PDF available?
		if( !is_dir( 'uploads/tx_caddy' ) )
		{
			$prompt = 'FATAL ERROR: uploads/tx_caddy not found!<br />
        ' . __METHOD__ . ' (' . __LINE__ . ')';
			die( $prompt );
		}
	}

	/**
	 * initCheckProducts( ) :  Checks, if there is one product at least.
	 *                         And dies, if not,
	 *
	 * @return	void
	 * @access private
	 * @version    2.0.0
	 * @since      2.0.0
	 */
	private function initCheckProducts()
	{
// Get the caddy session
		$sesArray = $GLOBALS[ 'TSFE' ]->fe_user->getKey( 'ses', $this->extKey . '_' . $GLOBALS[ "TSFE" ]->id );

		if( empty( $sesArray[ 'products' ] ) )
		{
			$prompt = 'ERROR: products are empty!<br />' . PHP_EOL .
							'Please go back to the page you visited before.<br />' . PHP_EOL .
							'Sorry for the trouble.<br />' . PHP_EOL .
							'TYPO3 Caddy<br />' . PHP_EOL .
							__METHOD__ . ' (' . __LINE__ . ')';
			die( $prompt );
		}
	}

	/**
	 * initInstances( )
	 *
	 * @return	void
	 * @access private
	 * @version    8.7.4
	 * @since      2.0.0
	 */
	private function initInstances()
	{
		// #i0226, 180312, dwildt, ~
		$this->caddy = \TYPO3\CMS\Core\Utility\GeneralUtility::makeInstance( \Netzmacher\Caddy\Lib\Caddy\Caddy::class );
		$this->dynamicMarkers = \TYPO3\CMS\Core\Utility\GeneralUtility::makeInstance( \Netzmacher\Caddy\Lib\Dynamicmarkers::class );

		// #i0191, 180109, #i0184, 171201, dwildt, ~
		switch( true )
		{
			// #i0195, 180118, dwildt, ~
			case($this->zz_getTYPO3Version() <= 7006000):
				require_once( \TYPO3\CMS\Core\Utility\ExtensionManagementUtility::extPath( 't3_tcpdf' ) . 'class.tx_t3_tcpdf.php' );
				break;
			case($this->zz_getTYPO3Version() > 7006000):
			default:
				require_once( \TYPO3\CMS\Core\Utility\ExtensionManagementUtility::extPath( 't3_tcpdf' ) . 'Resources/Private/PHP/tcpdf/tcpdf.php' );
				break;
		}
	}

	/**
	 * initTemplate( )
	 *
	 * @param	$string		$case : deliveryorder || invoice || revocation || terms
	 * @return	void
	 * @access private
	 * @version    8.7.0
	 * @since      2.0.0
	 */
	private function initTemplate( $case = 'invoice' )
	{
		switch( $case )
		{
			case( 'deliveryorder' ):
			case( 'invoice' ):
			case( 'orderconfirmation' ):
				$file = $this->conf[ 'templates.' ][ 'pdf.' ][ $case . '.' ][ 'file' ];
				$markerAll = $this->conf[ 'templates.' ][ 'pdf.' ][ $case . '.' ][ 'marker.' ][ 'all' ];
				$markerItem = $this->conf[ 'templates.' ][ 'pdf.' ][ $case . '.' ][ 'marker.' ][ 'item' ];
				break;
			case( 'revocation' ):
			case( 'terms' ):
				$file = $this->conf[ 'templates.' ][ 'pdf.' ][ $case . '.' ][ 'file' ];
				$markerAll = $this->conf[ 'templates.' ][ 'pdf.' ][ $case . '.' ][ 'marker.' ][ 'all' ];
				$markerItem = $this->conf[ 'templates.' ][ 'pdf.' ][ $case . '.' ][ 'marker.' ][ 'item' ];
				break;
			default:
				$prompt = 'Unproper value in switch. Case is "' . $case . '"'
								. PHP_EOL
								. __METHOD__ . '(line ' . __LINE__ . ')';
				die( $prompt );
				break;
		}
		// #t1817, 171213, dwildt, 1-/2+
		//$tmplFile = $GLOBALS[ 'TSFE' ]->cObj->fileResource( $file );
		$absFileName = \TYPO3\CMS\Core\Utility\GeneralUtility::getFileAbsFileName( $file );
		$tmplFile = file_get_contents( $absFileName );

		// :TODO: get pid of current plugin
		$pid = $GLOBALS[ "TSFE" ]->id;
		$CleanupTemplateUtility = GeneralUtility::makeInstance( CleanupTemplateUtility::class );
		$tmplFile = $CleanupTemplateUtility->RemoveUnusedSubparts( $tmplFile, $pid );
//		var_dump(__METHOD__, __LINE__, $tmplFile);
//		die();

		$this->tmpl = array();
		$this->tmpl[ 'all' ] = $this->MarkerBasedTemplateService->getSubpart( $tmplFile, $markerAll );
		$this->tmpl[ 'item' ] = $this->MarkerBasedTemplateService->getSubpart( $this->tmpl[ 'all' ], $markerItem );

		$this->initTemplateTable( $case );
	}

	/**
	 * initTemplate( )
	 *
	 * @param	$string		$case : deliveryorder || invoice || revocation || terms
	 * @return	void
	 * @access private
	 * @internal	 #i0288
	 * @version    8.17.0
	 * @since      2.0.0
	 */
	private function initTemplateTable( $case )
	{
		switch( $case )
		{
			case( 'deliveryorder' ):
			case( 'invoice' ):
			case( 'orderconfirmation' ):
				$table = $this->conf[ 'templates.' ][ 'pdf.' ][ $case . '.' ][ 'table.' ];
				break;
			case( 'revocation' ):
			case( 'terms' ):
				$table = $this->conf[ 'templates.' ][ 'pdf.' ][ $case . '.' ][ 'table.' ];
				break;
			default:
				$prompt = 'Unproper value in switch. Case is "' . $case . '"'
								. PHP_EOL
								. __METHOD__ . '(line ' . __LINE__ . ')';
				die( $prompt );
				break;
		}

		foreach( ( array ) $table as $property => $value )
		{
			$marker = '###' . strtoupper( $property ) . '###';
			$this->tmpl[ 'all' ] = str_replace( $marker, $value, $this->tmpl[ 'all' ] );
			$this->tmpl[ 'item' ] = str_replace( $marker, $value, $this->tmpl[ 'item' ] );
		}
	}

	/*	 * *********************************************
	 *
	 * Invoice
	 *
	 * ******************************************** */

	/**
	 * invoice( ) : Generates the PDF invoice and returns the path to the PDF file
	 *
	 * @return	string		$path : path to the rendered pdf file
	 * @access public
	 * @version    2.0.0
	 * @since      2.0.0
	 */
	public function invoice()
	{
		$destFile = null;
		$destPath = null;

		// Init caddy pdf
		$this->init( 'invoice' );

		// RETURN : any pdf is requested
		if( !$this->invoiceInit() )
		{
			return;
		}
		// RETURN : any pdf is requested
		// Get the caddy session
		$sesArray = $GLOBALS[ 'TSFE' ]->fe_user->getKey( 'ses', $this->extKey . '_' . $GLOBALS[ "TSFE" ]->id );

		// Add session data to the local cObj
		$this->local_cObj->start( $sesArray, $this->pObj->conf[ 'db.' ][ 'table' ] );

		// DRS
		if( $this->pObj->drsUserfunc )
		{
			$data = var_export( $this->local_cObj->data, true );
			$prompt = 'cObj->data: ' . $data;
			\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[INFO/COBJ] ' . $prompt, $this->extKey, 0 );
		}
		// DRS
		// Get the path of the destination file
		$destFile = $this->local_cObj->cObjGetSingle
						(
						$this->confPdf[ 'invoice.' ][ 'filename' ], $this->confPdf[ 'invoice.' ][ 'filename.' ]
		);
		$destPath = PATH_site . 'uploads/tx_caddy/' . $destFile;

		if( file_exists( $destPath ) )
		{
			if( $this->pObj->drsUserfunc )
			{
				$prompt = 'uploads/tx_caddy' . '/' . $destFile . ' exists and will overridden!';
				\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[WARN/USERFUNC] ' . $prompt, $this->extKey, 2 );
			}
		}

		$srceFile = $sesArray[ 'sendVendorInvoice' ];

		if( empty( $srceFile ) )
		{
			$srceFile = $sesArray[ 'sendCustomerInvoice' ];
		}
		if( empty( $srceFile ) )
		{
			$prompt = 'Can\'t get source file from session data ' .
							'sendCustomerInvoice/sendVendorInvoice<br />' . PHP_EOL .
							__METHOD__ . '(line ' . __LINE__ . ')';
			die( $prompt );
		}
		// PDF source file
		// Init tcpdf
		$this->tcpdfInit( $srceFile );
		$this->tcpdfAddPage();

		// Write the invoice address
		$fallBackToInvoiceAddress = false;
		$this->invoiceAddress( $fallBackToInvoiceAddress );

		// Write the invoice date
		$this->invoiceDate();

		// Write the invoice number
		$this->invoiceNumbers();

		// Write the caddy
		$this->caddy( 'invoice' );

		// Write the invoice term of credit
		$this->invoiceTermOfCredit();

		// Write additional textblocks
		$this->invoiceAdditionalTextblocks();

		// Create the PDF
		$this->tcpdfOutput( $destPath );

		// RETURN :
		return $destPath;
	}

	/**
	 * invoiceAdditionalTextblocks( ) : Write addional text blocks
	 *
	 * @return	void
	 * @access private
	 * @version    2.0.0
	 * @since      2.0.0
	 */
	private function invoiceAdditionalTextblocks()
	{
		$additionalTextblocks = $this->confPdf[ 'invoice.' ][ 'content.' ][ 'additionaltextblocks.' ];

// LOOP : additional textblocks
		foreach( array_keys( ( array ) $additionalTextblocks ) as $key )
		{
			if( !stristr( $key, '.' ) )
			{
				continue;
			}

			$this->writeTextblock( $additionalTextblocks[ $key ], 'invoice.additionaltextblocks.' . $key );
		}
// LOOP : additional textblocks
	}

	/**
	 * invoiceAddress( ) : Write the invoice address
	 *
	 * @return	void
	 * @access private
	 * @version    2.0.0
	 * @since      2.0.0
	 */
	private function invoiceAddress()
	{
		$invoiceaddress = $this->confPdf[ 'invoice.' ][ 'content.' ][ 'address.' ];

// LOOP : additional textblocks
		foreach( array_keys( ( array ) $invoiceaddress ) as $key )
		{
			if( !stristr( $key, '.' ) )
			{
				continue;
			}

			$this->writeTextblock( $invoiceaddress[ $key ], 'invoice.address.' . $key );
		}
// LOOP : additional textblocks
	}

	/**
	 * invoiceDate( ) : Write the current date
	 *
	 * @return	void
	 * @access private
	 * @version    2.0.0
	 * @since      2.0.0
	 */
	private function invoiceDate()
	{
		$date = $this->confPdf[ 'invoice.' ][ 'content.' ][ 'date.' ];
		$this->writeTextblock( $date, 'invoiceDate' );
	}

	/**
	 * invoiceInit( ) : Return false, if invoice should not sent to the customer or the vendor
	 *
	 * @return	boolean		$invoiceInit  : false, if invoice should not sent
	 * @access private
	 * @version    2.0.0
	 * @since      2.0.0
	 */
	private function invoiceInit()
	{
		$invoiceInit = null;

		// Get the caddy session
		$sesArray = $GLOBALS[ 'TSFE' ]->fe_user->getKey( 'ses', $this->extKey . '_' . $GLOBALS[ "TSFE" ]->id );

		// RETURN : any pdf is requested
		switch( true )
		{
			case(!empty( $sesArray[ 'sendCustomerInvoice' ] ) ):
			case(!empty( $sesArray[ 'sendVendorInvoice' ] ) ):
				$invoiceInit = true;
				break;
			default:
				$invoiceInit = false;
				break;
		}
		unset( $sesArray );

		if( !$invoiceInit )
		{
			// DRS
			if( $this->pObj->drsUserfunc )
			{
				$prompt = __METHOD__ . ' returns null.';
				\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[INFO/USERFUNC] ' . $prompt, $this->extKey, 0 );
			}
			// DRS
		}
		// RETURN : any pdf is requested

		return $invoiceInit;
	}

	/**
	 * invoiceNumbers( ) : Write the delivery order number and maybe numbers for invoice and order
	 *
	 * @return	void
	 * @access private
	 * @version    2.0.0
	 * @since      2.0.0
	 */
	private function invoiceNumbers()
	{
		$numbers = $this->confPdf[ 'invoice.' ][ 'content.' ][ 'numbers.' ];

// LOOP : fields, the elements of a product
		foreach( array_keys( ( array ) $numbers ) as $key )
		{
			if( !stristr( $key, '.' ) )
			{
				continue;
			}

			$this->writeTextblock( $numbers[ $key ], 'invoice.numbers.' . $key );
		}
	}

	/**
	 * invoiceTermOfCredit( ) : Write the current date
	 *
	 * @return	void
	 * @access private
	 * @version    2.0.8
	 * @since      2.0.8
	 * @internal: #i0021
	 */
	private function invoiceTermOfCredit()
	{
		$termOfCredit = $this->confPdf[ 'invoice.' ][ 'content.' ][ 'termOfCredit.' ];
		$this->writeTextblock( $termOfCredit, 'invoiceTermOfCredit' );
	}

	/*	 * *********************************************
	 *
	 * Order confirmation
	 *
	 * ******************************************** */

	/**
	 * orderconfirmation( ) : Generates the PDF orderconfirmation and returns the path to the PDF file
	 *
	 * @return	string		$path : path to the rendered pdf file
	 * @access public
	 * @version    7.1.0
	 * @since      7.1.0
	 * @internal   #i0109
	 */
	public function orderconfirmation()
	{
		$destFile = null;
		$destPath = null;

		// Init caddy pdf
		$this->init( 'orderconfirmation' );

		if( !$this->orderconfirmationInit() )
		{
			return;
		}
		// RETURN : any pdf is requested
		// Get the caddy session
		$sesArray = $GLOBALS[ 'TSFE' ]->fe_user->getKey( 'ses', $this->extKey . '_' . $GLOBALS[ "TSFE" ]->id );

		// Add session data to the local cObj
		$this->local_cObj->start( $sesArray, $this->pObj->conf[ 'db.' ][ 'table' ] );

		if( $this->pObj->drsUserfunc )
		{
			$data = var_export( $this->local_cObj->data, true );
			$prompt = 'cObj->data: ' . $data;
			\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[INFO/COBJ] ' . $prompt, $this->extKey, 0 );
		}

		// Get the path of the destination file
		$destFile = $this->local_cObj->cObjGetSingle
						(
						$this->confPdf[ 'orderconfirmation.' ][ 'filename' ], $this->confPdf[ 'orderconfirmation.' ][ 'filename.' ]
		);

		$destPath = PATH_site . 'uploads/tx_caddy/' . $destFile;

		if( file_exists( $destPath ) )
		{
			if( $this->pObj->drsUserfunc )
			{
				$prompt = 'uploads/tx_caddy' . '/' . $destFile . ' exists and will overridden!';
				\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[WARN/USERFUNC] ' . $prompt, $this->extKey, 2 );
			}
		}

		$srceFile = $sesArray[ 'sendVendorOrderconfirmation' ];

		if( empty( $srceFile ) )
		{
			$srceFile = $sesArray[ 'sendCustomerOrderconfirmation' ];
		}
		if( empty( $srceFile ) )
		{
			$prompt = 'Can\'t get source file from session data ' .
							'sendCustomerOrderconfirmation/sendVendorOrderconfirmation<br />' . PHP_EOL .
							__METHOD__ . '(line ' . __LINE__ . ')';
			die( $prompt );
		}

		// Init tcpdf
		$this->tcpdfInit( $srceFile );
		$this->tcpdfAddPage();

		// Write the orderconfirmation address
		$fallBackToInvoiceAddress = false;
		$this->orderconfirmationAddress( $fallBackToInvoiceAddress );

		// Write the orderconfirmation date
		$this->orderconfirmationDate();

		// Write the orderconfirmation number
		$this->orderconfirmationNumbers();

		// Write the caddy
		$this->caddy( 'orderconfirmation' );

		// Write the orderconfirmation term of credit
		$this->orderconfirmationTermOfCredit();

		// Write additional textblocks
		$this->orderconfirmationAdditionalTextblocks();

		// Create the PDF
		$this->tcpdfOutput( $destPath );

		// RETURN :
		return $destPath;
	}

	/**
	 * orderconfirmationAdditionalTextblocks( ) : Write addional text blocks
	 *
	 * @return	void
	 * @access private
	 * @version    7.1.0
	 * @since      7.1.0
	 * @internal   #i0109
	 */
	private function orderconfirmationAdditionalTextblocks()
	{
		$additionalTextblocks = $this->confPdf[ 'orderconfirmation.' ][ 'content.' ][ 'additionaltextblocks.' ];

// LOOP : additional textblocks
		foreach( array_keys( ( array ) $additionalTextblocks ) as $key )
		{
			if( !stristr( $key, '.' ) )
			{
				continue;
			}

			$this->writeTextblock( $additionalTextblocks[ $key ], 'orderconfirmation.additionaltextblocks.' . $key );
		}
// LOOP : additional textblocks
	}

	/**
	 * orderconfirmationAddress( ) : Write the orderconfirmation address
	 *
	 * @return	void
	 * @access private
	 * @version    7.1.0
	 * @since      7.1.0
	 * @internal   #i0109
	 */
	private function orderconfirmationAddress()
	{
		$invoiceaddress = $this->confPdf[ 'invoice.' ][ 'content.' ][ 'address.' ];

// LOOP : additional textblocks
		foreach( array_keys( ( array ) $invoiceaddress ) as $key )
		{
			if( !stristr( $key, '.' ) )
			{
				continue;
			}

			$this->writeTextblock( $invoiceaddress[ $key ], 'invoice.address.' . $key );
		}
// LOOP : additional textblocks
	}

	/**
	 * orderconfirmationDate( ) : Write the current date
	 *
	 * @return	void
	 * @access private
	 * @version    7.1.0
	 * @since      7.1.0
	 * @internal   #i0109
	 */
	private function orderconfirmationDate()
	{
		$date = $this->confPdf[ 'orderconfirmation.' ][ 'content.' ][ 'date.' ];
		$this->writeTextblock( $date, 'orderconfirmationDate' );
	}

	/**
	 * orderconfirmationInit( ) : Return false, if orderconfirmation should not sent to the customer or the vendor
	 *
	 * @return	boolean		$orderconfirmationInit  : false, if orderconfirmation should not sent
	 * @access private
	 * @version    7.1.0
	 * @since      7.1.0
	 * @internal   #i0109
	 */
	private function orderconfirmationInit()
	{
		$orderconfirmationInit = null;

		// Get the caddy session
		$sesArray = $GLOBALS[ 'TSFE' ]->fe_user->getKey( 'ses', $this->extKey . '_' . $GLOBALS[ "TSFE" ]->id );

		// RETURN : any pdf is requested
		switch( true )
		{
			case(!empty( $sesArray[ 'sendCustomerOrderconfirmation' ] ) ):
			case(!empty( $sesArray[ 'sendVendorOrderconfirmation' ] ) ):
				$orderconfirmationInit = true;
				break;
			default:
				$orderconfirmationInit = false;
				break;
		}
		unset( $sesArray );

		if( !$orderconfirmationInit )
		{
			// DRS
			if( $this->pObj->drsUserfunc )
			{
				$prompt = __METHOD__ . ' returns null.';
				\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[INFO/USERFUNC] ' . $prompt, $this->extKey, 0 );
			}
			// DRS
		}
		// RETURN : any pdf is requested

		return $orderconfirmationInit;
	}

	/**
	 * orderconfirmationNumbers( ) : Write the delivery order number and maybe numbers for orderconfirmation and order
	 *
	 * @return	void
	 * @access private
	 * @version    7.1.0
	 * @since      7.1.0
	 * @internal   #i0109
	 */
	private function orderconfirmationNumbers()
	{
		$numbers = $this->confPdf[ 'orderconfirmation.' ][ 'content.' ][ 'numbers.' ];

// LOOP : fields, the elements of a product
		foreach( array_keys( ( array ) $numbers ) as $key )
		{
			if( !stristr( $key, '.' ) )
			{
				continue;
			}

			$this->writeTextblock( $numbers[ $key ], 'orderconfirmation.numbers.' . $key );
		}
	}

	/**
	 * orderconfirmationTermOfCredit( ) : Write the current date
	 *
	 * @return	void
	 * @access private
	 * @version    7.1.0
	 * @since      7.1.0
	 * @internal   #i0021, #i0109
	 */
	private function orderconfirmationTermOfCredit()
	{
		$termOfCredit = $this->confPdf[ 'orderconfirmation.' ][ 'content.' ][ 'termOfCredit.' ];
		$this->writeTextblock( $termOfCredit, 'orderconfirmationTermOfCredit' );
	}

	/*	 * *********************************************
	 *
	 * Set
	 *
	 * ******************************************** */

	/**
	 * setParentObject( )  :
	 *
	 * @param	object		$pObj: ...
	 * @return	void
	 * @access public
	 * @version    2.0.0
	 * @since      2.0.0
	 */
	public function setParentObject( $pObj )
	{
		if( !is_object( $pObj ) )
		{
			$prompt = 'ERROR: no object!<br />' . PHP_EOL .
							'Sorry for the trouble.<br />' . PHP_EOL .
							'TYPO3 Caddy<br />' . PHP_EOL .
							__METHOD__ . ' (' . __LINE__ . ')';
			die( $prompt );
		}
		$this->pObj = $pObj;
	}

	/*	 * *********************************************
	 *
	 * TCPDF
	 *
	 * ******************************************** */

	/**
	 * tcpdfPageConf( )
	 *
	 * @return	array
	 * @access private
	 * @version    6.2.0
	 * @since      6.2.0
	 */
	private function tcpdfPageConf()
	{
		$confPages = $this->confPdf[ $this->currentAttachment . '.' ][ 'page.' ];

		// Return current array
		if( is_array( $confPages[ $this->pdfPageCurrent . '.' ] ) )
		{
			$confPage = $confPages[ $this->pdfPageCurrent . '.' ];
			return $confPage;
		}

		// Return last array
		for( $pageCurrent = $this->pdfPageCurrent; $pageCurrent >= 1; $pageCurrent = $pageCurrent - 1 )
		{
			if( is_array( $confPages[ $pageCurrent . '.' ] ) )
			{
				$confPage = $confPages[ $pageCurrent . '.' ];
				return $confPage;
			}
		}

		// Return default array
		$confPage = array(
				'autopagebreak' => 40,
				'margin.' => array(
						'top' => 20,
						'right' => 10,
						'bottom' => 20,
						'left' => 20,
				)
		);
		return $confPage;
	}

	/**
	 * tcpdfInit( ) : Init a TCPDF object
	 *
	 *             http://www.setasign.de/support/manuals/fpdi/
	 *             http://www.setasign.de/products/pdf-php-solutions/fpdi/demos/
	 *
	 * @param	string		$filename : current filename
	 * @return	void
	 * @access private
	 * @internal   http://www.tcpdf.org/doc/code/classTCPDF.html
	 * @version    6.2.0
	 * @since      2.0.0
	 */
	private function tcpdfInit( $srceFile )
	{
		$path2lib = \TYPO3\CMS\Core\Utility\ExtensionManagementUtility::extPath( 't3_tcpdf' ) . 'Resources/Private/PHP/';
		require_once( $path2lib . 'tcpdf/tcpdf.php' );
		$path2lib = \TYPO3\CMS\Core\Utility\ExtensionManagementUtility::extPath( 'caddy' ) . 'Resources/Private/PHP/';
		require_once( $path2lib . 'FPDI-1.5.4/fpdi.php' );
		$this->tcpdf = \TYPO3\CMS\Core\Utility\GeneralUtility::makeInstance( 'FPDI' );
//		//$this->tcpdf = new FPDI( );
//		$path2lib = \TYPO3\CMS\Core\Utility\ExtensionManagementUtility::extPath( 't3_tcpdf' ) . 'Resources/Private/PHP/';
//		require_once( $path2lib . 'tcpdf/tcpdf.php' );
//		$this->tcpdf = \TYPO3\CMS\Core\Utility\GeneralUtility::makeInstance( 'TCPDF' );

		$author = $this->pi_getLL( 'docauthor' );
		$title = $this->pi_getLL( 'doctitle' );

		$this->tcpdf->SetAutoPageBreak( false );
		$this->tcpdf->SetAuthor( $author );
		$this->tcpdf->SetTitle( $title );
		// $this->tcpdf->SetSubject('TYPO3 Caddy Order Subject');
		// $this->tcpdf->SetKeywords('TYPO3, caddy');
		// remove default header/footer
		$this->tcpdf->setPrintHeader( false );
		$this->tcpdf->setPrintFooter( false );

		// get the page count
		$this->pdfTemplateLastPage = $this->tcpdf->setSourceFile( $srceFile );
	}

	/**
	 * tcpdfAddPage( ) :
	 *
	 * @return	void
	 * @version    6.2.0
	 * @since      6.2.0
	 */
	private function tcpdfAddPage()
	{
		$this->tcpdfAddPageWiTemplate();
		$this->tcpdfAddPageWoTemplate();
	}

	/**
	 * tcpdfAddPageWiTemplate( ) :
	 *
	 * @param	object $tcpdf: TCPDF object
	 * @return	void
	 * @version    6.2.0
	 * @since      6.2.0
	 */
	private function tcpdfAddPageWiTemplate()
	{
		if( $this->pdfTemplateLastPage < 1 )
		{
			return;
		}

		$nextTemplatePage = $this->tcpdf->getPage() + 1;
		if( $nextTemplatePage > $this->pdfTemplateLastPage )
		{
			$nextTemplatePage = $this->pdfTemplateLastPage;
		}

		// get id of the next template page
		$templateId = $this->tcpdf->importPage( $nextTemplatePage );
		// get the size of the template
		$size = $this->tcpdf->getTemplateSize( $templateId );

		// create a page (landscape or portrait depending on the imported template size)
		if( $size[ 'w' ] > $size[ 'h' ] )
		{
			$this->tcpdf->AddPage( 'L', array( $size[ 'w' ], $size[ 'h' ] ) );
		}
		else
		{
			$this->tcpdf->AddPage( 'P', array( $size[ 'w' ], $size[ 'h' ] ) );
		}

		// use the template
		// Abscissa of the upper-left corner.
		$x = 0;
		// Ordinate of the upper-left corner.
		$y = 0;
		// Width of the template in the page. If not specified or equal to zero, it is automatically calculated.
		$w = 210;
		// Height of the template in the page. If not specified or equal to zero, it is automatically calculated.
		$h = 297;
		// If this parameter is set to true the page size will be adjusted to the size of the imported page.
		$adjustPageSize = true;
		$this->tcpdf->useTemplate( $templateId, $x, $y, $w, $h, $adjustPageSize );
	}

	/**
	 * tcpdfAddPageWoTemplate( ) :
	 *
	 * @param	object $tcpdf: TCPDF object
	 * @return	void
	 * @version    6.2.0
	 * @since      6.2.0
	 */
	private function tcpdfAddPageWoTemplate()
	{
		if( $this->pdfTemplateLastPage > 0 )
		{
			return;
		}

		$this->tcpdf->AddPage();
	}

	/**
	 * tcpdfOutput( ) :  Write the PDF file to the given path.
	 *                   Dies, if file could written.
	 *
	 * @param	string		$destPath : path of the destination file
	 * @return	void
	 * @access private
	 * @version    2.0.0
	 * @since      2.0.0
	 */
	private function tcpdfOutput( $destPath )
	{

		$this->tcpdf->Output( $destPath, 'F' );

// DIE  : PDF could not written
		if( !file_exists( $destPath ) )
		{
			$prompt = $destPath . ' could not written!<br />
        ' . __METHOD__ . ' (' . __LINE__ . ')';
			die( $prompt );
		}
// DIE  : PDF could not written
// DRS
		if( file_exists( $destPath ) )
		{
			if( $this->pObj->drsUserfunc )
			{
				$prompt = $destPath . ' is written.';
				\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[INFO/USERFUNC] ' . $prompt, $this->extKey, 0 );
			}
		}
// DRS
	}

	/**
	 * tcpdfSetFont( ) : Set the font properties family, size and style
	 *
	 * @param	[type]		$$font: ...
	 * @return	void
	 * @access private
	 * @version    2.0.0
	 * @since      2.0.0
	 */
	private function tcpdfSetFont( $font )
	{
		$family = $this->zz_cObjGetSingle( $font[ 'family' ], $font[ 'family.' ] );
		$size = $this->zz_cObjGetSingle( $font[ 'size' ], $font[ 'size.' ] );
		$style = $this->zz_cObjGetSingle( $font[ 'style' ], $font[ 'style.' ] );
		$this->tcpdf->SetFont( $family, $style, $size );

// DRS
		if( $this->pObj->drsUserfunc )
		{
			$prompt = 'SetFont( "' . $family . '", "' . $style . '", ' . $size . ' )';
			\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[INFO/USERFUNC] ' . $prompt, $this->extKey, 0 );
		}
// DRS

		$this->tcpdfSetFontStretching( $font );
	}

	/**
	 * tcpdfSetFontStretching( ) : Set the font properties family, size and style
	 *
	 * @param	[type]		$$font: ...
	 * @return	void
	 * @access private
	 * @version    2.0.0
	 * @since      2.0.0
	 */
	private function tcpdfSetFontStretching( $font )
	{
		$perc = ( int ) $this->zz_cObjGetSingle( $font[ 'stretching' ], $font[ 'stretching.' ] );
		if( $perc <= 0 )
		{
			$perc = 100;
		}
		$this->tcpdf->setFontStretching( $perc );

// DRS
		if( $this->pObj->drsUserfunc )
		{
			$prompt = 'SetFontStretching( "' . $perc . ' )';
			\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[INFO/USERFUNC] ' . $prompt, $this->extKey, 0 );
		}
// DRS
	}

	/**
	 * tcpdfSetTextColor( ) :  Set the text color.
	 *                         Given color must be in a hex notation like #FFF or #1200A3
	 *                         Hex format becomes a rgb format like "255 255 255"
	 *
	 * @param	[type]		$$properties: ...
	 * @return	void
	 * @access private
	 * @version    2.0.0
	 * @since      2.0.0
	 */
	private function tcpdfSetTextColor( $properties )
	{
		$textColor = $this->zz_cObjGetSingle( $properties[ 'textColor' ], $properties[ 'textColor.' ] );
		$textColor = $this->zz_hexToRgb( $textColor );
		$colors = explode( ' ', $textColor );

		switch( true )
		{
			case( count( $colors ) == 1 ):
// grey
				$this->tcpdf->SetTextColor( $colors[ 0 ] );
				$prompt = 'SetTextColor( ' . $colors[ 0 ] . ' )';
				break;
			case( count( $colors ) == 3 ):
// rgb
				$this->tcpdf->SetTextColor( $colors[ 0 ], $colors[ 1 ], $colors[ 2 ] );
				$prompt = 'SetTextColor( ' . $colors[ 0 ] . ', ' . $colors[ 1 ] . ', ' . $colors[ 2 ] . ' )';
				break;
			case( count( $colors ) == 4 ):
// cmyk
				$this->tcpdf->SetTextColor( $colors[ 0 ], $colors[ 1 ], $colors[ 2 ], $colors[ 3 ] );
				$prompt = 'SetTextColor( ' . $colors[ 0 ] . ', ' . $colors[ 1 ] . ', ' . $colors[ 2 ] . ', ' . $colors[ 3 ] . ' )';
				break;
			default:
				$prompt = 'FATAL ERROR: textColor<br />
          ' . __METHOD__ . ' (line ' . __LINE__ . ')';
				break;
		}

// DRS
		if( $this->pObj->drsUserfunc )
		{
			\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[INFO/USERFUNC] ' . $prompt, $this->extKey, 0 );
		}
// DRS
	}

	/**
	 * tcpdfWriteHtmlCell( ) : Writes the content to the PDF. Dimensions, font and textColor are taken
	 *                         from the properties
	 *
	 *                                 hr, i, img, li, ol, p, pre, small, span, strong, sub, sup, table, tcpdf,
	 *                                 td, th, thead, tr, tt, u, ul
	 *             NOTE: all the HTML attributes must be enclosed in double-quote
	 *
	 * @param	array		$properties   : Array with the properties font, textColor and cell
	 * @param	string		$htmlContent  : Current HTML content
	 * @param	string		$drsLabel     : Label for teh DRS prompt. Usually the path of the used typoscript
	 * @return	void
	 * @access private
	 * @internal   Supported tags are: a, b, blockquote, br, dd, del, div, dl, dt, em, font, h1, h2, h3, h4, h5, h6,
	 * @version    6.2.0
	 * @since      2.0.0
	 */
	private function tcpdfWriteHtmlCell( $properties, $htmlContent, $drsLabel )
	{
		// Get cursor position
		$cursor = $this->tcpdfWriteHtmlCellCursor( $properties );
		$x = $cursor[ 'x' ];
		$y = $cursor[ 'y' ];

		// Get the current page number
		$pageNumber = $this->tcpdf->getPage();

		// Short variable
		$cell = $properties[ 'cell.' ];

		// Set some properties for writeHtmlCell()
		$this->tcpdfWriteHtmlCellProperties( $properties );
		// Set some more properties for writeHtmlCell()
		$align = $this->zz_cObjGetSingle( $cell[ 'align' ], $cell[ 'align.' ] ); // Allows to center or align the text. Possible values are: L, C, R, ''
		$autopadding = true; // if true, uses internal padding and automatically adjust it to account for line width.
		$border = 0; // Indicates if borders must be drawn around the cell. The value can be a number, a string or an array. 0: no border.
		$fill = false; // Indicates if the cell background must be painted (true) or transparent (false).
		$h = $this->zz_cObjGetSingle( $cell[ 'height' ], $cell[ 'height.' ] ); // Cell minimum height. The cell extends automatically if needed.
		$ln = 1; // 1: to the beginning of the next line [DEFAULT]
		$reseth = true; // if true reset the last cell height (default true).
		$w = $this->zz_cObjGetSingle( $cell[ 'width' ], $cell[ 'width.' ] ); // Cell width. If 0, the cell extends up to the right margin.
		// ---------- START TRANSACTION ----------
		$this->tcpdf->startTransaction();

		$confPage = $this->tcpdfPageConf();
		$this->tcpdf->SetAutoPageBreak( true, $confPage[ 'autopagebreak' ] );
		//170421
		$left = $confPage[ 'margin.' ][ 'left' ];
		$top = $confPage[ 'margin.' ][ 'top' ];
		$right = $confPage[ 'margin.' ][ 'right' ];
		$keepmargins = true;
		$this->tcpdf->SetMargins( $left, $top, $right, $keepmargins );
//$this->zz_debugToConsole( 'dev only.', __METHOD__, __LINE__ );
		// Write HTML cell
		$this->tcpdf->writeHtmlCell( $w, $h, $cursor[ 'x' ], $cursor[ 'y' ], $htmlContent, $border, $ln, $fill, $reseth, $align, $autopadding );
//$this->zz_debugToConsole( 'dev only.', __METHOD__, __LINE__ );
		$pageNumberAfterTransaction = $this->tcpdf->getPage();
		$this->tcpdf = $this->tcpdf->rollbackTransaction();
		// ---------- ROLLBACK TRANSACTION ----------
		// Add a page, if content will placed on next page
		if( $pageNumberAfterTransaction > $pageNumber )
		{
			$this->tcpdfAddPage();
		}

		// Get cursor position from before the transaction
		$x = $cursor[ 'x' ];
		$y = $cursor[ 'y' ];
		$this->tcpdf->writeHtmlCell( $w, $h, $x, $y, $htmlContent, $border, $ln, $fill, $reseth, $align, $autopadding );

		// DRS
		if( $this->pObj->drsUserfunc )
		{
			$prompt = 'writeHtmlCell( ' . $w . ', ' . $h . ', ' . $cursor[ 'x' ] . ', ' . $cursor[ 'y' ] . ', ' . $drsLabel . ' )';
			\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[INFO/USERFUNC] ' . $prompt, $this->extKey, 0 );
		}
		// DRS
	}

	/**
	 * tcpdfWriteHtmlCellCursor( ) :
	 *
	 * @param	array		$properties   : Array with the properties font, textColor and cell
	 * @return	array $cursor: x, y
	 * @access private
	 * @version    6.2.0
	 * @since      6.2.0
	 */
	private function tcpdfWriteHtmlCellCursor( $properties )
	{
		$cursor = array();

		$cell = $properties[ 'cell.' ];
		$x = ( float ) $this->zz_cObjGetSingle( $cell[ 'x' ], $cell[ 'x.' ] );
		if( empty( $x ) )
		{
			$x = $this->tcpdf->GetX();
		}
		$y = ( float ) $this->zz_cObjGetSingle( $cell[ 'y' ], $cell[ 'y.' ] );
		if( empty( $y ) )
		{
			$y = $this->tcpdf->GetY();
		}

		$cursor = array(
				'x' => $x,
				'y' => $y
		);

		return $cursor;
	}

	/**
	 * tcpdfWriteHtmlCellProperties( ) :
	 *
	 * @param	array		$properties   : Array with the properties font, textColor and cell
	 * @return	void
	 * @access private
	 * @version    6.2.0
	 * @since      2.0.0
	 */
	private function tcpdfWriteHtmlCellProperties( $properties )
	{
		// Set textColor
		$this->tcpdfSetTextColor( $properties );

		// Set font
		$this->tcpdfSetFont( $properties[ 'font.' ] );
	}

	/*	 * *********************************************
	 *
	 * Revocation
	 *
	 * ******************************************** */

	/**
	 * revocation( ) : Generates the PDF revocation and returns the path to the PDF file
	 *
	 * @return	string		$path : path to the rendered pdf file
	 * @access public
	 * @version    2.0.0
	 * @since      2.0.0
	 */
	public function revocation()
	{
		$destFile = null;
		$destPath = null;

// Init caddy pdf
		$this->init( 'revocation' );

// RETURN : any pdf is requested
		if( !$this->revocationInit() )
		{
			return;
		}
// RETURN : any pdf is requested
// Get the caddy session
		$sesArray = $GLOBALS[ 'TSFE' ]->fe_user->getKey( 'ses', $this->extKey . '_' . $GLOBALS[ "TSFE" ]->id );

// Add session data to the local cObj
		$this->local_cObj->start( $sesArray, $this->pObj->conf[ 'db.' ][ 'table' ] );

// DRS
		if( $this->pObj->drsUserfunc )
		{
			$data = var_export( $this->local_cObj->data, true );
			$prompt = 'cObj->data: ' . $data;
			\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[INFO/COBJ] ' . $prompt, $this->extKey, 0 );
		}
// DRS
// Get the path of the destination file
		$destFile = $this->local_cObj->cObjGetSingle
						(
						$this->confPdf[ 'revocation.' ][ 'filename' ], $this->confPdf[ 'revocation.' ][ 'filename.' ]
		);
// #i0079, 150416, dwildt, 1+/-
//$destPath = 'uploads/tx_caddy/' . $destFile;
		$destPath = PATH_site . 'uploads/tx_caddy/' . $destFile;
// Get the path of the destination file
// RETURN : destination file already exists
// #i0079, 150416, dwildt, 1+/-
//if( file_exists( 'uploads/tx_caddy' . '/' . $destFile ) )
		if( file_exists( $destPath ) )
		{
// DRS
			if( $this->pObj->drsUserfunc )
			{
// #i0015, 130611, dwildt, 1-
//$prompt = 'RETURN : uploads/tx_caddy' . '/' . $destFile . ' exists!';
// #i0015, 130611, dwildt, 1+
				$prompt = 'uploads/tx_caddy' . '/' . $destFile . ' exists and will overridden!';
				\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[WARN/USERFUNC] ' . $prompt, $this->extKey, 2 );
			}
// DRS
// #i0015, 130611, dwildt, 1-
//return $destPath;
		}
// RETURN : destination file already exists
// PDF source file
		$srceFile = $sesArray[ 'sendVendorRevocation' ];
		if( empty( $srceFile ) )
		{
			$srceFile = $sesArray[ 'sendCustomerRevocation' ];
		}
		if( empty( $srceFile ) )
		{
			$prompt = 'Can\'t get source file from session data ' .
							'sendCustomerRevocation/sendVendorRevocation<br />' . PHP_EOL .
							__METHOD__ . '(line ' . __LINE__ . ')';
			die( $prompt );
		}
// PDF source file
// Init tcpdf
		$this->tcpdfInit( $srceFile );
		$this->tcpdfAddPage();

// Write the delivery order address
		$this->revocationAddress();

// Write the delivery order date
		$this->revocationDate();

// Write additional textblocks
		$this->revocationAdditionalTextblocks();

// Write the caddy
		$this->caddy( 'invoice' );
// Create the PDF
		$this->tcpdfOutput( $destPath );

// RETURN :
		return $destPath;
	}

	/**
	 * revocationAdditionalTextblocks( ) : Write addional text blocks
	 *
	 * @return	void
	 * @access private
	 * @version    2.0.0
	 * @since      2.0.0
	 */
	private function revocationAdditionalTextblocks()
	{
		$additionalTextblocks = $this->confPdf[ 'revocation.' ][ 'content.' ][ 'additionaltextblocks.' ];

// LOOP : additional textblocks
		foreach( array_keys( ( array ) $additionalTextblocks ) as $key )
		{
			if( !stristr( $key, '.' ) )
			{
				continue;
			}

			$this->writeTextblock( $additionalTextblocks[ $key ], 'revocation.additionaltextblocks.' . $key );
		}
// LOOP : additional textblocks
	}

	/**
	 * revocationAddress( ) :  Write the invoice address
	 *
	 * @return	void
	 * @access private
	 * @version    2.0.0
	 * @since      2.0.0
	 */
	private function revocationAddress()
	{
		$invoiceaddress = $this->confPdf[ 'revocation.' ][ 'content.' ][ 'address.' ][ 'invoice.' ];
		$this->writeTextblock( $invoiceaddress, 'invoiceAddress' );
	}

	/**
	 * revocationDate( ) : Write the current date
	 *
	 * @return	void
	 * @access private
	 * @version    2.0.0
	 * @since      2.0.0
	 */
	private function revocationDate()
	{
		$date = $this->confPdf[ 'revocation.' ][ 'content.' ][ 'date.' ];
		$this->writeTextblock( $date, 'revocationDate' );
	}

	/**
	 * revocationInit( ) : Return false, if revocation should not sent to the customer or the vendor
	 *
	 * @return	boolean		$revocationInit  : false, if invoice should not sent
	 * @access private
	 * @version    2.0.0
	 * @since      2.0.0
	 */
	private function revocationInit()
	{
		$revocationInit = null;

// Get the caddy session
		$sesArray = $GLOBALS[ 'TSFE' ]->fe_user->getKey( 'ses', $this->extKey . '_' . $GLOBALS[ "TSFE" ]->id );

// RETURN : any pdf is requested
		switch( true )
		{
			case(!empty( $sesArray[ 'sendCustomerRevocation' ] ) ):
			case(!empty( $sesArray[ 'sendVendorRevocation' ] ) ):
				$revocationInit = true;
				break;
			default:
				$revocationInit = false;
				break;
		}
		unset( $sesArray );

		if( !$revocationInit )
		{
// DRS
			if( $this->pObj->drsUserfunc )
			{
				$prompt = __METHOD__ . ' returns null.';
				\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[INFO/USERFUNC] ' . $prompt, $this->extKey, 0 );
			}
// DRS
		}
// RETURN : any pdf is requested

		return $revocationInit;
	}

	/*	 * *********************************************
	 *
	 * Terms
	 *
	 * ******************************************** */

	/**
	 * terms( ) : Generates the PDF terms and returns the path to the PDF file
	 *
	 * @return	string		$path : path to the rendered pdf file
	 * @access public
	 * @version    2.0.0
	 * @since      2.0.0
	 */
	public function terms()
	{
		$destFile = null;
		$destPath = null;

// Init caddy pdf
		$this->init( 'terms' );

// RETURN : any pdf is requested
		if( !$this->termsInit() )
		{
			return;
		}
// RETURN : any pdf is requested
// Get the caddy session
		$sesArray = $GLOBALS[ 'TSFE' ]->fe_user->getKey( 'ses', $this->extKey . '_' . $GLOBALS[ "TSFE" ]->id );

// Add session data to the local cObj
		$this->local_cObj->start( $sesArray, $this->pObj->conf[ 'db.' ][ 'table' ] );

// DRS
		if( $this->pObj->drsUserfunc )
		{
			$data = var_export( $this->local_cObj->data, true );
			$prompt = 'cObj->data: ' . $data;
			\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[INFO/COBJ] ' . $prompt, $this->extKey, 0 );
		}
// DRS
// Get the path of the destination file
		$destFile = $this->local_cObj->cObjGetSingle
						(
						$this->confPdf[ 'terms.' ][ 'filename' ], $this->confPdf[ 'terms.' ][ 'filename.' ]
		);
// #i0079, 150416, dwildt, 1+/-
//$destPath = 'uploads/tx_caddy/' . $destFile;
		$destPath = PATH_site . 'uploads/tx_caddy/' . $destFile;
// Get the path of the destination file
// RETURN : destination file already exists
// #i0079, 150416, dwildt, 1+/-
//if( file_exists( 'uploads/tx_caddy' . '/' . $destFile ) )
		if( file_exists( $destPath ) )
		{
// DRS
			if( $this->pObj->drsUserfunc )
			{
// #i0015, 130611, dwildt, 1-
//$prompt = 'RETURN : uploads/tx_caddy' . '/' . $destFile . ' exists!';
// #i0015, 130611, dwildt, 1+
				$prompt = 'uploads/tx_caddy' . '/' . $destFile . ' exists and will overridden!';
				\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[WARN/USERFUNC] ' . $prompt, $this->extKey, 2 );
			}
// DRS
// #i0015, 130611, dwildt, 1-
//return $destPath;
		}
// RETURN : destination file already exists
// PDF source file
		$srceFile = $sesArray[ 'sendVendorTerms' ];
		if( empty( $srceFile ) )
		{
			$srceFile = $sesArray[ 'sendCustomerTerms' ];
		}
		if( empty( $srceFile ) )
		{
			$prompt = 'Can\'t get source file from session data ' .
							'sendCustomerTerms/sendVendorTerms<br />' . PHP_EOL .
							__METHOD__ . '(line ' . __LINE__ . ')';
			die( $prompt );
		}
// PDF source file
// Init tcpdf
		$this->tcpdfInit( $srceFile );
		$this->tcpdfAddPage();

// Write the delivery order address
		$this->termsAddress();

// Write the delivery order date
		$this->termsDate();

// Write additional textblocks
		$this->termsAdditionalTextblocks();

// Write the caddy
		$this->caddy( 'invoice' );
// Create the PDF
		$this->tcpdfOutput( $destPath );

// RETURN :
		return $destPath;
	}

	/**
	 * termsAdditionalTextblocks( ) : Write addional text blocks
	 *
	 * @return	void
	 * @access private
	 * @version    2.0.0
	 * @since      2.0.0
	 */
	private function termsAdditionalTextblocks()
	{
		$additionalTextblocks = $this->confPdf[ 'terms.' ][ 'content.' ][ 'additionaltextblocks.' ];

// LOOP : additional textblocks
		foreach( array_keys( ( array ) $additionalTextblocks ) as $key )
		{
			if( !stristr( $key, '.' ) )
			{
				continue;
			}

			$this->writeTextblock( $additionalTextblocks[ $key ], 'terms.additionaltextblocks.' . $key );
		}
// LOOP : additional textblocks
	}

	/**
	 * termsAddress( ) :  Write the invoice address
	 *
	 * @return	void
	 * @access private
	 * @version    2.0.0
	 * @since      2.0.0
	 */
	private function termsAddress()
	{
		$invoiceaddress = $this->confPdf[ 'terms.' ][ 'content.' ][ 'address.' ][ 'invoice.' ];
		$this->writeTextblock( $invoiceaddress, 'invoiceAddress' );
	}

	/**
	 * termsDate( ) : Write the current date
	 *
	 * @return	void
	 * @access private
	 * @version    2.0.0
	 * @since      2.0.0
	 */
	private function termsDate()
	{
		$date = $this->confPdf[ 'terms.' ][ 'content.' ][ 'date.' ];
		$this->writeTextblock( $date, 'termsDate' );
	}

	/**
	 * termsInit( ) : Return false, if terms should not sent to the customer or the vendor
	 *
	 * @return	boolean		$termsInit  : false, if invoice should not sent
	 * @access private
	 * @version    2.0.0
	 * @since      2.0.0
	 */
	private function termsInit()
	{
		$termsInit = null;

// Get the caddy session
		$sesArray = $GLOBALS[ 'TSFE' ]->fe_user->getKey( 'ses', $this->extKey . '_' . $GLOBALS[ "TSFE" ]->id );

// RETURN : any pdf is requested
		switch( true )
		{
			case(!empty( $sesArray[ 'sendCustomerTerms' ] ) ):
			case(!empty( $sesArray[ 'sendVendorTerms' ] ) ):
				$termsInit = true;
				break;
			default:
				$termsInit = false;
				break;
		}
		unset( $sesArray );

		if( !$termsInit )
		{
// DRS
			if( $this->pObj->drsUserfunc )
			{
				$prompt = __METHOD__ . ' returns null.';
				\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[INFO/USERFUNC] ' . $prompt, $this->extKey, 0 );
			}
// DRS
		}
// RETURN : any pdf is requested

		return $termsInit;
	}

	/*	 * *********************************************
	 *
	 * Write
	 *
	 * ******************************************** */

	/**
	 * writeTextblock( ) : Write a texblock. If textblock has a header, prepend it to the body / content
	 *
	 * @param	array		$textBlock  : the additional text block
	 * @param	string		$drsLabel   : label for the drs prompt. Usually name of the calling function.
	 * @return	void
	 * @access private
	 * @version    2.0.0
	 * @since      2.0.0
	 */
	private function writeTextblock( $textBlock, $drsLabel )
	{
		$body = $textBlock[ 'body.' ];
		$htmlContent = $GLOBALS[ 'TSFE' ]->cObj->cObjGetSingle( $body[ 'content' ], $body[ 'content.' ] );

		$content = strip_tags( $htmlContent );
		$content = trim( $content );
		$content = str_replace( ' ', null, $content );

		// RETURN : HTML content is empty
		if( empty( $content ) )
		{
			return;
		}
		// RETURN : HTML content is empty

		$header = $textBlock[ 'header.' ];
		$htmlContent = $this->writeTextblockAddHeader( $header ) . $htmlContent;
		$this->tcpdfWriteHtmlCell( $body[ 'properties.' ], $htmlContent, $drsLabel );
	}

	/**
	 * writeTextblockAddHeader( ) : Returns a HTML rendered header
	 *
	 * @param	[type]		$$header: ...
	 * @return	void
	 * @access private
	 * @version    2.0.0
	 * @since      2.0.0
	 */
	private function writeTextblockAddHeader( $header )
	{
// RETURN : there is no header
		if( empty( $header ) )
		{
			return;
		}
// RETURN : there is no header
// Get the header content
		$htmlContent = $GLOBALS[ 'TSFE' ]->cObj->cObjGetSingle( $header[ 'content' ], $header[ 'content.' ] );

		return $htmlContent;
	}

	/*	 * *********************************************
	 *
	 * ZZ
	 *
	 * ******************************************** */

	/**
	 * zz_hexToRgb( )  : Returns the rgb notation for the given value - a CSS color name or
	 *                   a color in a hex notation.
	 *
	 * @param	string		$hex        : color in HTML notation like #FFF or #CC00CC or CSS color name
	 * @return	string		$textColor  : color in RGB format like "255 255 255" or "144 0 144"
	 * @access private
	 * @version    2.0.0
	 * @since      2.0.0
	 */
	private function zz_hexToRgb( $hex )
	{
		switch( true )
		{
			case( $hex == 'aqua' ):
				$hex = '#00FFFF';
				break;
			case( $hex == 'beige' ):
				$hex = '#F5F5DC';
				break;
			case( $hex == 'black' ):
				$hex = '#000000';
				break;
			case( $hex == 'blue' ):
				$hex = '#0000FF';
				break;
			case( $hex == 'brown' ):
				$hex = '#A52A2A';
				break;
			case( $hex == 'fuchsia' ):
				$hex = '#FF00FF';
				break;
			case( $hex == 'gold' ):
				$hex = '#FFD700';
				break;
			case( $hex == 'gray' ):
				$hex = '#808080';
				break;
			case( $hex == 'green' ):
				$hex = '#008000';
				break;
			case( $hex == 'lime' ):
				$hex = '#00FF00';
				break;
			case( $hex == 'maroon' ):
				$hex = '#800000';
				break;
			case( $hex == 'navy' ):
				$hex = '#000080';
				break;
			case( $hex == 'olive' ):
				$hex = '#808000';
				break;
			case( $hex == 'orange' ):
				$hex = '#FFA500';
				break;
			case( $hex == 'purple' ):
				$hex = '#800080';
				break;
			case( $hex == 'red' ):
				$hex = '#FF0000';
				break;
			case( $hex == 'silver' ):
				$hex = '#C0C0C0';
				break;
			case( $hex == 'tan' ):
				$hex = '#D2B48C';
				break;
			case( $hex == 'teal' ):
				$hex = '#008080';
				break;
			case( $hex == 'white' ):
				$hex = '#FFFFFF';
				break;
			case( $hex == 'yellow' ):
				$hex = '#FFFF00';
				break;
		}

		$hex = str_replace( '#', null, $hex );

		if( strlen( $hex ) == 3 )
		{
			$r = hexdec( substr( $hex, 0, 1 ) . substr( $hex, 0, 1 ) );
			$g = hexdec( substr( $hex, 1, 1 ) . substr( $hex, 1, 1 ) );
			$b = hexdec( substr( $hex, 2, 1 ) . substr( $hex, 2, 1 ) );
		}
		else
		{
			$r = hexdec( substr( $hex, 0, 2 ) );
			$g = hexdec( substr( $hex, 2, 2 ) );
			$b = hexdec( substr( $hex, 4, 2 ) );
		}
		$rgb = array( $r, $g, $b );

		$textColor = implode( ' ', $rgb );
		return $textColor;
	}

	/**
	 * zz_cObjGetSingle( ) : Renders a typoscript property with cObjGetSingle, if it is an array.
	 *                       Otherwise returns the property unchanged.
	 *
	 * @param	string		$cObj_name  : value or the name of the property like TEXT, COA, IMAGE, ...
	 * @param	array		$cObj_conf  : null or the configuration of the property
	 * @return	string		$value      : unchanged value or rendered typoscript property
	 * @access private
	 * @version    2.0.0
	 * @since      2.0.0
	 */
	private function zz_cObjGetSingle( $cObj_name, $cObj_conf )
	{
		switch( true )
		{
			case( is_array( $cObj_conf ) ):
				$value = $GLOBALS[ 'TSFE' ]->cObj->cObjGetSingle( $cObj_name, $cObj_conf );
				break;
			case(!( is_array( $cObj_conf ) ) ):
			default:
				$value = $cObj_name;
				break;
		}

		return $value;
	}

	/**
	 * zz_debugToConsole( ): 
	 *
	 * @param	mixed		$data: 
	 * @param	string		$method: 
	 * @param	string		$line: 
	 * @return	string		$path : absolute path or false in case of an error
	 * @internal    #i0163
	 * @version     8.9.6
	 * @since       8.9.6
	 */
	private function zz_debugToConsole( $data, $method, $line )
	{
		$promptOutput = $this->conf[ 'debug.' ][ 'console' ];
		$promptOutput = true; // 181012: test
		\Netzmacher\Caddy\Utility\Debug\Console::log( $data, $method, $line, $promptOutput );
		return;
	}

	/**
	 * zz_getTYPO3Version( ): Get the current TYPO3 version, move it to an integer
	 *                      and set the global $bool_typo3_43
	 *                      This method is independent from
	 *                        * \TYPO3\CMS\Core\Utility\GeneralUtility::int_from_ver (upto 4.7)
	 *                        * t3lib_utility_VersionNumber::convertVersionNumberToInteger (from 4.7)
	 *
	 * @return	void
	 * @version 8.7.4
	 * @since   8.7.4
	 * @internal #i0191
	 * @access private
	 */
	private function zz_getTYPO3Version()
	{
		// Set TYPO3 version as integer (sample: 4.7.7 -> 4007007)
		list( $main, $sub, $bugfix ) = explode( '.', TYPO3_version );
		$version = ( ( int ) $main ) * 1000000;
		$version = $version + ( ( int ) $sub ) * 1000;
		$version = $version + ( ( int ) $bugfix ) * 1;

		return $version;
	}

}
