<?php

namespace Netzmacher\Caddy\Lib;

/* * *************************************************************
 *  Copyright notice
 *
 *  (c) 2013-2020 - Dirk Wildt <http://wildt.at.die-netzmacher.de>
 *  All rights reserved
 *
 *  This script is part of the TYPO3 project. The TYPO3 project is
 *  free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  The GNU General Public License can be found at
 *  http://www.gnu.org/copyleft/gpl.html.
 *
 *  This script is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  This copyright notice MUST APPEAR in all copies of the script!
 * ************************************************************* */

/**
 * Plugin for the 'caddy' extension.
 *
 * @author	Dirk Wildt <http://wildt.at.die-netzmacher.de>
 * @package    TYPO3
 * @subpackage    tx_caddy
 * @version     9.1.4
 * @since       1.4.6
 */
class Session
{

	public $extKey = 'caddy';
	public $prefixId = 'tx_caddy_pi1';
	public $scriptRelPath = 'Resources/Private/Language/Plugins/pi1/class.tx_caddy_pi1.php';
	public $conf = null; // Array  : the TypoScript configuration of the parent object
	public $cObj = null; // Object : the content object of the parent object
	public $drs = null; // Object : the DRS object of the parent object
	public $pObj = null; // Object : the parent object
	private $sql = null; // Object : the SQL object

	/*	 * *********************************************
	 *
	 * Getting methods - user methods
	 *
	 * ******************************************** */

	/**
	 * getCObj( ) : Get the content object
	 *
	 * @return	object    $cObj : content object
	 * @access public
	 * @internal   #55726
	 * @version    4.0.7
	 * @since      4.0.7
	 */
	public function getCObj()
	{
		return $this->cObj;
	}

	/**
	 * getConf( ) : Get the TypoScript configuration array
	 *
	 * @return	array   $conf : TypoScript configuration
	 * @access public
	 * @internal   #55726
	 * @version    4.0.7
	 * @since      4.0.7
	 */
	public function getConf()
	{
		return $this->conf;
	}

	/**
	 * getDrs( ) : Get the DRS object
	 *
	 * @return	object   $drs : DRS object
	 * @access public
	 * @internal   #55726
	 * @version    4.0.7
	 * @since      4.0.7
	 */
	public function getDrs()
	{
		return $this->drs;
	}

	/**
	 * getNumberDeliveryorder( )  : Get the current order number
	 *
	 * @param	integer		: $pid  : uid of the page, which contains the caddy plugin
	 * @param	[type]		$conf: ...
	 * @return	string
	 * @access public
	 * @internal    #54634
	 * @version     4.0.3
	 * @since       2.0.0
	 */
	public function getNumberDeliveryorder( $content = '', $conf = array() )
	{
		unset( $content );
		$this->conf = $conf;

		// DRS
		if( $this->conf[ 'userFunc.' ][ 'drs' ] )
		{
			$this->drsUserfunc = true;
			$prompt = 'DRS is enabled by userfunc ' . __METHOD__ . '[userFunc.][drs].';
			\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[INFO/USERFUNC] ' . $prompt, $this->extKey, 0 );
		}
		// DRS

		$pid = $this->getPid();

		// #54634, 131128, dwildt, 1-
		//$sesArray       = $GLOBALS['TSFE']->fe_user->getKey( 'ses', $this->extKey . '_' . $GLOBALS["TSFE"]->id );
		// #54634, 131128, dwildt, 1+
		$sesArray = $GLOBALS[ 'TSFE' ]->fe_user->getKey( 'ses', $this->extKey . '_' . $pid );
		$currentNumber = $sesArray[ 'numberDeliveryorderCurrent' ];
		return $currentNumber;
	}

	/**
	 * getNumberInvoice( )  : Get the current order number
	 *
	 * @param	integer		: $pid  : uid of the page, which contains the caddy plugin
	 * @param	[type]		$conf: ...
	 * @return	string
	 * @access public
	 * @internal    #54634
	 * @version     4.0.3
	 * @since       2.0.0
	 */
	public function getNumberInvoice( $content = '', $conf = array() )
	{
		unset( $content );
		$this->conf = $conf;

		// DRS
		if( $this->conf[ 'userFunc.' ][ 'drs' ] )
		{
			$this->drsUserfunc = true;
			$prompt = 'DRS is enabled by userfunc ' . __METHOD__ . '[userFunc.][drs].';
			\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[INFO/USERFUNC] ' . $prompt, $this->extKey, 0 );
		}
		// DRS

		$pid = $this->getPid();

		// #54634, 131128, dwildt, 1-
		//$sesArray       = $GLOBALS['TSFE']->fe_user->getKey( 'ses', $this->extKey . '_' . $GLOBALS["TSFE"]->id );
		// #54634, 131128, dwildt, 1+
		$sesArray = $GLOBALS[ 'TSFE' ]->fe_user->getKey( 'ses', $this->extKey . '_' . $pid );
		$currentNumber = $sesArray[ 'numberInvoiceCurrent' ];
		return $currentNumber;
	}

	/**
	 * getNumberOrder( )  : Get the current order number
	 *
	 * @param	integer		: $pid  : uid of the page, which contains the caddy plugin
	 * @param	[type]		$conf: ...
	 * @return	string
	 * @access public
	 * @internal    #54634
	 * @version     4.0.3
	 * @since       2.0.0
	 */
	public function getNumberOrder( $content = '', $conf = array() )
	{
		unset( $content );
		$this->conf = $conf;

		// DRS
		if( $this->conf[ 'userFunc.' ][ 'drs' ] )
		{
			$this->drsUserfunc = true;
			$prompt = 'DRS is enabled by userfunc ' . __METHOD__ . '[userFunc.][drs].';
			\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[INFO/USERFUNC] ' . $prompt, $this->extKey, 0 );
		}
		// DRS

		$pid = $this->getPid();

		// #54634, 131128, dwildt, 1-
		//$sesArray       = $GLOBALS['TSFE']->fe_user->getKey( 'ses', $this->extKey . '_' . $GLOBALS["TSFE"]->id );
		// #54634, 131128, dwildt, 1+
		$sesArray = $GLOBALS[ 'TSFE' ]->fe_user->getKey( 'ses', $this->extKey . '_' . $pid );
		$currentNumber = $sesArray[ 'numberOrderCurrent' ];
//    if( $this->drs->drsInfo )
//    {
//      $prompt = '$GLOBALS["TSFE"]->id = ' . $GLOBALS["TSFE"]->id;
//      \TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[ERROR/SESSION] ' . $prompt, $this->extKey, 3 );
//      $prompt = 'pid = ' . $pid;
//      \TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[ERROR/SESSION] ' . $prompt, $this->extKey, 3 );
//      $prompt = 'ordernumber = ' . $currentNumber;
//      \TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[ERROR/SESSION] ' . $prompt, $this->extKey, 3 );
//      exit;
//    }
		return $currentNumber;
	}

	/**
	 * getNumberOrder( )  : Get the current order number
	 *
	 * @param	integer		: $pid  : uid of the page, which contains the caddy plugin
	 * @param	[type]		$conf: ...
	 * @return	string
	 * @access public
	 * @internal    #i0078
	 * @version     7.2.4
	 * @since       7.2.4
	 */
	public function getPaymentId( $content = '', $conf = array() )
	{
		unset( $content );
		$this->conf = $conf;

		// DRS
		if( $this->conf[ 'userFunc.' ][ 'drs' ] )
		{
			$this->drsUserfunc = true;
			$prompt = 'DRS is enabled by userfunc ' . __METHOD__ . '[userFunc.][drs].';
			\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[INFO/USERFUNC] ' . $prompt, $this->extKey, 0 );
		}
		// DRS

		$pid = $this->getPid();

		// #54634, 131128, dwildt, 1-
		//$sesArray       = $GLOBALS['TSFE']->fe_user->getKey( 'ses', $this->extKey . '_' . $GLOBALS["TSFE"]->id );
		// #54634, 131128, dwildt, 1+
		$sesArray = $GLOBALS[ 'TSFE' ]->fe_user->getKey( 'ses', $this->extKey . '_' . $pid );
		$paymentId = $sesArray[ 'payment' ];
//    if( $this->drs->drsInfo )
//    {
//      $prompt = '$GLOBALS["TSFE"]->id = ' . $GLOBALS["TSFE"]->id;
//      \TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[ERROR/SESSION] ' . $prompt, $this->extKey, 3 );
//      $prompt = 'pid = ' . $pid;
//      \TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[ERROR/SESSION] ' . $prompt, $this->extKey, 3 );
//      $prompt = 'ordernumber = ' . $paymentId;
//      \TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[ERROR/SESSION] ' . $prompt, $this->extKey, 3 );
//      exit;
//    }
		return $paymentId;
	}

	/**
	 * getPid( )  : Returns the globlas tsfe id, if the given pid is null
	 *
	 * @param	integer		$pid  : given pid (may be null)
	 * @return	integer		$pid  : id of the page with the caddy plugin
	 * @internal    #54634
	 * @access      private
	 * @version     9.1.4
	 * @since       3.0.1
	 */
	private function getPid( $pid = null )
	{
		$pid = \Netzmacher\Caddy\Userfunc\GetPid::Caddy( $this->conf );
		if( $this->drs->drsError || $this->drsUserfunc )
		{
			$prompt = 'Given pid of the Caddy page is empty.';
			\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[INFO/SESSION] ' . $prompt, $this->extKey, 0 );
			$prompt = 'pid is set to the pid of the current page: #' . $pid;
			\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[HELP/SESSION] ' . $prompt, $this->extKey, 1 );
		}

		return $pid;
	}

	/*	 * *********************************************
	 *
	 * Init
	 *
	 * ******************************************** */

	/**
	 * init( )
	 *
	 * @return	void
	 * @access private
	 * @internal   #55726
	 * @version    4.0.7
	 * @since      4.0.7
	 */
	private function init()
	{
		$this->initVars();
		$this->initInstances();
	}

	/**
	 * initInstances( )
	 *
	 * @return		void
	 * @access		private
	 * @internal  #55726
	 * @version   8.13.0
	 * @since     4.0.7
	 */
	private function initInstances()
	{
		if( !( $this->initInstances === null ) )
		{
			return;
		}

		// #i0262, 180913, dwildt, 1+
		$this->initInstancesSql();
		$this->initInstancesStockmanager();

		$this->initInstances = true;
	}

	/**
	 * initInstancesSql( )
	 *
	 * @return		void
	 * @access		private
	 * @version   8.13.0
	 * @since     8.13.0
	 */
	private function initInstancesSql()
	{
		$this->sql = \TYPO3\CMS\Core\Utility\GeneralUtility::makeInstance( \Netzmacher\Caddy\Lib\Sql::class );
		$this->sql->SetParentObject( $this->pObj );
	}

	/**
	 * initInstancesStockmanager( )
	 *
	 * @return	void
	 * @access private
	 * @internal   #55726
	 * @version    4.0.7
	 * @since      4.0.7
	 */
	private function initInstancesStockmanager()
	{
		// #i0226, 180312, dwildt, ~
		$this->stockmanager = \TYPO3\CMS\Core\Utility\GeneralUtility::makeInstance( \Netzmacher\Caddy\Lib\Caddy\Stockmanager::class );
		$this->stockmanager->setParentObject( $this );
	}

	/**
	 * initVars( )
	 *
	 * @return	void
	 * @access private
	 * @internal   #55726
	 * @version    4.0.7
	 * @since      4.0.7
	 */
	private function initVars()
	{
//var_dump(__METHOD__, __LINE__);		
		$this->conf = $this->pObj->conf;
//var_dump(__METHOD__, __LINE__, $this->conf);		
		$this->cObj = $this->pObj->cObj;
		$this->drs = $this->pObj->drs;
	}

	/*	 * *********************************************
	 *
	 * get
	 *
	 * ******************************************** */

	/**
	 * getVariantsStyle( )  :
	 *
	 * @return	array		$variants :
	 * @access public
	 * @internal     #t0915
	 * @version     7.2.0
	 * @since       7.2.0
	 */
	public function getVariantsStyle()
	{
		static $getVariantsStyle = NULL;

		if( $getVariantsStyle !== NULL )
		{
			return $getVariantsStyle;
		}
		// RETURN 'wt_cart': variants like wt_cart style are configured
		if( is_array( $this->pObj->conf[ 'settings.' ][ 'variant.' ] ) )
		{
			$getVariantsStyle = $this->getVariantsStyleDRSwtcart();
			return $getVariantsStyle;
		}

		// RETURN 'caddy': variants like caddy style are configured
		if( !empty( $this->pObj->conf[ 'constant_editor.' ][ 'variants.' ][ 'tables.' ][ 'variant' ] ) )
		{
			$getVariantsStyle = $this->getVariantsStyleDRScaddy();
			return $getVariantsStyle;
		}

		$getVariantsStyle = $this->getVariantsStyleDRSnovariants();
		return $getVariantsStyle;
	}

	/**
	 * getVariantsStyleDRScaddy( )  :
	 *
	 * @return	string  style:
	 * @access public
	 * @internal     #t0915
	 * @version     7.2.0
	 * @since       7.2.0
	 */
	public function getVariantsStyleDRScaddy()
	{
		$getVariantsStyle = 'caddy';

		if( !$this->drs->drsVariants )
		{
			return $getVariantsStyle;
		}

		$prompt = 'constant_editor.variants.tables.variant is '
						. $this->pObj->conf[ 'constant_editor.' ][ 'variants.' ][ 'tables.' ][ 'variant' ] . '. '
						. 'We have the caddy style for variants.';
		\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[INFO/VARIANTS] ' . $prompt, $this->extKey, 0 );
		$getVariantsStyle = 'caddy';
		return $getVariantsStyle;
	}

	/**
	 * getVariantsStyleDRSnovariants( )  :
	 *
	 * @return	string  style:
	 * @access public
	 * @internal     #t0915
	 * @version     7.2.0
	 * @since       7.2.0
	 */
	public function getVariantsStyleDRSnovariants()
	{
		$getVariantsStyle = 'novariants';

		if( !$this->drs->drsVariants )
		{
			return $getVariantsStyle;
		}

		$prompt = 'settings.variant hasn\'t any element and db.tablevariants is empty: there isn\'t any configuration for variants.';
		\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[INFO/VARIANTS] ' . $prompt, $this->extKey, 0 );
		return $getVariantsStyle;
	}

	/**
	 * getVariantsStyle( )  :
	 *
	 * @return	string  style:
	 * @access public
	 * @internal     #t0915
	 * @version     7.2.0
	 * @since       7.2.0
	 */
	public function getVariantsStyleDRSwtcart()
	{
		$getVariantsStyle = 'wt_cart';

		if( !$this->drs->drsVariants )
		{
			return $getVariantsStyle;
		}

		$prompt = 'settings.variant has elements: we have the wt_cart style for variants.';
		\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[INFO/VARIANTS] ' . $prompt, $this->extKey, 0 );
		return $getVariantsStyle;
	}

	/*	 * *********************************************
	 *
	 * Payment
	 *
	 * ******************************************** */

	/**
	 * paymentUpdate( ) : Change the payment method in session
	 *
	 * @param	integer		: $value
	 * @param	integer		: $pid   : uid of the page, which contains the caddy plugin
	 * @return	void
	 * @access public
	 * @version     6.0.3
	 * @since       1.4.6
	 */
	public function paymentUpdate( $value, $pid = null )
	{
		$pid = $this->getPid( $pid );

		$sesArray = $GLOBALS[ 'TSFE' ]->fe_user->getKey( 'ses', $this->extKey . '_' . $pid );
		$sesArray[ 'payment' ] = intval( $value );
		$epayment = $this->pObj->conf[ 'api.' ][ 'options.' ][ 'payment.' ][ 'options.' ][ $value . '.' ][ 'e-payment' ];
		$sesArray[ 'e-payment' ][ 'woEpayment' ] = !$epayment;
//var_dump( __METHOD__, __LINE__, $sesArray, $this->pObj->conf );
//var_dump( __METHOD__, __LINE__, $sesArray );
//die( ":(" );
//var_dump( __METHOD__, __LINE__, $sesArray );
		$GLOBALS[ 'TSFE' ]->fe_user->setKey( 'ses', $this->extKey . '_' . $pid, $sesArray );
		$GLOBALS[ 'TSFE' ]->storeSessionData();
	}

	/**
	 * paymentGet( )  : get the payment method from session
	 *
	 * @param	integer		: $pid  : uid of the page, which contains the caddy plugin
	 * @return	integer
	 * @access public
	 * @version     3.0.1
	 * @since       1.4.6
	 */
	public function paymentGet( $pid = null )
	{
		$pid = $this->getPid( $pid );

		// #54634, 131128, dwildt, 1-
		//$sesArray = $GLOBALS['TSFE']->fe_user->getKey('ses', $this->extKey . '_' . $GLOBALS["TSFE"]->id);
		// #54634, 131128, dwildt, 1+
		$sesArray = $GLOBALS[ 'TSFE' ]->fe_user->getKey( 'ses', $this->extKey . '_' . $pid );

		return $sesArray[ 'payment' ];
	}

	/*	 * *********************************************
	 *
	 * Product
	 *
	 * ******************************************** */

	/**
	 * productAdd( )  : Add product to session
	 *
	 *    array (
	 *      'title' => 'this is the title',
	 *      'amount' => 2,
	 *      'gross' => '1,49',
	 *      'tax' => 1,
	 *      'uid' => 234,
	 *      'sku' => 'P234whatever'
	 *    )
	 *
	 * @param	array		: $newProduct :
	 * @param	integer		: $pid        : uid of the page, which contains the caddy plugin
	 * @return	void
	 * @access public
	 * @version     7.2.6
	 * @since       2.0.11
	 */
	public function productAdd( $newProduct, $pid = null )
	{
		$variants = null;
		$sesArray = array();

		// RETURN : requirements aren't matched
		if( !$this->productAddRequirements( $newProduct ) )
		{
			return false;
		}

		// #i0059
		$newProduct[ 'qty' ] = $this->productAddGetDefaultQty( $newProduct[ 'qty' ] );

		// RETURN : requirements aren't matched
		// #55726, 140206, dwildt, 1+
		$this->init();

		// Get current products
		$currProducts = $this->productsGet( $pid );
		// Reset error messages
		$currProducts = $this->productResetErrorPrompt( $currProducts );

		// #i0173, 171124, dwildt, 1+
		$variants = $this->variantsInit( $newProduct[ 'uid' ] );

		$arrResult = $this->productAddQuantity( $newProduct, $currProducts, $variants );
		$currProducts = $arrResult[ 'currProducts' ];
		$newProduct = $arrResult[ 'newProduct' ];

		// limit quantity
		$newProduct = $this->quantityCheckMinMax( $newProduct );
		// move comma to dot
		$newProduct[ 'gross' ] = $this->zz_commaToDot( $newProduct[ 'gross' ] );

		$newProduct = $this->variantsSet( $variants, $newProduct );

		// add product to the session array
		$currProducts[] = $newProduct;

		$pid = $this->getPid( $pid );
		// #i0044, 140206, dwildt, 1+
		$sesArray = $GLOBALS[ 'TSFE' ]->fe_user->getKey( 'ses', $this->extKey . '_' . $pid );
//var_dump(__METHOD__, __LINE__, $this->extKey . '_' . $pid);
//die();
		// Overwrite products
		$sesArray[ 'products' ] = $currProducts;

		$GLOBALS[ 'TSFE' ]->fe_user->setKey( 'ses', $this->extKey . '_' . $pid, $sesArray );
		// save session
		$GLOBALS[ 'TSFE' ]->storeSessionData();
	}

	/**
	 * productAddQuantity( )  : Calculate quantity, if item is part of caddy (update)
	 *
	 * @param	array		: $newProduct
	 * @param	array		: $currProducts
	 * @param	array		: $variants
	 * @return	array : $newProduct, $currProducts
	 * @access private
	 * @internal #i0121
	 * @version     7.5.8
	 * @since       7.2.6
	 */
	private function productAddQuantity( $newProduct, $currProducts, $variants )
	{
		switch( TRUE )
		{
			case( $variants !== NULL ):
				$arrResult = $this->productAddQuantityWiVariants( $variants, $newProduct, $currProducts );
				break;
			case( $variants === NULL ):
			default:
				$arrResult = $this->productAddQuantityWoVariants( $newProduct, $currProducts );
				break;
		}

		return $arrResult;
	}

	/**
	 * productAddQuantityWiVariants( )  : Calculate quantity, if item is part of caddy (update)
	 *
	 * @param	array		: $newProduct
	 * @param	array		: $currProducts
	 * @return	array : $newProduct, $currProducts
	 * @access private
	 * @internal #i0121, #i0173
	 * @version     7.5.7
	 * @since       7.2.6
	 */
	private function productAddQuantityWiVariants( $variants, $newProduct, $currProducts )
	{
		$drsCurrProductQty = 0;
		foreach( ( array ) $currProducts as $key => $currProduct )
		{
			$matchedVariants = $this->variantsMatched( $variants, $currProduct );
			// all conditions fit
			if( $matchedVariants == count( $variants ) )
			{
				$newProduct[ 'qty' ] = $currProducts[ $key ][ 'qty' ] + $newProduct[ 'qty' ];
				$drsCurrProductQty = $currProducts[ $key ][ 'qty' ];
				// remove product from current products, it will replaced by new product
				unset( $currProducts[ $key ] );
			}
		}
		$this->productAddQuantityWiVariantsDRS( $newProduct, $drsCurrProductQty );

		return array(
				'newProduct' => $newProduct,
				'currProducts' => $currProducts,
		);
	}

	/**
	 * productAddQuantityWiVariantsDRS( )  : 
	 * 
	 * @param	array		: $newProduct
	 * @param	array		: $currProduct
	 * @param	integer		: $key
	 * @return	void
	 * @access private
	 * @version     7.5.8
	 * @since       7.5.8
	 */
	private function productAddQuantityWiVariantsDRS( $newProduct, $drsCurrProductQty )
	{
		if( !$this->drs->drsQuantity )
		{
			return;
		}

		if( $drsCurrProductQty == 0 )
		{
			$prompt = 'Quantity of the new item: #' . $newProduct[ 'qty' ];
			\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[QUANTITY] ' . $prompt, $this->extKey, 0 );
			return;
		}

		$prompt = 'Quantity of the item in the caddy before updating #' . $drsCurrProductQty . ' and after updating #' . $newProduct[ 'qty' ];
		\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[QUANTITY] ' . $prompt, $this->extKey, 0 );
	}

	/**
	 * productAddQuantityWoVariants( )  : Calculate quantity, if item is part of caddy (update)
	 *
	 * @param	array		: $newProduct
	 * @param	array		: $currProducts
	 * @return	array : $newProduct, $currProducts
	 * @access private
	 * @internal #i0121
	 * @version     7.2.6
	 * @since       7.2.6
	 */
	private function productAddQuantityWoVariants( $newProduct, $currProducts )
	{
		// remove product from current products, sum quantity
		// FOREACH  : current product
		foreach( ( array ) $currProducts as $key => $currProduct )
		{
			if( $currProduct[ 'uid' ] != $newProduct[ 'uid' ] )
			{
				continue;
			}
			$newProduct[ 'qty' ] = $currProducts[ $key ][ 'qty' ] + $newProduct[ 'qty' ];
			// remove product from current products
			unset( $currProducts[ $key ] );
		}
		// FOREACH  : current product
		// remove product from current products, sum quantity

		return array(
				'newProduct' => $newProduct,
				'currProducts' => $currProducts,
		);
	}

	/**
	 * productAddGetDefaultQty( )  : Returns 1, if given qty isn't set or is 0
	 *
	 * @param	integer     $qty  : quantity of the current item
	 * @return	integer
	 * @access private
	 * @version     6.0.0
	 * @since       6.0.0
	 */
	private function productAddGetDefaultQty( $qty )
	{
		if( $qty > 0 )
		{
			return $qty;
		}

		return 1;
	}

	/**
	 * productAddRequirements( ) :
	 *
	 * @param	array		$product:
	 * @return	boolean		true, if requirements are matched
	 * @access private
	 * @internal    #i0024
	 * @version     2.0.11
	 * @since       2.0.11
	 */
	private function productAddRequirements( $product )
	{
		// RETURN : without price or without title
		switch( true )
		{
			// #i0211, 180216, dwildt, 1-/+
			//case( empty( $product[ 'gross' ] ) ):
			case(!isset( $product[ 'gross' ] ) ):
				// DRS
				if( $this->drs->drsWarn )
				{
					$prompt = 'Aborted: Item record is without the element "gross"!';
					\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[WARN/SESSION] ' . $prompt, $this->extKey, 3 );
				}
				return false;
			case( empty( $product[ 'title' ] ) ):
				// DRS
				if( $this->drs->drsWarn )
				{
					$prompt = 'Aborted: Item record is without the element "title"!';
					\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[WARN/SESSION] ' . $prompt, $this->extKey, 3 );
				}
				return false;
		}
		// RETURN : without price or without title

		unset( $product );

		return true;
	}

	/**
	 * productDelete( )  : Remove product from session with given uid
	 *
	 * @param	integer		: $pid   : uid of the page, which contains the caddy plugin
	 * @return	void
	 * @version  4.0.7
	 * @since    1.4.6
	 */
	public function productDelete( $pid = null )
	{
		// #55726, 140206, dwildt, 1+
		$this->init();

		$pid = $this->getPid( $pid );

		// variants
		// add variant key/value pairs from piVars
		$arr_variant = $this->productGetVariantGpvar();
		// add product id to variant array
		$arr_variant[ 'uid' ] = $this->pObj->piVars[ 'del' ];
		// get already exting products from session
		// #54634, 131128, dwildt, 1-
		//$sesArray = $GLOBALS['TSFE']->fe_user->getKey( 'ses', $this->extKey . '_' . $GLOBALS["TSFE"]->id );
		// #54634, 131128, dwildt, 1+
		$sesArray = $GLOBALS[ 'TSFE' ]->fe_user->getKey( 'ses', $this->extKey . '_' . $pid );

		// loop every product
		foreach( array_keys( ( array ) $sesArray[ 'products' ] ) as $key )
		{
			// Reset error messages
			unset( $sesArray[ 'products' ][ $key ][ 'error' ] );

			// Counter for condition
			$int_counter = 0;

			// loop through conditions
			foreach( ( array ) $arr_variant as $key_variant => $value_variant )
			{
				// condition fits
				if( $sesArray[ 'products' ][ $key ][ $key_variant ] == $value_variant )
				{
					$int_counter++;
				}
			}
			// loop through conditions
			// all conditions fit
			if( $int_counter == count( $arr_variant ) )
			{
				// remove product from session
				unset( $sesArray[ 'products' ][ $key ] );
			}
		}
		// loop every product
		// generate new session
		// #54634, 131128, dwildt, 1-
		//$GLOBALS['TSFE']->fe_user->setKey( 'ses', $this->extKey . '_' . $GLOBALS["TSFE"]->id, $sesArray );
		// #54634, 131128, dwildt, 1+
		$GLOBALS[ 'TSFE' ]->fe_user->setKey( 'ses', $this->extKey . '_' . $pid, $sesArray );
		// save session
		$GLOBALS[ 'TSFE' ]->storeSessionData();


		// #54634, 131128, dwildt, 1-
		//$sesArray = $GLOBALS['TSFE']->fe_user->getKey( 'ses', $this->extKey . '_' . $GLOBALS["TSFE"]->id );
		// #54634, 131128, dwildt, 1+
		$sesArray = $GLOBALS[ 'TSFE' ]->fe_user->getKey( 'ses', $this->extKey . '_' . $pid );
		$productId = $this->productsGetFirstKey();
		if( empty( $productId ) )
		{
			return;
		}
		$sesArray[ 'products' ][ $productId ] = $this->quantityCheckMinMax( $sesArray[ 'products' ][ $productId ] );
		// #54634, 131128, dwildt, 1-
		//$GLOBALS['TSFE']->fe_user->setKey( 'ses', $this->extKey . '_' . $GLOBALS["TSFE"]->id, $sesArray );
		// #54634, 131128, dwildt, 1+
		$GLOBALS[ 'TSFE' ]->fe_user->setKey( 'ses', $this->extKey . '_' . $pid, $sesArray );
		// save session
		$GLOBALS[ 'TSFE' ]->storeSessionData();
	}

	/**
	 * read product details (title, price from table)
	 * the method productGetDetails of version 1.2.1 became productGetDetailsTs from version 1.2.2
	 *
	 * @param	array		$uid: product uid
	 * @return	array		$arr: array with title and price
	 * @version 6.0.0
	 * @since 1.2.2
	 */
	public function productGet( $uid )
	{
		if( !empty( $this->pObj->conf[ 'db.' ][ 'sql' ] ) )
		{
			$product = $this->productGetDetailsSql( $uid );
			return $product;
		}

		// #i0261, 180913, dwildt, 1+
		$this->initInstancesSql();
		$product = $this->productGetDetailsTs( $uid );
		//var_dump( __METHOD__, __LINE__, $product );
		return $product;
	}

	/**
	 * productsGetFirstKey( ) :
	 *
	 * @return	integer		$uid: uid of the first item in the caddy
	 * @access private
	 * @version 2.0.0
	 * @since 2.0.0
	 */
	private function productsGetFirstKey()
	{
		$products = $this->productsGet();
		$productsKey = array_keys( ( array ) $products );
		$uid = $productsKey[ 0 ];

		return $uid;
	}

	/**
	 * productGetDetailsSql( )  :   read product details by a manually configured sql query
	 *
	 * @param	array		$uid: product uid
	 * @return	array		$arr: array with title and price
	 * @access private
	 * @version 2.0.11
	 * @since 1.4.6
	 */
	private function productGetDetailsSql( $uid )
	{
		// RETURN : any sql result isn't needed
		if( $this->productGetDetailsSqlIsNotNeeded() )
		{
			return false;
		}
		// RETURN : any sql result isn't needed
		// replace gp:marker and enable_fields:marker in $pObj->conf['db.']['sql']
		$this->zz_sqlReplaceMarker();
		// #42154, 101218, dwildt, 1-
		//$query = $pObj->cObj->stdWrap($pObj->conf['db.']['sql'], $pObj->conf['db.']['sql.']);
		// #42154, 101218, dwildt, 1+
		$name = $this->pObj->conf[ 'db.' ][ 'sql' ];
		$conf = $this->pObj->conf[ 'db.' ][ 'sql.' ];
		$query = $this->pObj->cObj->cObjGetSingle( $name, $conf );
		// execute the query
		$res = $GLOBALS[ 'TYPO3_DB' ]->sql_query( $query );

		// DRS
		$boolPromptGP = true;
		$prompt = NULL;
		$this->zz_sqlDRS( $query, $prompt, __METHOD__, __LINE__, $boolPromptGP );

		// RETURN false : no SQL ressource
		if( !$res )
		{
			return false;
		}
		// RETURN false : no SQL ressource
		// WHILE  : get a row with a title field
		while( $row = $GLOBALS[ 'TYPO3_DB' ]->sql_fetch_assoc( $res ) )
		{
			if( $row[ 'title' ] != null )
			{
				break;
			}
		}
		// WHILE  : get a row with a title field

		if( empty( $row ) )
		{
			// DRS
			$boolPromptGP = true;
			$prompt = 'Row is empty or row doesn\'t have any field called title.';
			$this->zz_sqlDRS( $query, $prompt, __METHOD__, __LINE__, $boolPromptGP );
			return false;
		}

		// Add or overwrite uid field
		$row[ 'uid' ] = $uid;

		// DRS
		if( $this->drs->drsSql )
		{
			$prompt = var_export( $row, true );
			\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[INFO/SQL] ' . $prompt, $this->extKey, 0 );
		}
		// DRS

		return $row;
	}

	/**
	 * productGetDetailsSqlIsNotNeeded( )  :
	 *
	 * @return	boolean		true, if SQL result is not needed
	 * @access private
	 * @version 2.0.11
	 * @since   2.0.11
	 */
	private function productGetDetailsSqlIsNotNeeded()
	{
		// RETURN : there isn't any GET or POST parameter
		if( (!\TYPO3\CMS\Core\Utility\GeneralUtility::_GET() ) && (!\TYPO3\CMS\Core\Utility\GeneralUtility::_POST() ) )
		{
			return true;
		}
		// RETURN : there isn't any GET or POST parameter

		$GP = \TYPO3\CMS\Core\Utility\GeneralUtility::_GET() + \TYPO3\CMS\Core\Utility\GeneralUtility::_POST()
		;
//var_dump( $GP, $this->pObj->piVars );
		// RETURN : no sql result needed
		switch( true )
		{
			// caddy is called without the page id only
			case( count( $GP ) == 1 && empty( $this->pObj->piVars ) ):
			// case is update of quantity only
			case( intval( $this->pObj->piVars[ 'updateByCaddy' ] ) ):
			// case is removing the current item
			case( intval( $this->pObj->piVars[ 'del' ] ) ):
			// case is confirmation of ordering
			case( $GP[ 'tx_powermail_pi1' ][ 'action' ] == 'confirmation' ):
			// case is ordering is sent
			case( $GP[ 'tx_powermail_pi1' ][ 'action' ] == 'create' ):
			// ordering is canceled
			case( $GP[ 'tx_powermail_pi1' ][ 'action' ] == 'form' ):
				return true;
			// 130913, dwildt, 3+
			case( isset( $GP[ 'tx_powermail_pi1' ][ 'mailID' ] ) ):
				return true;
			default:
				// Follow the workflow
				break;
		}
		// RETURN : no sql result needed

		unset( $GP );

		return false;
	}

	/**
	 * productGetDetailsTs( ) : read product details (title, price from table)
	 *
	 * @param	integer		$uid: uid of the current product
	 * @return	array		$arr: array with title and price
	 * @access private
	 * @version 8.13.0
	 * @since 1.4.6
	 */
	private function productGetDetailsTs( $uid )
	{
		if( empty( $uid ) )
		{
			return false;
		}

		// #i0262, 180912, dwildt, 2+
		$row = $this->sql->Row( $uid );
		return $row;
	}

	/**
	 * productGetVariantGpvar(): Get variant values from piVars
	 *                              variant values have to be content of
	 *                              ts array variant and of piVars
	 *
	 * @return	array		$arr_variant: array with variant key/value pairs
	 * @access private
	 * @internal     #t0915
	 * @version 7.2.0
	 * @since 1.4.6
	 */
	private function productGetVariantGpvar()
	{
		$arr_variant = NULL;

		switch( $this->getVariantsStyle() )
		{
			case('caddy'):
				$arr_variant = $this->productGetVariantGpvarCaddy();
				break;
			case('wt_cart'):
				$arr_variant = $this->productGetVariantGpvarWtcart();
				break;
			case('novariants'):
			default:
				// Do nothing;
				break;
		}

		return $arr_variant;
	}

	/**
	 * productGetVariantGpvarCaddy(): Get variant values from piVars
	 *                              variant values have to be content of
	 *                              ts array variant and of piVars
	 *
	 * @return	array		$arr_variant: array with variant key/value pairs
	 * @access private
	 * @internal     #t0915
	 * @version 7.2.0
	 * @since 1.4.6
	 */
	private function productGetVariantGpvarCaddy()
	{
		$arr_variant = NULL;

		$arr_variant = NULL;

		// POST has precedence!
		$GP = \TYPO3\CMS\Core\Utility\GeneralUtility::_POST() + \TYPO3\CMS\Core\Utility\GeneralUtility::_GET();

		$table = $this->pObj->conf[ 'constant_editor.' ][ 'variants.' ][ 'tables.' ][ 'property' ];
		$field = 'uid';

		if( !isset( $GP[ $table ][ $field ] ) )
		{
			return $arr_variant;
		}

		$tableField = $table . '.' . $field;
		$csvValues = implode( ',', $GP[ $table ][ $field ] );
		$arr_variant[ $tableField ] = $csvValues;

//    var_dump( __METHOD__, __LINE__, $arr_variant );
//array(1) {
//  ["tx_quickshop_variantproperty.uid"]=>
//  string(3) "3,6"
//}
//    die();

		return $arr_variant;
	}

	/**
	 * productGetVariantGpvarWtcart(): Get variant values from piVars
	 *                              variant values have to be content of
	 *                              ts array variant and of piVars
	 *
	 * @return	array		$arr_variant: array with variant key/value pairs
	 * @access private
	 * @internal     #t0915
	 * @version 7.2.0
	 * @since 1.4.6
	 */
	private function productGetVariantGpvarWtcart()
	{
		$arr_variant = NULL;

		foreach( ( array ) $this->pObj->conf[ 'settings.' ][ 'variant.' ] as $tableField )
		{
			// piVars tableField doesn't contain variant key
			if( empty( $this->pObj->piVars[ $tableField ] ) )
			{
				continue;
			}
			list($table) = explode( '.', $tableField );
			$arr_variant[ $tableField ] = $this->zz_escapeStrForLike( $this->pObj->piVars[ $tableField ], $table );
		}

//    var_dump( __METHOD__, __LINE__, $arr_variant );
//array(1) {
//  ["tx_org_calentrance.uid"]=>
//  string(1) "1"
//}
		return $arr_variant;
	}

	/**
	 * productGetVariantTs():  Get an array with the variant values
	 *                                out of the current product
	 *
	 * @param	array		$product: array with product uid, title, tax, etc...
	 * @return	array		$arr_variant: array with variant key/value pairs
	 * @access public
	 * @internal     #t0915
	 * @version 7.2.0
	 * @since 1.4.6
	 */
	public function productGetVariantTs( $product )
	{
//    var_dump( __METHOD__, __LINE__);
		$arr_variant = NULL;

		switch( $this->getVariantsStyle() )
		{
			case('caddy'):
				$arr_variant = $this->productGetVariantTsCaddy( $product );
				break;
			case('wt_cart'):
				$arr_variant = $this->productGetVariantTsWtcart( $product );
				break;
			case('novariants'):
			default:
				// Do nothing;
				break;
		}

		return $arr_variant;
	}

	/**
	 * productGetVariantTsCaddy():  Get an array with the variant values
	 *                                out of the current product
	 *
	 * @param	array		$product: array with product uid, title, tax, etc...
	 * @return	array		$arr_variant: array with variant key/value pairs
	 * @access private
	 * @internal     #t0915
	 * @version 7.2.0
	 * @since 1.4.6
	 */
	private function productGetVariantTsCaddy( $product )
	{
		$arr_variant = NULL;

		$table = $this->pObj->conf[ 'constant_editor.' ][ 'variants.' ][ 'tables.' ][ 'property' ];
		$field = 'uid';
		$tableField = $table . '.' . $field;

		// RETURN NULL: product doesn't contain the variant tableField
		if( !in_array( $tableField, array_keys( $product ) ) )
		{
			return $arr_variant;
		}

		// RETURN NULL: product contains the variant tableField, but the value is empty
		if( empty( $product[ $tableField ] ) )
		{
			return $arr_variant;
		}

		// RETURN variant: product contains the variant tableField and a value is set
		$arr_variant[ $tableField ] = $product[ $tableField ];

//    var_dump( __METHOD__, __LINE__, $arr_variant );
//    array(2)
//    {
//      [ "uid" ] =>
//      string(1) "6"
//      [ "tx_quickshop_variantproperty.uid" ] =>
//      string(3) "3,6"
//    }
//    die();

		return $arr_variant;
	}

	/**
	 * productGetVariantTsWtcart():  Get an array with the variant values
	 *                                out of the current product
	 *
	 * @param	array		$product: array with product uid, title, tax, etc...
	 * @return	array		$arr_variant: array with variant key/value pairs
	 * @access private
	 * @internal     #t0915
	 * @version 7.2.0
	 * @since 7.2.0
	 */
	private function productGetVariantTsWtcart( $product )
	{
		$arr_variant = NULL;

		// loop through ts array variant
		foreach( ( array ) $this->pObj->conf[ 'settings.' ][ 'variant.' ] as $tableField )
		{
			// product doesn't contains variant $tableField
			if( !in_array( $tableField, array_keys( $product ) ) )
			{
				continue;
			}
			// product contains variant $tableField, but $tableField has no value
			if( empty( $product[ $tableField ] ) )
			{
				continue;
			}
			// product contains variant $tableField and but has a value
			$arr_variant[ $tableField ] = $product[ $tableField ];
		}

//		var_dump( __METHOD__, __LINE__, $arr_variant );
//    array(2)
//    {
//      [ "uid" ] =>
//      string(1) "6"
//      [ "tx_quickshop_variantproperty.uid" ] =>
//      string(3) "3,6"
//    }
//    die();

		return $arr_variant;
	}

	/**
	 * productResetErrorPrompt( )  :
	 *
	 * @param	array		$products:
	 * @return	array		$products:
	 * @version     2.0.11
	 * @since       1.4.6
	 */
	public function productResetErrorPrompt( $products )
	{
		// FOREACH  : current product
		foreach( array_keys( ( array ) $products ) as $product )
		{
			// Reset error messages
			unset( $products[ $product ][ 'error' ] );
		}

		return $products;
	}

	/**
	 * productsGet( ) : Read products from session
	 *
	 * @param	integer		: $pidCaddy : uid of the page, which contains the caddy plugin
	 * @return	array		: $arr      : array with all products from session
	 * @version    3.0.1
	 * @since      2.0.0
	 */
	public function productsGet( $pid = null )
	{
		$pid = $this->getPid( $pid );
		$sesArray = $GLOBALS[ 'TSFE' ]->fe_user->getKey( 'ses', $this->extKey . '_' . $pid );

		return $sesArray[ 'products' ];
	}

	/**
	 * productsGetGross( )  : Count gross price of all products in a cart
	 *                        Is used by pi3 only
	 *
	 * @param	integer		: $pidCaddy : uid of the page, which contains the caddy plugin
	 * @return	integer
	 * @internal #t0915
	 * @version    7.2.0
	 * @since      2.0.0
	 */
	public function productsGetGross( $pid = null )
	{
		// #54634, 131128, dwildt, 1+
		$pid = $this->getPid( $pid );
		// get already exting products from session
		$sesArray = $GLOBALS[ 'TSFE' ]->fe_user->getKey( 'ses', $this->extKey . '_' . $pid );

		$gross = 0;
		foreach( ( array ) $sesArray[ 'products' ] as $val )
		{
			$gross += ($val[ 'gross' ] + $val[ 'grossvariants' ]) * $val[ 'qty' ];
		}

		return $gross;
	}

	/**
	 * quantityCheckMinMax( )  : Checks
	 *                           * min and max limits depending on an item (database)
	 *                           * min and max limits depending on the caddy (plugin/flexform)
	 *                           If a limit is passed over, quantities will updated and there will be
	 *                           error prompts near the items.
	 *                           min and max limits of the caddy have precedence!
	 *                           Example:
	 *                           * An item have a maximum limit of 2
	 *                           * The caddy have a minimum limit of 4
	 *                           * quantityCheckMinMax( ) will update the quantity to 4 of the product
	 *
	 * @param	array		$product  : the current product
	 * @return	array		$product  : the current or the updated product
	 * @access private
	 * @version 7.5.8
	 * @since 2.0.0
	 */
	private function quantityCheckMinMax( $product )
	{
		// #i0226, 180312, dwildt, ~
		$this->minmax = \TYPO3\CMS\Core\Utility\GeneralUtility::makeInstance( \Netzmacher\Caddy\Lib\Caddy\Minmax::class );
		$this->minmax->setParentObject( $this );

		$product = $this->minmax->evaluate( $product );
		return $product;
	}

	/**
	 * quantityGetVariant(): Get variant values out of the name of the qty field
	 *                              variant values have to be content of
	 *                              ts array variant and of qty field
	 *
	 * @return	array		$arr_variant: array with variant key/value pairs
	 * @access private
	 * @internal     #t0915
	 * @version     7.2.0
	 * @since 1.4.6
	 */
	private function quantityGetVariant()
	{
		$arr_variant = NULL;
		switch( $this->getVariantsStyle() )
		{
			case('caddy'):
				$arr_variant = $this->quantityGetVariantCaddy();
				break;
			case('wt_cart'):
				$arr_variant = $this->quantityGetVariantWtcart();
				break;
			case('novariants'):
			default:
				// Do nothing
				break;
		}
//    var_dump( __METHOD__, __LINE__, $arr_variant );
//    die( ":(" );

		return $arr_variant;
	}

	/**
	 * quantityGetVariantCaddy(): Get variant values out of the name of the qty field
	 *                              variant values have to be content of
	 *                              ts array variant and of qty field
	 *
	 * @return	array		$arr_variant: array with variant key/value pairs
	 * @access private
	 * @internal     #t0915
	 * @version     7.2.0
	 * @since 7.2.0
	 */
	private function quantityGetVariantCaddy()
	{
		$arr_qty = null;
		$arr_variant = null;

		$int_counter = 0;
		foreach( ( array ) $this->pObj->piVars[ 'qty' ] as $key => $value )
		{
			$arr_qty[ $int_counter ][ 'qty' ][ $key ] = $value;
			$int_counter++;
		}

		$table = $this->pObj->conf[ 'constant_editor.' ][ 'variants.' ][ 'tables.' ][ 'property' ];
		$field = 'uid';
		$tableField = $table . '.' . $field;

		foreach( ( array ) $arr_qty as $key => $piVarsQty )
		{
			$arr_variant[ $key ] = $this->quantityGetVariantIterator( $piVarsQty );

// piVars contain variant key
			if( empty( $arr_variant[ $key ][ $tableField ] ) )
			{
				continue;
			}
			list( $table ) = explode( '.', $tableField );
			$arr_variant[ $key ][ $tableField ] = $this->zz_escapeStrForLike( $arr_variant[ $key ][ $tableField ], $table );
		}

		return $arr_variant;
	}

	/**
	 * quantityGetVariantIterator(): Get variant values out of the name of the qty field
	 *                              variant values have to be content of
	 *                              ts array variant and of qty field
	 *
	 * @return	array		$arr_variant: array with variant key/value pairs
	 * @access private
	 * @internal     #t0915
	 * @version     8.11.10
	 * @since 7.2.0
	 */
	private function quantityGetVariantIterator( $piVarsQty )
	{
		$arr_variant = array();

		// #i0244, 180618, dwildt, 3-/+
// iterator object
//		$data = new RecursiveArrayIterator( $piVarsQty[ 'qty' ] );
//		$iterator = new RecursiveIteratorIterator( $data, true );
		$qty = $piVarsQty[ 'qty' ];
		$data = \TYPO3\CMS\Core\Utility\GeneralUtility::makeInstance( \RecursiveArrayIterator::class, $qty );
		$iterator = \TYPO3\CMS\Core\Utility\GeneralUtility::makeInstance( \RecursiveIteratorIterator::class, $data, true );

// top level of ecursive array
		$iterator->rewind();

// get all variant key/value pairs from qty name
// #i0173, 171124, dwildt, 1-/+ 
//foreach( (array) $iterator as $key_iterator => $value_iterator )
		foreach( $iterator as $key_iterator => $value_iterator )
		{
// i.e for a key: tx_org_calentrance.uid=4
			list($tableField, $value) = explode( '=', $key_iterator );
			if( $tableField == 'uid' )
			{
				$arr_variant[ 'uid' ] = $value;
			}
// i.e arr_var[tx_org_calentrance.uid] = 4
			$arr_variant[ $tableField ] = $value;

			switch( TRUE )
			{
				case(is_array( $value_iterator )):
					list($tableField, $value) = explode( '=', key( $value_iterator ) );
					if( $tableField == 'uid' )
					{
						$arr_variant[ 'uid' ] = $value;
					}
					$arr_variant[ $tableField ] = $value;
					break;
				case(!is_array( $value_iterator )):
				default:
					$arr_variant[ 'qty' ] = $value_iterator;
					break;
			}
		}
		return $arr_variant;
	}

	/**
	 * quantityGetVariantWtcart(): Get variant values out of the name of the qty field
	 *                              variant values have to be content of
	 *                              ts array variant and of qty field
	 *
	 * @return	array		$arr_variant: array with variant key/value pairs
	 * @access private
	 * @internal     #t0915
	 * @version     7.2.0
	 * @since 1.4.6
	 */
	private function quantityGetVariantWtcart()
	{
		$arr_qty = null;
		$arr_variant = null;

		$int_counter = 0;
		foreach( ( array ) $this->pObj->piVars[ 'qty' ] as $key => $value )
		{
			$arr_qty[ $int_counter ][ 'qty' ][ $key ] = $value;
			$int_counter++;
		}

		foreach( ( array ) $arr_qty as $key => $piVarsQty )
		{
			$arr_variant[ $key ] = $this->quantityGetVariantIterator( $piVarsQty );

// loop through ts variant array
			foreach( ( array ) $this->pObj->conf[ 'settings.' ][ 'variant.' ] as $tableField )
			{
				// piVars contain variant key
				if( empty( $arr_variant[ $key ][ $tableField ] ) )
				{
					continue;
				}
				list( $table ) = explode( '.', $tableField );
				$arr_variant[ $key ][ $tableField ] = $this->zz_escapeStrForLike( $arr_variant[ $key ][ $tableField ], $table );
			}
		}

		return $arr_variant;
	}

	/**
	 * quantityUpdate( )  : Change quantity of a product in session
	 *
	 * @param       integer : $pid   : uid of the page, which contains the caddy plugin
	 * @return	void
	 * @access public
	 * @version 7.2.5
	 * @since 1.4.6
	 */
	public function quantityUpdate( $pid = null )
	{
		$this->init();
		$pid = $this->getPid( $pid );
		$arr_variant = $this->quantityGetVariant();

		// get products from session
		$sesArray = $GLOBALS[ 'TSFE' ]->fe_user->getKey( 'ses', $this->extKey . '_' . $pid );
		// #i0082, 161115, dwildt, 1+
		$sesArray[ 'products' ] = $this->productResetErrorPrompt( $sesArray[ 'products' ] );

		// LOOP : each product
		foreach( array_keys( ( array ) $sesArray[ 'products' ] ) as $productId )
		{
			switch( TRUE )
			{
				case( is_array( $arr_variant ) ):
					$sesArray = $this->quantityUpdateWiVariant( $arr_variant, $sesArray, $productId, $pid );
					break;
				case(!is_array( $arr_variant ) ):
				default:
					$sesArray = $this->quantityUpdateWoVariant( $sesArray, $productId, $pid );
					break;
			}
		} // LOOP : each product
		$GLOBALS[ 'TSFE' ]->fe_user->setKey( 'ses', $this->extKey . '_' . $pid, $sesArray );
		$GLOBALS[ 'TSFE' ]->storeSessionData();
	}

	/**
	 * quantityUpdateWiVariant( ) : 
	 *
	 * @param       array   : variant
	 * @param       array   : session
	 * @param       string
	 * @return      array   : session
	 * @access private
	 * @version 7.5.8
	 * @since 7.2.5
	 */
	private function quantityUpdateWiVariant( $arr_variant, $sesArray, $productId, $pid )
	{
		$session_uid = $sesArray[ 'products' ][ $productId ][ 'uid' ];
		$updateByCaddy = intval( $this->pObj->piVars[ 'updateByCaddy' ] );

		foreach( ( array ) $arr_variant as $key_variant => $arr_condition )
		{
			$sesArray = $this->quantityUpdateWiVariantConditions( $arr_variant, $key_variant, $session_uid, $sesArray, $productId, $updateByCaddy, $pid, $arr_condition );
		}

		if( isset( $sesArray[ 'products' ][ $productId ] ) )
		{
			$sesArray[ 'products' ][ $productId ] = $this->quantityCheckMinMax( $sesArray[ 'products' ][ $productId ] );
		}

		return $sesArray;
	}

	/**
	 * quantityUpdateWiVariantConditions( ) : 
	 *
	 * @param       array   : variant
	 * @param       array   : session
	 * @param       string
	 * @return      array   : session
	 * @access private
	 * @version 7.5.8
	 * @since 7.5.8
	 */
	private function quantityUpdateWiVariantConditions( $arr_variant, $key_variant, $session_uid, $sesArray, $productId, $updateByCaddy, $pid, $arr_condition )
	{
		$uid = $this->quantityUpdateWiVariantGetUid( $arr_variant, $key_variant );
		$qty = $this->quantityUpdateWiVariantGetQty( $arr_variant, $key_variant, $uid );
		$conditionCounter = $this->quantityUpdateWiVariantConditionsCounter( $session_uid, $sesArray, $productId, $uid, $arr_condition );
		$sesArray = $this->quantityUpdateWiVariantItem( $conditionCounter, $arr_condition, $qty, $updateByCaddy, $sesArray, $productId, $pid );

		return $sesArray;
	}

	/**
	 * quantityUpdateWiVariantConditionsCounter( ) : 
	 *
	 * @param       array   : variant
	 * @param       array   : session
	 * @param       string
	 * @return      integer   : conditionCounter
	 * @access private
	 * @version 7.5.8
	 * @since 7.5.8
	 */
	private function quantityUpdateWiVariantConditionsCounter( $session_uid, $sesArray, $productId, $uid, $arr_condition )
	{
// counter for condition
		$conditionCounter = 0;
// uid: condition fits
		if( $session_uid == $uid )
		{
			$conditionCounter++;
		}

// loop through conditions
		foreach( ( array ) $arr_condition as $key_condition => $value_condition )
		{
// workaround (it would be better, if qty and uid won't be elements of $arr_condition
			if( in_array( $key_condition, array( 'qty', 'uid' ) ) )
			{
				// workaround: uid and qty should fit in every case
				$conditionCounter++;
			}
// workaround (it would be better, if qty and uid won't be elements of $arr_condition
			if( !in_array( $key_condition, array( 'qty', 'uid' ) ) )
			{
				// variants: condition fits
				if( $sesArray[ 'products' ][ $productId ][ $key_condition ] == $value_condition )
				{
					//$prompt_315[] = 'div 315: true - session_key : ' . $productId . ', ' . $key_condition . ' : ' . $value_condition;
					$conditionCounter++;
				}
			}
		}
		return $conditionCounter;
	}

	/**
	 * quantityUpdateWiVariantGetUid( ) : 
	 *
	 * @param       array   : variant
	 * @param       string
	 * @return      integer : uid
	 * @access private
	 * @version 7.5.8
	 * @since 7.5.8
	 */
	private function quantityUpdateWiVariantGetQty( $arr_variant, $key_variant, $uid )
	{
		switch( true )
		{
			case(isset( $arr_variant[ $key_variant ][ 'qty' ] )):
				$qty = intval( $arr_variant[ $key_variant ][ 'qty' ] );
				break;
			case(!isset( $arr_variant[ $key_variant ][ 'qty' ] )):
				$qty = intval( $this->pObj->piVars[ 'qty' ][ $uid ] );
				break;
		}

		return $qty;
	}

	/**
	 * quantityUpdateWiVariantGetUid( ) : 
	 *
	 * @param       array   : variant
	 * @param       string
	 * @return      integer : uid
	 * @access private
	 * @version 7.5.8
	 * @since 7.5.8
	 */
	private function quantityUpdateWiVariantGetUid( $arr_variant, $key_variant )
	{
		switch( true )
		{
			case(isset( $arr_variant[ $key_variant ][ 'uid' ] )):
				$uid = $arr_variant[ $key_variant ][ 'uid' ];
				break;
			case(!isset( $arr_variant[ $key_variant ][ 'uid' ] )):
				$uid = key( $this->pObj->piVars[ 'qty' ] );
				break;
		}

		return $uid;
	}

	/**
	 * quantityUpdateWiVariantItem( ) : 
	 *
	 * @param       array   : variant
	 * @param       array   : session
	 * @param       string
	 * @return      array   : session
	 * @access private
	 * @version 7.5.8
	 * @since 7.5.8
	 */
	private function quantityUpdateWiVariantItem( $conditionCounter, $arr_condition, $qty, $updateByCaddy, $sesArray, $productId, $pid )
	{
		// conditions doesn't fit
		if( $conditionCounter != ( count( $arr_condition ) + 1 ) )
		{
			return $sesArray;
		}

		switch( true )
		{
			case( $qty > 0):
				$sesArray = $this->quantityUpdateWiVariantItemUpdate( $qty, $updateByCaddy, $sesArray, $productId );
				break;
			case(!($qty > 0)):
			default:
				$sesArray = $this->quantityUpdateWiVariantItemDelete( $sesArray, $productId, $pid );
				break;
		}

		return $sesArray;
	}

	/**
	 * quantityUpdateWiVariantItemDelete( ) : 
	 *
	 * @param       array   : variant
	 * @param       array   : session
	 * @param       string
	 * @return      array   : session
	 * @access private
	 * @version 7.5.8
	 * @since 7.5.8
	 */
	private function quantityUpdateWiVariantItemDelete( $sesArray, $productId, $pid )
	{
// remove product from session
		$this->productDelete( $pid );
// remove product from current session array
		unset( $sesArray[ 'products' ][ $productId ] );

		$this->quantityUpdateWiVariantItemDeleteDRS( $pid );
		return $sesArray;
	}

	/**
	 * quantityUpdateWiVariantItemDeleteDRS( )  : 
	 * 
	 * @param	integer		: $pid
	 * @return	void
	 * @access private
	 * @version     7.5.8
	 * @since       7.5.8
	 */
	private function quantityUpdateWiVariantItemDeleteDRS( $pid )
	{
		if( !$this->drs->drsQuantity )
		{
			return;
		}

		$prompt = 'Quantity is set to 0. Item #' . $pid . ' is removed from session.';
		\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[QUANTITY] ' . $prompt, $this->extKey, 0 );
	}

	/**
	 * quantityUpdateWiVariantItemUpdate( ) : 
	 *
	 * @param       array   : variant
	 * @param       array   : session
	 * @param       string
	 * @return      array   : session
	 * @access private
	 * @version 7.5.8
	 * @since 7.5.8
	 */
	private function quantityUpdateWiVariantItemUpdate( $qty, $updateByCaddy, $sesArray, $productId )
	{
		switch( true )
		{
			case($updateByCaddy):
				$sesArray[ 'products' ][ $productId ][ 'qty' ] = $qty;
				break;
			default:
				$sesArray[ 'products' ][ $productId ][ 'qty' ] = $sesArray[ 'products' ][ $productId ][ 'qty' ] + $qty;
				break;
		}
		$this->quantityUpdateWiVariantItemUpdateDRS( $sesArray[ 'products' ][ $productId ][ 'qty' ] );
		return $sesArray;
	}

	/**
	 * quantityUpdateWiVariantItemDeleteDRS( )  : 
	 * 
	 * @param	integer		: $pid
	 * @return	void
	 * @access private
	 * @version     7.5.8
	 * @since       7.5.8
	 */
	private function quantityUpdateWiVariantItemUpdateDRS( $qty )
	{
		if( !$this->drs->drsQuantity )
		{
			return;
		}

		$prompt = 'Quantity is updated to #' . $qty;
		\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[QUANTITY] ' . $prompt, $this->extKey, 0 );
	}

	/**
	 * quantityUpdateWoVariant( ) : 
	 *
	 * @param       array   : session
	 * @param       string
	 * @return      array   : session
	 * @access private
	 * @version 7.2.5
	 * @since 7.2.5
	 */
	private function quantityUpdateWoVariant( $sesArray, $productId, $pid )
	{
// no variant, nothing to loop
		$session_uid = $sesArray[ 'products' ][ $productId ][ 'uid' ];
		$int_qty = intval( $this->pObj->piVars[ 'qty' ][ $session_uid ] );
		$updateByCaddy = intval( $this->pObj->piVars[ 'updateByCaddy' ] );

		switch( TRUE )
		{
			case($int_qty > 0):
				// update quantity
				if( $updateByCaddy )
				{
					// update from cart then set new qty
					$sesArray[ 'products' ][ $productId ][ 'qty' ] = $int_qty;
				}
				else
				{
					// update not from cart then add qty
					$sesArray[ 'products' ][ $productId ][ 'qty' ] = $sesArray[ 'products' ][ $productId ][ 'qty' ] + $int_qty;
				}
				$productId = $productId;
				break;
			case($int_qty == 0):
			default:
				// remove product from session
				$this->productDelete( $pid );
				// remove product from current session array
				unset( $sesArray[ 'products' ][ $productId ] );
				$productId = $this->productsGetFirstKey();
				break;
		}
		$product = $this->quantityCheckMinMax( $sesArray[ 'products' ][ $productId ] );

		if( !empty( $product ) )
		{
			$sesArray[ 'products' ][ $productId ] = $product;
		}
		return $sesArray;
	}

	/*	 * *********************************************
	 *
	 * Session
	 *
	 * ******************************************** */

	/**
	 * sessionDelete( )
	 *
	 * @param	string		: $content  : current content (will be empty)
	 * @param	array		: $conf     : current TypoScript configuration
	 * @param	integer		: $pid      : uid of the page, which contains the caddy plugin
	 * @return	void
	 * @access public
	 * @version    9.1.4
	 * @since      2.0.0
	 */
	public function sessionDelete( $content = '', $conf = array(), $pid = null )
	{
		\Netzmacher\Caddy\Userfunc\ClearAll::Caddy( $content, $conf, $pid );
		return;

// #i0322, 200218, dwildt, -		
//		$pid = $this->getPid( $pid );
//
//// DRS
//		unset( $content );
//		$drs = false;
//		if( $conf[ 'userFunc.' ][ 'drs' ] )
//		{
//			$drs = true;
//			$prompt = 'DRS is enabled by userfunc ' . __METHOD__ . '[userFunc.][drs].';
//			\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[INFO/USERFUNC] ' . $prompt, $this->extKey, 0 );
//		}
//		if( $this->drs->drsSession || $drs )
//		{
//			$prompt = 'Session is cleared.';
//			\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[INFO/SESSION] ' . $prompt, $this->extKey, 0 );
//		}
//// DRS
//// Increase numbers
//		$this->sessionDeleteIncreaseNumbers( $drs, $pid );
//
//// Delete the session
//// #54634, 131128, dwildt, 1-
////$GLOBALS['TSFE']->fe_user->setKey( 'ses', $this->extKey . '_' . $GLOBALS["TSFE"]->id, array( ) );
//// #54634, 131128, dwildt, 1+
//		$GLOBALS[ 'TSFE' ]->fe_user->setKey( 'ses', $this->extKey . '_' . $pid, array() );
//		$GLOBALS[ 'TSFE' ]->storeSessionData();
	}

// #i0322, 200218, dwildt, -		
//	
//	/**
//	 * sessionDeleteIncreaseNumbers( )
//	 *
//	 * @param	[type]		$$drs: ...
//	 * @param	[type]		$pid: ...
//	 * @return	void
//	 * @access private
//	 * @version    2.0.0
//	 * @since      2.0.0
//	 */
//	private function sessionDeleteIncreaseNumbers( $drs, $pid )
//	{
//		$products = $this->productsGet( $pid );
//
//// RETURN : any product, don't increase numbers!
//		if( empty( $products ) )
//		{
//			if( $this->drs->drsSession || $drs )
//			{
//				$prompt = 'Session is empty! Maybe powermail form is sent twice!';
//				\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[ERROR/SESSION] ' . $prompt, $this->extKey, 3 );
//			}
//			return;
//		}
//// RETURN : any product, don't increase numbers!
//
//		$registry = new \TYPO3\CMS\Core\Registry;
//// #54634, 131128, dwildt, 1-
////$prefix = 'page_' . $GLOBALS["TSFE"]->id . '_';
//// #54634, 131128, dwildt, 1+
//		$prefix = 'page_' . $pid . '_';
//
//// Get current numbers
//		$numberDeliveryorder = ( int ) $registry->get( 'tx_caddy', $prefix . 'deliveryorder' );
//		$numberInvoice = ( int ) $registry->get( 'tx_caddy', $prefix . 'invoice' );
//		$numberOrder = ( int ) $registry->get( 'tx_caddy', $prefix . 'order' );
//// Get current numbers
//// Increase current numbers
//		$numberDeliveryorder = $numberDeliveryorder + 1;
//		$numberInvoice = $numberInvoice + 1;
//		$numberOrder = $numberOrder + 1;
//// Increase current numbers
//// Set current numbers
//		$registry->set( 'tx_caddy', $prefix . 'deliveryorder', $numberDeliveryorder );
//		$registry->set( 'tx_caddy', $prefix . 'invoice', $numberInvoice );
//		$registry->set( 'tx_caddy', $prefix . 'order', $numberOrder );
//// Set current numbers
//// DRS
//		if( $drs )
//		{
//			$prompt = 'New delivery order number: ' . $numberDeliveryorder;
//			\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( ' [INFO/SESSION] ' . $prompt, $this->extKey, 0 );
//			$prompt = 'New invoice number: ' . $numberInvoice;
//			\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( ' [INFO/SESSION] ' . $prompt, $this->extKey, 0 );
//			$prompt = 'New order number: ' . $numberOrder;
//			\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( ' [INFO/SESSION] ' . $prompt, $this->extKey, 0 );
//		}
//// DRS
//	}

	/*	 * *********************************************
	 *
	 * Setting methods
	 *
	 * ******************************************** */

	/**
	 * setParentObject( )  : Returns a caddy with HTML form and HTML options among others
	 *
	 * @param	[type]		$$pObj: ...
	 * @return	void
	 * @access public
	 * @version    2.0.0
	 * @since      2.0.0
	 */
	public function setParentObject( $pObj )
	{
		if( !is_object( $pObj ) )
		{
			$prompt = 'ERROR: no parent object!<br />' . PHP_EOL .
							'Sorry for the trouble.<br />' . PHP_EOL .
							'TYPO3 Caddy<br />' . PHP_EOL .
							__METHOD__ . ' (' . __LINE__ . ')';
			die( $prompt );
		}
		$this->pObj = $pObj;

//$prompt = 'debug trail: ' . t3lib_utility_Debug::debugTrail( ) . PHP_EOL .
//          'TYPO3 Caddy<br />' . PHP_EOL .
//        __METHOD__ . ' (' . __LINE__ . ')';
//echo $prompt;

		$this->setParentObjectDrs();
	}

	/**
	 * setParentObjectDrs( )  :
	 *
	 * @return	void
	 * @access private
	 * @version    2.0.0
	 * @since      2.0.0
	 */
	private function setParentObjectDrs()
	{
		if( !is_object( $this->pObj->drs ) )
		{
			$prompt = 'ERROR: no DRS object!<br />' . PHP_EOL .
							'Sorry for the trouble.<br />' . PHP_EOL .
//                'debug trail: ' . t3lib_utility_Debug::debugTrail( ) . PHP_EOL .
							'TYPO3 Caddy<br />' . PHP_EOL .
							__METHOD__ . ' (' . __LINE__ . ')';
			die( $prompt );
		}
		$this->drs = $this->pObj->drs;
	}

	/*	 * *********************************************
	 *
	 * Shipping
	 *
	 * ******************************************** */

	/**
	 * insuranceGet( ) : Get the insurance method from session
	 *
	 * @param	integer		: $pid  : uid of the page, which contains the caddy plugin
	 * @return	integer
	 * @internal	 #i0286
	 * @version    8.17.0
	 * @since      8.17.0
	 */
	public function insuranceGet( $pid = null )
	{
		$pid = $this->getPid( $pid );
		$sesArray = $GLOBALS[ 'TSFE' ]->fe_user->getKey( 'ses', $this->extKey . '_' . $pid );

		return $sesArray[ 'insurance' ];
	}

	/**
	 * insuranceUpdate( )  : Change the insurance method in session
	 *
	 * @param	integer		:
	 * @param	integer		: $pid      : uid of the page, which contains the caddy plugin
	 * @return	void
	 * @internal	 #i0286
	 * @version    8.17.0
	 * @since      8.17.0
	 */
	public function insuranceUpdate( $value, $pid = null )
	{
		$pid = $this->getPid( $pid );

		$sesArray = $GLOBALS[ 'TSFE' ]->fe_user->getKey( 'ses', $this->extKey . '_' . $pid );

		$sesArray[ 'insurance' ] = intval( $value ); // overwrite with new qty
		$GLOBALS[ 'TSFE' ]->fe_user->setKey( 'ses', $this->extKey . '_' . $pid, $sesArray );
		$GLOBALS[ 'TSFE' ]->storeSessionData(); // Save session
	}

	/**
	 * recyclingfeeGet( ) : Get the recyclingfee method from session
	 *
	 * @param	integer		: $pid  : uid of the page, which contains the caddy plugin
	 * @return	integer
	 * @internal	 #i0286
	 * @version    8.17.0
	 * @since      8.17.0
	 */
	public function recyclingfeeGet( $pid = null )
	{
		$pid = $this->getPid( $pid );
		$sesArray = $GLOBALS[ 'TSFE' ]->fe_user->getKey( 'ses', $this->extKey . '_' . $pid );

		return $sesArray[ 'recyclingfee' ];
	}

	/**
	 * recyclingfeeUpdate( )  : Change the recyclingfee method in session
	 *
	 * @param	integer		:
	 * @param	integer		: $pid      : uid of the page, which contains the caddy plugin
	 * @return	void
	 * @internal	 #i0286
	 * @version    8.17.0
	 * @since      8.17.0
	 */
	public function recyclingfeeUpdate( $value, $pid = null )
	{
		$pid = $this->getPid( $pid );

		$sesArray = $GLOBALS[ 'TSFE' ]->fe_user->getKey( 'ses', $this->extKey . '_' . $pid );

		$sesArray[ 'recyclingfee' ] = intval( $value ); // overwrite with new qty
		$GLOBALS[ 'TSFE' ]->fe_user->setKey( 'ses', $this->extKey . '_' . $pid, $sesArray );
		$GLOBALS[ 'TSFE' ]->storeSessionData(); // Save session
	}

	/**
	 * shippingGet( ) : Get the shipping method from session
	 *
	 * @param	integer		: $pid  : uid of the page, which contains the caddy plugin
	 * @return	integer
	 * @version     3.0.1
	 * @since       1.4.0
	 */
	public function shippingGet( $pid = null )
	{
		$pid = $this->getPid( $pid );
		$sesArray = $GLOBALS[ 'TSFE' ]->fe_user->getKey( 'ses', $this->extKey . '_' . $pid );

		return $sesArray[ 'shipping' ];
	}

	/**
	 * shippingUpdate( )  : Change the shipping method in session
	 *
	 * @param	integer		:
	 * @param	integer		: $pid      : uid of the page, which contains the caddy plugin
	 * @return	void
	 * @version     3.0.1
	 * @since       1.4.0
	 */
	public function shippingUpdate( $value, $pid = null )
	{
		$pid = $this->getPid( $pid );

		$sesArray = $GLOBALS[ 'TSFE' ]->fe_user->getKey( 'ses', $this->extKey . '_' . $pid );

		$sesArray[ 'shipping' ] = intval( $value ); // overwrite with new qty
		$GLOBALS[ 'TSFE' ]->fe_user->setKey( 'ses', $this->extKey . '_' . $pid, $sesArray );
		$GLOBALS[ 'TSFE' ]->storeSessionData(); // Save session
	}

	/*	 * *********************************************
	 *
	 * Special
	 *
	 * ******************************************** */

	/**
	 * specialUpdate( ) : Change the special method in session
	 *
	 * @param	array		: $arr  : array to change
	 * @param	integer		: $pid  : uid of the page, which contains the caddy plugin
	 * @return	void
	 * @version     3.0.1
	 * @since       1.4.0
	 */
	public function specialUpdate( $specials_arr, $pid = null )
	{
		$pid = $this->getPid( $pid );

// get products from session
// #54634, 131128, dwildt, 1-
//$sesArray = $GLOBALS['TSFE']->fe_user->getKey('ses', $this->extKey . '_' . $GLOBALS["TSFE"]->id);
// #54634, 131128, dwildt, 1+
		$sesArray = $GLOBALS[ 'TSFE' ]->fe_user->getKey( 'ses', $this->extKey . '_' . $pid );

		$sesArray[ 'specials' ] = $specials_arr; // overwrite with new qty
// #54634, 131128, dwildt, 1-
//$GLOBALS['TSFE']->fe_user->setKey('ses', $this->extKey . '_' . $GLOBALS["TSFE"]->id, $sesArray);
// #54634, 131128, dwildt, 1+
		$GLOBALS[ 'TSFE' ]->fe_user->setKey( 'ses', $this->extKey . '_' . $pid, $sesArray );
		$GLOBALS[ 'TSFE' ]->storeSessionData(); // Save session
	}

	/**
	 * specialGet( )  : get the special method from session
	 *
	 * @param	integer		: $pid  : uid of the page, which contains the caddy plugin
	 * @return	integer
	 * @version     3.0.1
	 * @since       1.4.0
	 */
	public function specialGet( $pid = null )
	{
		$pid = $this->getPid( $pid );

// get products from session
// #54634, 131128, dwildt, 1-
//$sesArray = $GLOBALS['TSFE']->fe_user->getKey('ses', $this->extKey . '_' . $GLOBALS["TSFE"]->id);
// #54634, 131128, dwildt, 1+
		$sesArray = $GLOBALS[ 'TSFE' ]->fe_user->getKey( 'ses', $this->extKey . '_' . $pid );

		return $sesArray[ 'specials' ];
	}

	/*	 * *********************************************
	 *
	 * Variants
	 *
	 * ******************************************** */

	/**
	 * variantsInit( )  :
	 *
	 * @param	array		$newProduct :
	 * @return	array		$variants :
	 * @access private
	 * @internal     #t0915
	 * @version     7.2.0
	 * @since       2.0.11
	 */
	private function variantsInit( $newProductUid )
	{
		$variants = NULL;

		switch( $this->getVariantsStyle() )
		{
			case('caddy'):
				$variants = $this->variantsInitCaddy( $newProductUid );
				break;
			case('wt_cart'):
				$variants = $this->variantsInitWtcart( $newProductUid );
				break;
			case('novariants'):
			default:
				// Do nothing
				break;
		}

		return $variants;
	}

	/**
	 * variantsInitCaddy( )  :
	 *
	 * @param	array		$newProduct :
	 * @return	array		$variants :
	 * @access private
	 * @internal     #t0915
	 * @version     7.2.0
	 * @since       2.0.11
	 */
	private function variantsInitCaddy( $newProductUid )
	{

// POST has precedence!
		$GP = \TYPO3\CMS\Core\Utility\GeneralUtility::_POST() + \TYPO3\CMS\Core\Utility\GeneralUtility::_GET();

		$table = $this->pObj->conf[ 'constant_editor.' ][ 'variants.' ][ 'tables.' ][ 'property' ];
		$field = 'uid';
		$tableField = $table . '.' . $field;
		$uids = array();
//var_dump( __METHOD__, __LINE__, $GP, $table, $field );

		foreach( ( array ) $GP[ $table ][ $field ] as $uid )
		{
			$uids[] = ( int ) $uid;
		}
		$csvUids = implode( ',', $uids );

		$variants = array
				(
				'uid' => $newProductUid,
				$tableField => $csvUids
		);

//    var_dump( __METHOD__, __LINE__, $variants );
////    array(2)
////    {
////      [ "uid" ] =>
////      string(1) "6"
////      [ "tx_quickshop_variantproperty.uid" ] =>
////      string(3) "3,6"
////    }
//    die();

		return $variants;
	}

	/**
	 * variantsInitWtcart( )  :
	 *
	 * @param	array		$newProduct :
	 * @return	array		$variants :
	 * @access private
	 * @internal     #t0915
	 * @version     7.2.0
	 * @since       2.0.11
	 */
	private function variantsInitWtcart( $newProductUid )
	{
		$variants = array
				(
				'uid' => $newProductUid
		);

// POST has precedence!
		$GP = \TYPO3\CMS\Core\Utility\GeneralUtility::_POST() + \TYPO3\CMS\Core\Utility\GeneralUtility::_GET();

// FOREACH variant
		foreach( ( array ) $this->pObj->conf[ 'settings.' ][ 'variant.' ] as $key => $tableField )
		{
			if( stristr( $key, '.' ) )
			{
				continue;
			}

// DRS
			if( $this->drs->drsVariants )
			{
				$prompt = 'settings.variant. ' . $key . ' = ' . $tableField;
				\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[INFO/VARIANTS] ' . $prompt, $this->extKey, 0 );
			}
// DRS

			list( $table, $field ) = explode( '.', $tableField );
			if( !isset( $GP[ $table ][ $field ] ) )
			{
				// DRS
				if( $this->drs->drsVariants )
				{
					$prompt = 'GET/POST[' . $table . '][' . $field . '] is not set.';
					\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[WARN/VARIANTS] ' . $prompt, $this->extKey, 2 );
				}
				// DRS
				continue;
			}

			$variant = $this->zz_escapeStrForLike( $GP[ $table ][ $field ], $table );

// DRS
			if( $this->drs->drsVariants )
			{
				$prompt = 'Variant: GET/POST[' . $table . '][' . $field . '] is set.';
				\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[OK/VARIANTS] ' . $prompt, $this->extKey, -1 );
			}
// DRS

			$variants[ $tableField ] = $variant;
		}
// FOREACH variant
//var_dump(__METHOD__, __LINE__, $variants);
//    array {
//      "uid" => "2",
//      "tx_org_calentrance.uid" => "3"
//    }
//die();
		return $variants;
	}

	/**
	 * variantsMatched( )
	 *
	 * @param	array		$variants :
	 * @param	array		$product  :
	 * @return	integer		$matchedVariants
	 * @access private
	 * @version     2.0.11
	 * @since       2.0.11
	 */
	private function variantsMatched( $variants, $product )
	{
		$matchedVariants = 0;

// FOREACH variant
		foreach( ( array ) $variants as $tableField => $variant )
		{
// variant fits
			if( $product[ $tableField ] == $variant )
			{
				$matchedVariants++;
			}
		}

		if( !$this->drs->drsVariants )
		{
			return $matchedVariants;
		}

// DRS
		switch( $matchedVariants )
		{
			case( 0 ):
				$prompt = 'Any variant doesn\'t match!';
				break;
			case( 1 ):
				$prompt = '#1 variant matchs!';
				break;
			default:
				$prompt = '#' . $matchedVariants . ' variants matchs!';
				break;
		}
		\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[INFO/VARIANTS] ' . $prompt, $this->extKey, 0 );

		return $matchedVariants;
	}

	/**
	 * variantsMatched( )
	 *
	 * @param	array		$variants :
	 * @param	array		$product  :
	 * @return	integer		$matchedVariants
	 * @access private
	 * @version     2.0.11
	 * @since       2.0.11
	 */
	private function variantsSet( $variants, $product )
	{
		if( empty( $variants ) )
		{
			return $product;
		}

// remove uid from variant array
		unset( $variants[ 'uid' ] );

// add variant
		foreach( ( array ) $variants as $tableField => $variant )
		{
			$product[ $tableField ] = $variant;
		}
		return $product;
	}

	/*	 * *********************************************
	 *
	 * ZZ
	 *
	 * ******************************************** */

	/**
	 * zz_commaToDot( ) :
	 *
	 * @param	string		$strWiComma :
	 * @return	string		$strWiDot   :
	 * @access private
	 * @version 2.0.0
	 * @since 1.4.6
	 */
	private function zz_commaToDot( $strWiComma )
	{
		$strWiDot = str_replace( ',', '.', $strWiComma );

		return $strWiDot;
	}

	/**
	 * returns message with optical flair
	 *
	 * @param	string		$str: Message to show
	 * @param	int		$pos: Is this a positive message? (0,1,2)
	 * @param	boolean		$die: Process should be died now
	 * @param	boolean		$prefix: Activate or deactivate prefix "$extKey: "
	 * @param	string		$id: id to add to the message (maybe to do some javascript effects)
	 * @return	string		$string: Manipulated string
	 * @access private
	 * @version 2.0.0
	 * @since 1.4.6
	 */
	private function zz_msg( $str, $pos = 0, $die = 0, $prefix = 1, $id = '' )
	{
// config
		if( $prefix )
		{
			$string = $this->extKey . ($pos != 1 && $pos != 2 ? ' Error' : '') . ': ';
		}
		$string .= $str; // add string
		$URLprefix = \TYPO3\CMS\Core\Utility\GeneralUtility::getIndpEnv( 'TYPO3_SITE_URL' ) . '/'; // URLprefix with domain
		if( \TYPO3\CMS\Core\Utility\GeneralUtility::getIndpEnv( 'TYPO3_REQUEST_HOST' ) . '/' != \TYPO3\CMS\Core\Utility\GeneralUtility::getIndpEnv( 'TYPO3_SITE_URL' ) )
		{ // if request_host is different to site_url (TYPO3 runs in a subfolder)
			$URLprefix .= str_replace( \TYPO3\CMS\Core\Utility\GeneralUtility::getIndpEnv( 'TYPO3_REQUEST_HOST' ) . '/', '', \TYPO3\CMS\Core\Utility\GeneralUtility::getIndpEnv( 'TYPO3_SITE_URL' ) ); // add folder (like "subfolder/")
		}

// let's go
		switch( $pos )
		{
			default: // error
				$wrap = '<div class="' . $this->extKey . '_msg_error" style="background-color: #FBB19B; background-position: 4px 4px; background-image: url(' . $URLprefix . 'typo3/gfx/error.png); background-repeat: no-repeat; padding: 5px 30px; font-weight: bold; border: 1px solid #DC4C42; margin-bottom: 20px; font-family: arial, verdana; color: #444; font-size: 12px;"';
				if( $id )
				{
					$wrap .= ' id="' . $id . '"'; // add css id
				}
				$wrap .= '>';
				break;

			case 1: // success
				$wrap = '<div class="' . $this->extKey . '_msg_status" style="background-color: #CDEACA; background-position: 4px 4px; background-image: url(' . $URLprefix . 'typo3/gfx/ok.png); background-repeat: no-repeat; padding: 5px 30px; font-weight: bold; border: 1px solid #58B548; margin-bottom: 20px; font-family: arial, verdana; color: #444; font-size: 12px;"';
				if( $id )
				{
					$wrap .= ' id="' . $id . '"'; // add css id
				}
				$wrap .= '>';
				break;

			case 2: // note
				$wrap = '<div class="' . $this->extKey . '_msg_error" style="background-color: #DDEEF9; background-position: 4px 4px; background-image: url(' . $URLprefix . 'typo3/gfx/information.png); background-repeat: no-repeat; padding: 5px 30px; font-weight: bold; border: 1px solid #8AAFC4; margin-bottom: 20px; font-family: arial, verdana; color: #444; font-size: 12px;"';
				if( $id )
				{
					$wrap .= ' id="' . $id . '"'; // add css id
				}
				$wrap .= '>';
				break;
		}

		if( !$die )
		{
			return $wrap . $string . '</div>'; // return message
		}
		else
		{
			die( $string ); // die process and write message
		}
	}

	/**
	 * zz_escapeStrForLike( )  :
	 *
	 * @param	string		$string :
	 * @param	string		$table :
	 * @return	string		$string :
	 * @access private
	 * @internal    #61632
	 * @version     6.0.0
	 * @since       2.0.11
	 */
	private function zz_escapeStrForLike( $string, $table )
	{
// #61632, 140916, dwildt, 1-
//$mysqlEscapeString = mysql_escape_string( $string );
// #61870, 140925, dwildt, 2-
//    // #61632, 140916, dwildt, 2+
//    $string = $GLOBALS['TYPO3_DB']->escapeStrForLike( $string, $table ) . "'";
// #61870, 140925, dwildt, 2+
// #61632, 140916, dwildt, 1+
		$string = $GLOBALS[ 'TYPO3_DB' ]->escapeStrForLike( $string, $table );
		return $string;

// #61632, 140916, dwildt, -
//    if ( $mysqlEscapeString )
//    {
//      return $mysqlEscapeString;
//    }
//
//    // DRS
//    if ( $this->drs->drsWarn )
//    {
//      $prompt = 'mysql_escape_string( ) returns false! Unproper connection to database!';
//      \TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[WARN/PHP] ' . $prompt, $this->extKey, 2 );
//      $prompt = '", \' and \\ will replaced manually. Maybe this is a security risk!';
//      \TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[WARN/PHP] ' . $prompt, $this->extKey, 2 );
//    }
//    // DRS
//
//    $string = str_replace( '"', null, $string );
//    $string = str_replace( "'", null, $string );
//    $string = str_replace( '\\', null, $string );
//
//    return $string;
	}

	/**
	 * zz_sqlReplaceMarker(): Replace marker in the SQL query
	 *                             MARKERS are
	 *                             - GET/POST markers
	 *                             - enable_field markers
	 *                             SYNTAX is
	 *                             - ###GP:TABLE###
	 *                             - ###GP:TABLE.FIELD###
	 *                             - ###ENABLE_FIELD:TABLE.FIELD###
	 *
	 * @return	void
	 * @version 6.0.0
	 * @since 1.2.2
	 */
	private function zz_sqlReplaceMarker()
	{
		$arr_result = null;

// set marker array with values from GET
		foreach( \TYPO3\CMS\Core\Utility\GeneralUtility::_GET() as $table => $arr_fields )
		{
			if( is_array( $arr_fields ) )
			{
				foreach( ( array ) $arr_fields as $field => $value )
				{
					$tableField = strtoupper( $table . '.' . $field );
//            $marker['###GP:' . strtoupper($tableField) . '###'] = $this->zz_escapeStrForLike($value);
				}
			}
			if( !is_array( $arr_fields ) )
			{
				// #61632, 140916, dwildt: +$table
				$marker[ '###GP:' . strtoupper( $table ) . '###' ] = $this->zz_escapeStrForLike( $arr_fields, $table );
			}
		}

// set and overwrite marker array with values from POST
		foreach( \TYPO3\CMS\Core\Utility\GeneralUtility::_POST() as $table => $arr_fields )
		{
			if( is_array( $arr_fields ) )
			{
				foreach( ( array ) $arr_fields as $field => $value )
				{
					$tableField = strtoupper( $table . '.' . $field );
					// #61632, 140916, dwildt: +$table
					$marker[ '###GP:' . strtoupper( $tableField ) . '###' ] = $this->zz_escapeStrForLike( $value, $table );
				}
			}
			if( !is_array( $arr_fields ) )
			{
				// #61632, 140916, dwildt: +$table
				$marker[ '###GP:' . strtoupper( $table ) . '###' ] = $this->zz_escapeStrForLike( $arr_fields, $table );
			}
		}

// get the SQL query from ts, allow stdWrap
// #42154, 101218, dwildt, 1-
//$query = $pObj->cObj->stdWrap($pObj->conf['db.']['sql'], $pObj->conf['db.']['sql.']);
// #42154, 101218, dwildt, 1+
		$query = $this->pObj->cObj->cObjGetSingle( $this->pObj->conf[ 'db.' ][ 'sql' ], $this->pObj->conf[ 'db.' ][ 'sql.' ] );

// get all gp:marker out of the query
		$arr_gpMarker = array();
		preg_match_all( '|###GP\:(.*)###|U', $query, $arr_result, PREG_PATTERN_ORDER );
		if( isset( $arr_result[ 0 ] ) )
		{
			$arr_gpMarker = $arr_result[ 0 ];
		}

// get all enable_fields:marker out of the query
		$arr_efMarker = array();
		preg_match_all( '|###ENABLE_FIELDS\:(.*)###|U', $query, $arr_result, PREG_PATTERN_ORDER );
		if( isset( $arr_result[ 0 ] ) )
		{
			$arr_efMarker = $arr_result[ 0 ];
		}

// replace gp:marker
		foreach( ( array ) $arr_gpMarker as $str_gpMarker )
		{
			$value = null;
			if( isset( $marker[ $str_gpMarker ] ) )
			{
				$value = $marker[ $str_gpMarker ];
			}
			$query = str_replace( $str_gpMarker, $value, $query );
		}

// #i0105, 160702, dwildt, 1+
		$cObj = new \TYPO3\CMS\Frontend\ContentObject\ContentObjectRenderer;

// replace enable_fields:marker
		foreach( ( array ) $arr_efMarker as $str_efMarker )
		{
			$str_efTable = trim( strtolower( $str_efMarker ), '#' );
			list( $dummy, $str_efTable ) = explode( ':', $str_efTable );
			unset( $dummy );
// #i0105, 160702, dwildt, 1-/+
//$andWhere_ef = tslib_cObj::enableFields( $str_efTable );
			$andWhere_ef = $cObj->enableFields( $str_efTable );
			$query = str_replace( $str_efMarker, $andWhere_ef, $query );
		}

//var_dump(__METHOD__, __LINE__, $query);		
// #42154, 121203, dwildt, 1-
//    $pObj->conf['db.']['sql'] = $query;
// #61889, 140925, dwildt, 2+
		unset( $this->pObj->conf[ 'db.' ][ 'sql' ] );
		unset( $this->pObj->conf[ 'db.' ][ 'sql.' ] );
// #42154, 121203, dwildt, 2+
		$this->pObj->conf[ 'db.' ][ 'sql' ] = 'TEXT';
		$this->pObj->conf[ 'db.' ][ 'sql.' ][ 'value' ] = $query;
	}

	/**
	 * zz_sqlDRS(): 
	 * 
	 * @return	void
	 * @version 7.2.0
	 * @since 7.2.0
	 */
	private function zz_sqlDRS( $query, $prompt = NULL, $method = __METHOD__, $line = __LINE__, $boolPromptGP = false )
	{
		$error = $GLOBALS[ 'TYPO3_DB' ]->sql_error();
		if( !$error )
		{
			if( !$this->drs->drsSql )
			{
				return;
			}
			$prompt = $query;
			\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[OK/SQL] ' . $prompt, $this->extKey, -1 );
			return;
		}

		if( $this->drs->drsError )
		{
			$prompt = $query;
			\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[ERROR/SQL] ' . $prompt, $this->extKey, 3 );
			$prompt = $error;
			\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[ERROR/SQL] ' . $prompt, $this->extKey, 2 );
		}

		if( $boolPromptGP )
		{
			$GP = \TYPO3\CMS\Core\Utility\GeneralUtility::_GET() + \TYPO3\CMS\Core\Utility\GeneralUtility::_POST();
			$GPprompt = ''
							. '<pre>' . PHP_EOL
							. var_export( $GP, true ) . PHP_EOL
							. '</pre>' . PHP_EOL
			;
		}

// DRS
		if( $prompt )
		{
			$prompt = '<h2>' . $prompt . '</h2>' . PHP_EOL;
		}
		$prompt = '<h1>caddy: SQL-Error</h1>' . PHP_EOL
						. '<p>' . $error . '</p>' . PHP_EOL
						. '<p>' . $query . '</p>' . PHP_EOL
						. '<h2>' . $query . '</h2>' . PHP_EOL
						. $GPprompt
						. '<p>' . PHP_EOL
						. 'Please take care for a proper configuration<br />' . PHP_EOL
						. 'Maybe there is a problem with plugin.tx_caddy_pi1.db.sql<br />' . PHP_EOL
						. 'Sorry for the trouble.<br />' . PHP_EOL
						. 'TYPO3 Caddy<br />' . PHP_EOL
						. $method . ' (' . $line . ')'
						. '</p>' . PHP_EOL
		;
		die( $prompt );
	}

}
