<?php

namespace Netzmacher\Caddy\Plugins\pi1;

use TYPO3\CMS\Core\Utility\GeneralUtility;

/* * *************************************************************
 *  Copyright notice
 *
 *  (c) 2013-2020 - Dirk Wildt <http://wildt.at.die-netzmacher.de>
 *  All rights reserved
 *
 *  This script is part of the TYPO3 project. The TYPO3 project is
 *  free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  The GNU General Public License can be found at
 *  http://www.gnu.org/copyleft/gpl.html.
 *
 *  This script is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  This copyright notice MUST APPEAR in all copies of the script!
 * ************************************************************* */

/**
 * plugin 'Cart to powermail' for the 'caddy' extension.
 *
 * @author	Dirk Wildt <http://wildt.at.die-netzmacher.de>
 * @package	TYPO3
 * @subpackage	tx_caddy
 * @version		9.1.4
 * @since     1.4.6
 */
class Clean
{

	public $prefixId = 'tx_caddy_pi1';
	// same as class name
	public $scriptRelPath = 'Resources/Private/Language/Plugins/pi1/Clean.php';
	// path to this script relative to the extension dir.
	public $extKey = 'caddy';
	// Parent object
	public $pObj = null;
	// Current row
	public $row = null;
	private $pidCaddy = null;
	private $local_cObj = null;

	/**
	 * main( )
	 *
	 * @return	void
	 * @access public
	 * @version    2.0.0
	 * @since      2.0.0
	 */
	public function main()
	{
		$this->_init( $this->pidCaddy );

		// RETURN : powermail form isn't sent. Nothing to clean
		if( empty( $this->pObj->powermail->getSent() ) )
		{
			// DRS
			if( $this->pObj->drs->drsClean )
			{
				$prompt = 'The powermail form isn\'t sent, nothing to clean up.';
				\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[INFO/CLEAN] ' . $prompt, $this->pObj->extKey, 0 );
			}
			// DRS
			return;
		}
		// RETURN : powermail form isn't sent. Nothing to clean
		// DRS
		if( $this->pObj->drs->drsClean )
		{
			$prompt = 'The powermail form is sent. Database, numbers and session will cleaned up.';
			\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[INFO/CLEAN] ' . $prompt, $this->pObj->extKey, 0 );
		}
		// DRS
		// #55726, 140207, dwildt, 1+
		$this->_stock();
		$this->_sessionUpdate();
		$this->_database();
	}

	/*	 * *********************************************
	 *
	 * Database
	 *
	 * ******************************************** */

	/**
	 * _database( )
	 *
	 * @return	void
	 * @access private
	 * @version    7.3.0
	 * @since      2.0.0
	 */
	private function _database()
	{
		if( $this->_databaseCancelWoItems() )
		{
			return;
		}

		if( $this->_databaseCancelWiEpaymentError() )
		{
			return;
		}

		$this->_databaseInitCObj();

		$orderUid = $this->_sqlInsertOrder();
		$this->_sqlInsertItems( $orderUid );
	}

	/**
	 * _databaseCancelWiEpaymentError( )
	 *
	 * @return	boolen
	 * @access private
	 * @version    4.0.5
	 * @since      4.0.5
	 */
	private function _databaseCancelWiEpaymentError()
	{
		$sesArray = $GLOBALS[ 'TSFE' ]->fe_user->getKey( 'ses', $this->extKey . '_' . $this->pidCaddy );
		// RETURN : e-payment transaction was successful
		if( !$sesArray[ 'e-payment' ][ 'powermail' ][ 'error' ] )
		{
			return false;
		}

		unset( $_GET[ 'tx_powermail_pi1' ] );
		unset( $_POST[ 'tx_powermail_pi1' ] );

		$prompt = 'E-payment error is set by powermail';
		\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[ERROR/SESSION] ' . $prompt, $this->extKey, 3 );
		$prompt = 'There won\'t added any new record to the database.';
		\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[WARN/SESSION] ' . $prompt, $this->extKey, 2 );
		$prompt = 'GET/POST tx_powermail_pi1 will removed! This prevents another e-mail.';
		\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[WARN/SESSION] ' . $prompt, $this->extKey, 2 );

		return true;
	}

	/**
	 * _databaseCancelWoItems( )
	 *
	 * @return	boolen
	 * @access private
	 * @version    4.0.5
	 * @since      2.0.0
	 */
	private function _databaseCancelWoItems()
	{
		$sesArray = $GLOBALS[ 'TSFE' ]->fe_user->getKey( 'ses', $this->extKey . '_' . $this->pidCaddy );

		// RETURN : any product, don't increase numbers!
		if( !empty( $sesArray[ 'products' ] ) )
		{
			return false;
		}

		unset( $_GET[ 'tx_powermail_pi1' ] );
		unset( $_POST[ 'tx_powermail_pi1' ] );

		if( !$this->pObj->drs->drsError )
		{
			return true;
		}

		$prompt = 'There isn\'t any product! Maybe powermail form is sent twice!';
		\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[ERROR/SESSION] ' . $prompt, $this->extKey, 3 );
		$prompt = 'There won\'t added any new record to the database.';
		\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[WARN/SESSION] ' . $prompt, $this->extKey, 2 );
		$prompt = 'GET/POST tx_powermail_pi1 will removed! This prevents another e-mail.';
		\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[WARN/SESSION] ' . $prompt, $this->extKey, 2 );

		return true;
	}

	/**
	 * _databaseFieldsFiles( )
	 *
	 * @param	array		$sesArray :
	 * @return	void
	 * @access private
	 * @version    2.0.2
	 * @since      2.0.2
	 */
	private function _databaseFieldsFiles( $sesArray )
	{
		// Initiate files
		$fileDeliveryorder = null;
		$fileInvoice = null;
		#i0109, 160930, dwildt, 1+
		$fileOrderconfirmation = null;
		$fileRevocation = null;
		$fileTerms = null;
		// Initiate files
		// Get pdf is sent to ...
		$pdfDeliveryorderToCustomer = false;
		$pdfDeliveryorderToVendor = false;
		$pdfInvoiceToCustomer = false;
		$pdfInvoiceToVendor = false;
		#i0109, 160930, dwildt, 2+
		$pdfOrderconfirmationToCustomer = false;
		$pdfOrderconfirmationToVendor = false;
		$pdfRevocationToCustomer = false;
		$pdfRevocationToVendor = false;
		$pdfTermsToCustomer = false;
		$pdfTermsToVendor = false;

		if( !empty( $sesArray[ 'sendCustomerDeliveryorder' ] ) )
		{
			$pdfDeliveryorderToCustomer = true;
		}
		if( !empty( $sesArray[ 'sendVendorDeliveryorder' ] ) )
		{
			$pdfDeliveryorderToVendor = true;
		}
		if( !empty( $sesArray[ 'sendCustomerInvoice' ] ) )
		{
			$pdfInvoiceToCustomer = true;
		}
		if( !empty( $sesArray[ 'sendVendorInvoice' ] ) )
		{
			$pdfInvoiceToVendor = true;
		}
		#i0109, 160930, dwildt, 4+
		if( !empty( $sesArray[ 'sendCustomerOrderconfirmation' ] ) )
		{
			$pdfOrderconfirmationToCustomer = true;
		}
		if( !empty( $sesArray[ 'sendVendorOrderconfirmation' ] ) )
		{
			$pdfOrderconfirmationToVendor = true;
		}
		if( !empty( $sesArray[ 'sendCustomerRevocation' ] ) )
		{
			$pdfRevocationToCustomer = true;
		}
		if( !empty( $sesArray[ 'sendVendorRevocation' ] ) )
		{
			$pdfRevocationToVendor = true;
		}
		if( !empty( $sesArray[ 'sendCustomerTerms' ] ) )
		{
			$pdfTermsToCustomer = true;
		}
		if( !empty( $sesArray[ 'sendVendorTerms' ] ) )
		{
			$pdfTermsToVendor = true;
		}
		// Get pdf is sent to ...
		// Set files
		switch( true )
		{
			case( $pdfDeliveryorderToCustomer ):
			case( $pdfDeliveryorderToVendor ):
				$fileDeliveryorder = $this->local_cObj->cObjGetSingle
								(
								$this->pObj->conf[ 'pdf.' ][ 'deliveryorder.' ][ 'filename' ], $this->pObj->conf[ 'pdf.' ][ 'deliveryorder.' ][ 'filename.' ]
				);
				break;
		}
		switch( true )
		{
			case( $pdfInvoiceToCustomer ):
			case( $pdfInvoiceToVendor ):
				$fileInvoice = $this->local_cObj->cObjGetSingle
								(
								$this->pObj->conf[ 'pdf.' ][ 'invoice.' ][ 'filename' ], $this->pObj->conf[ 'pdf.' ][ 'invoice.' ][ 'filename.' ]
				);
				break;
		}
		#i0109, 160930, dwildt, 11+
		switch( true )
		{
			case( $pdfOrderconfirmationToCustomer ):
			case( $pdfOrderconfirmationToVendor ):
				$fileOrderconfirmation = $this->local_cObj->cObjGetSingle
								(
								$this->pObj->conf[ 'pdf.' ][ 'orderconfirmation.' ][ 'filename' ], $this->pObj->conf[ 'pdf.' ][ 'orderconfirmation.' ][ 'filename.' ]
				);
				break;
		}
		switch( true )
		{
			case( $pdfRevocationToCustomer ):
			case( $pdfRevocationToVendor ):
				$fileRevocation = $this->local_cObj->cObjGetSingle
								(
								$this->pObj->conf[ 'pdf.' ][ 'revocation.' ][ 'filename' ], $this->pObj->conf[ 'pdf.' ][ 'revocation.' ][ 'filename.' ]
				);
				break;
		}
		switch( true )
		{
			case( $pdfTermsToCustomer ):
			case( $pdfTermsToVendor ):
				$fileTerms = $this->local_cObj->cObjGetSingle
								(
								$this->pObj->conf[ 'pdf.' ][ 'terms.' ][ 'filename' ], $this->pObj->conf[ 'pdf.' ][ 'terms.' ][ 'filename.' ]
				);
				break;
		}

		$record = array
				(
				'fileDeliveryorder' => $fileDeliveryorder,
				'fileInvoice' => $fileInvoice,
				#i0109, 160930, dwildt, 1+
				'fileOrderconfirmation' => $fileOrderconfirmation,
				'fileRevocation' => $fileRevocation,
				'fileTerms' => $fileTerms,
				'pdfDeliveryorderToCustomer' => $pdfDeliveryorderToCustomer,
				'pdfDeliveryorderToVendor' => $pdfDeliveryorderToVendor,
				'pdfInvoiceToCustomer' => $pdfInvoiceToCustomer,
				'pdfInvoiceToVendor' => $pdfInvoiceToVendor,
				#i0109, 160930, dwildt, 2+
				'pdfOrderconfirmationToCustomer' => $pdfOrderconfirmationToCustomer,
				'pdfOrderconfirmationToVendor' => $pdfOrderconfirmationToVendor,
				'pdfRevocationToCustomer' => $pdfRevocationToCustomer,
				'pdfRevocationToVendor' => $pdfRevocationToVendor,
				'pdfTermsToCustomer' => $pdfTermsToCustomer,
				'pdfTermsToVendor' => $pdfTermsToVendor,
		);
		// Set record

		return $record;
	}

	/**
	 * databaseFieldsNumbers( )
	 *
	 * @param	array		$sesArray :
	 * @return	void
	 * @access private
	 * @version    2.0.2
	 * @since      2.0.2
	 */
	private function _databaseFieldsNumbers( $sesArray )
	{
		$record = array
				(
				'numberDeliveryorder' => $sesArray[ 'numberDeliveryorderCurrent' ],
				'numberInvoice' => $sesArray[ 'numberInvoiceCurrent' ],
				'numberOrder' => $sesArray[ 'numberOrderCurrent' ],
		);
		// Set record

		return $record;
	}

	/**
	 * _databaseFieldsSum( )
	 *
	 * @param	array		$sesArray :
	 * @return	void
	 * @access private
	 * @version    8.15.1
	 * @since      2.0.2
	 */
	private function _databaseFieldsSum( $sesArray )
	{
		$record = array
				(
				'sumGross' => $sesArray[ 'sum' ][ 'sum' ][ 'gross' ],
				'sumNet' => $sesArray[ 'sum' ][ 'sum' ][ 'net' ],
				'sumTaxNormal' => $sesArray[ 'sum' ][ 'sum' ][ 'tax' ][ 'normal' ],
				'sumTaxReduced' => $sesArray[ 'sum' ][ 'sum' ][ 'tax' ][ 'reduced' ],
				// #i0281, 191218, dwildt, 1+
				'sumWeight' => $sesArray[ 'sum' ][ 'sum' ][ 'weight' ],
		);
		// Set record

		return $record;
	}

	/**
	 * _databaseInitCObj( )
	 *
	 * @return	void
	 * @access private
	 * @version    7.3.0
	 * @since      2.0.0
	 */
	private function _databaseInitCObj()
	{
		$sesArray = null;
		$sesArray = $GLOBALS[ 'TSFE' ]->fe_user->getKey( 'ses', $this->extKey . '_' . $this->pidCaddy );

		$this->local_cObj->start( $sesArray, $this->pObj->conf[ 'db.' ][ 'table' ] );

		if( !$this->pObj->drs->drsCobj )
		{
			return;
		}

		$data = var_export( $this->local_cObj->data, true );
		$prompt = 'cObj->data: ' . $data;
		\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[INFO/COBJ] ' . $prompt, $this->extKey, 0 );
	}

	/*	 * *********************************************
	 *
	 * Getting methods
	 *
	 * ******************************************** */

	/**
	 * getCObj( ) : Get the content object
	 *
	 * @return	object    $cObj : content object
	 * @access public
	 * @internal   #55726
	 * @version    4.0.7
	 * @since      4.0.7
	 */
	public function getCObj()
	{
		return $this->cObj;
	}

	/**
	 * getConf( ) : Get the TypoScript configuration array
	 *
	 * @return	array   $conf : TypoScript configuration
	 * @access public
	 * @internal   #55726
	 * @version    4.0.7
	 * @since      4.0.7
	 */
	public function getConf()
	{
		return $this->conf;
	}

	/**
	 * getDrs( ) : Get the DRS object
	 *
	 * @return	object   $drs : DRS object
	 * @access public
	 * @internal   #55726
	 * @version    4.0.7
	 * @since      4.0.7
	 */
	public function getDrs()
	{
		return $this->drs;
	}

	/**
	 * getPid( )  : Returns the globlas tsfe id, if the given pid is null
	 *
	 * @param	integer		$pid  : given pid (may be null)
	 * @return	integer		$pid  : id of the page with the caddy plugin
	 * @internal    #54634
	 * @access      public
	 * @version     4.0.7
	 * @since       4.0.7
	 */
	public function getPid( $pid = null )
	{
		$this->initPidCaddy( $pid );
		return $this->pidCaddy;
	}

	/**
	 * _createfeuser( ) : Returns 1 (true), if powermail form contains a field createfeuser with the value 1
	 *
	 * @return	integer		$value  : the value
	 * @access private
	 * @internal   #i0321
	 * @version    9.1.4
	 * @since      9.1.4
	 */
	private function _createfeuser()
	{
		if( ( int ) $GLOBALS[ 'TSFE' ]->fe_user->user[ 'uid' ] > 0 )
		{
			// User is logged in. Don't create a new user!
			return false;
		}

		$tx_powermail_pi1 = GeneralUtility::_GP( 'tx_powermail_pi1' );

		$key = $this->pObj->conf[ 'constant_editor.' ][ 'getpost.' ][ 'createfeuser' ];
		$value = ( int ) $tx_powermail_pi1[ 'field' ][ $key ];

		//var_dump( __METHOD__, __LINE__, $_GET, $_POST, $value );
		return $value;
	}

	/**
	 * _getFeuserNewId( )
	 *
	 * @return	integer
	 * @access private
	 * @internal   #i0321
	 * @version    9.1.4
	 * @since      9.1.4
	 */
	private function _getFeuserNewId()
	{

		$maxUid = \Netzmacher\Caddy\Userfunc\SQL::SelectFeuserMaxUid();
		return $maxUid + 1;
	}

	/**
	 * _getFeuserUid( )
	 *
	 * @return	integer
	 * @access private
	 * @internal   #i0317
	 * @version    9.1.4
	 * @since      9.1.1
	 */
	private function _getFeuserUid()
	{
		if( $this->_createfeuser() )
		{
			return $this->_getFeuserNewId();
		}

		return ( int ) $GLOBALS[ 'TSFE' ]->fe_user->user[ 'uid' ];
	}

	/*	 * *********************************************
	 *
	 * Get powermail fields
	 *
	 * ******************************************** */

	/**
	 * _getPmFieldDebitBic( ) : Get the debit BIC from powermail POST params
	 *
	 * @return	string		$value  : the value
	 * @access private
	 * @internal   #i0142
	 * @version    7.3.2
	 * @since      7.3.2
	 */
	private function _getPmFieldDebitBic()
	{
		$pmUid = $this->pObj->flexform->debitBic;
		$value = $this->pObj->powermail->getFieldById( $pmUid );
		return $value;
	}

	/**
	 * _getPmFieldDebitIban( ) : Get the debit IBAN from powermail POST params
	 *
	 * @return	string		$value  : the value
	 * @access private
	 * @internal   #i0142
	 * @version    7.3.2
	 * @since      7.3.2
	 */
	private function _getPmFieldDebitIban()
	{
		$pmUid = $this->pObj->flexform->debitIban;
		$value = $this->pObj->powermail->getFieldById( $pmUid );
		return $value;
	}

	/**
	 * _getPmFieldDebitOwner( ) : Get the debit owner from powermail POST params
	 *
	 * @return	string		$value  : the value
	 * @access private
	 * @internal   #i0142
	 * @version    7.3.2
	 * @since      7.3.2
	 */
	private function _getPmFieldDebitOwner()
	{
		$pmUid = $this->pObj->flexform->debitOwner;
		$value = $this->pObj->powermail->getFieldById( $pmUid );
		return $value;
	}

	/**
	 * getDeliveryorderAddress( )  : Get the delivery order Address from powermail POST params
	 *
	 * @return	string		$value  : the value
	 * @access private
	 * @version    2.0.0
	 * @since      2.0.0
	 */
	private function _getPmFieldDeliveryorderAddress()
	{
		$pmUid = $this->pObj->flexform->deliveryorderAddress;
		$value = $this->pObj->powermail->getFieldById( $pmUid );
		return $value;
	}

	/**
	 * getDeliveryorderCity( )  : Get the delivery order City from powermail POST params
	 *
	 * @return	string		$value  : the value
	 * @access private
	 * @version    2.0.0
	 * @since      2.0.0
	 */
	private function _getPmFieldDeliveryorderCity()
	{
		$pmUid = $this->pObj->flexform->deliveryorderCity;
		$value = $this->pObj->powermail->getFieldById( $pmUid );
		return $value;
	}

	/**
	 * getDeliveryorderCompany( )  : Get the delivery order Company from powermail POST params
	 *
	 * @return	string		$value  : the value
	 * @access private
	 * @version    2.0.0
	 * @since      2.0.0
	 */
	private function _getPmFieldDeliveryorderCompany()
	{
		$pmUid = $this->pObj->flexform->deliveryorderCompany;
		$value = $this->pObj->powermail->getFieldById( $pmUid );
		return $value;
	}

	/**
	 * getDeliveryorderCountry( )  : Get the delivery order Country from powermail POST params
	 *
	 * @return	string		$value  : the value
	 * @access private
	 * @version    2.0.0
	 * @since      2.0.0
	 */
	private function _getPmFieldDeliveryorderCountry()
	{
		$pmUid = $this->pObj->flexform->deliveryorderCountry;
		$value = $this->pObj->powermail->getFieldById( $pmUid );
		return $value;
	}

	/**
	 * getDeliveryorderFirstname( )  : Get the delivery order Firstname from powermail POST params
	 *
	 * @return	string		$value  : the value
	 * @access private
	 * @version    2.0.0
	 * @since      2.0.0
	 */
	private function _getPmFieldDeliveryorderFirstname()
	{
		$pmUid = $this->pObj->flexform->deliveryorderFirstname;
		$value = $this->pObj->powermail->getFieldById( $pmUid );
		return $value;
	}

	/**
	 * getDeliveryorderLastname( )  : Get the delivery order Lastname from powermail POST params
	 *
	 * @return	string		$value  : the value
	 * @access private
	 * @version    2.0.0
	 * @since      2.0.0
	 */
	private function _getPmFieldDeliveryorderLastname()
	{
		$pmUid = $this->pObj->flexform->deliveryorderLastname;
		$value = $this->pObj->powermail->getFieldById( $pmUid );
		return $value;
	}

	/**
	 * getDeliveryorderZip( )  : Get the delivery order Zip from powermail POST params
	 *
	 * @return	string		$value  : the value
	 * @access private
	 * @version    2.0.0
	 * @since      2.0.0
	 */
	private function _getPmFieldDeliveryorderZip()
	{
		$pmUid = $this->pObj->flexform->deliveryorderZip;
		$value = $this->pObj->powermail->getFieldById( $pmUid );
		return $value;
	}

	/**
	 * _getPmFieldEmailCustomerEmail( )  : Get the customer e-mail from powermail POST params
	 *
	 * @return	string		$value  : the value
	 * @access private
	 * @version    2.0.0
	 * @since      2.0.0
	 */
	private function _getPmFieldEmailCustomerEmail()
	{
		$pmUid = $this->pObj->flexform->emailCustomerEmail;
		$value = $this->pObj->powermail->getFieldById( $pmUid );

		// Workaround in context with PayPal
		if( empty( $value ) )
		{
			$sesArray = $GLOBALS[ 'TSFE' ]->fe_user->getKey( 'ses', $this->extKey . '_' . $this->pidCaddy );
			$value = $sesArray[ 'powermail' ][ 'values' ][ 'customer' ][ 'email' ];
		}

		return $value;
	}

	/**
	 * _getPmFieldStatisticValue01( )  : Get the statisticValue01 field from powermail POST params
	 *
	 * @return	string		$value  : the value
	 * @access private
	 * @version    7.3.0
	 * @since      7.3.0
	 */
	private function _getPmFieldStatisticValue01()
	{
		$pmUid = $this->pObj->flexform->statisticValue01;
		$value = $this->pObj->powermail->getFieldById( $pmUid );
		if( is_array( $value ) )
		{
			$value = implode( ', ', $value );
		}
		return $value;
	}

	/**
	 * getInvoiceAddress( )  : Get the invoice Address from powermail POST params
	 *
	 * @return	string		$value  : the value
	 * @access private
	 * @version    2.0.0
	 * @since      2.0.0
	 */
	private function _getPmFieldInvoiceAddress()
	{
		$pmUid = $this->pObj->flexform->invoiceAddress;
		$value = $this->pObj->powermail->getFieldById( $pmUid );
		return $value;
	}

	/**
	 * getInvoiceCity( )  : Get the invoice City from powermail POST params
	 *
	 * @return	string		$value  : the value
	 * @access private
	 * @version    2.0.0
	 * @since      2.0.0
	 */
	private function _getPmFieldInvoiceCity()
	{
		$pmUid = $this->pObj->flexform->invoiceCity;
		$value = $this->pObj->powermail->getFieldById( $pmUid );
		return $value;
	}

	/**
	 * getInvoiceCompany( )  : Get the invoice Company from powermail POST params
	 *
	 * @return	string		$value  : the value
	 * @access private
	 * @version    2.0.0
	 * @since      2.0.0
	 */
	private function _getPmFieldInvoiceCompany()
	{
		$pmUid = $this->pObj->flexform->invoiceCompany;
		$value = $this->pObj->powermail->getFieldById( $pmUid );
		return $value;
	}

	/**
	 * getInvoiceCountry( )  : Get the invoice Country from powermail POST params
	 *
	 * @return	string		$value  : the value
	 * @access private
	 * @version    2.0.0
	 * @since      2.0.0
	 */
	private function _getPmFieldInvoiceCountry()
	{
		$pmUid = $this->pObj->flexform->invoiceCountry;
		$value = $this->pObj->powermail->getFieldById( $pmUid );
		return $value;
	}

	/**
	 * getInvoiceFirstname( )  : Get the invoice Firstname from powermail POST params
	 *
	 * @return	string		$value  : the value
	 * @access private
	 * @version    2.0.0
	 * @since      2.0.0
	 */
	private function _getPmFieldInvoiceFirstname()
	{
		$pmUid = $this->pObj->flexform->invoiceFirstname;
		$value = $this->pObj->powermail->getFieldById( $pmUid );
		return $value;
	}

	/**
	 * getInvoiceLastname( )  : Get the invoice Lastname from powermail POST params
	 *
	 * @return	string		$value  : the value
	 * @access private
	 * @version    2.0.0
	 * @since      2.0.0
	 */
	private function _getPmFieldInvoiceLastname()
	{
		$pmUid = $this->pObj->flexform->invoiceLastname;
		$value = $this->pObj->powermail->getFieldById( $pmUid );
		return $value;
	}

	/**
	 * getInvoiceZip( )  : Get the invoice Zip from powermail POST params
	 *
	 * @return	string		$value  : the value
	 * @access private
	 * @version    2.0.0
	 * @since      2.0.0
	 */
	private function _getPmFieldInvoiceZip()
	{
		$pmUid = $this->pObj->flexform->invoiceZip;
		$value = $this->pObj->powermail->getFieldById( $pmUid );
		return $value;
	}

	/*	 * *********************************************
	 *
	 * Init
	 *
	 * ******************************************** */

	/**
	 * _init( )
	 *
	 * @param	integer		$pidCaddy: pid of the page with the current Caddy plugin
	 * @return	void
	 * @access private
	 * @internal   #54628
	 * @version    4.0.7
	 * @since      4.0.7
	 */
	private function _init( $pidCaddy = null )
	{
		$this->_initVars();
		$this->_initInstances();
		$this->initPidCaddy( $pidCaddy );
	}

	/**
	 * _initInstances( )
	 *
	 * @return	void
	 * @access private
	 * @internal   #54628
	 * @version    4.0.7
	 * @since      4.0.7
	 */
	private function _initInstances()
	{
		if( !( $this->_initInstances === null ) )
		{
			return;
		}

		$this->_initInstancesSession();
		$this->_initInstancesStockmanager();

		$this->_initInstances = true;
	}

	/**
	 * _initInstancesSession( )
	 *
	 * @return	void
	 * @access private
	 * @internal   #54628
	 * @version    4.0.7
	 * @since      4.0.7
	 */
	private function _initInstancesSession()
	{
		// #i0226, 180312, dwildt, ~
		$this->_session = \TYPO3\CMS\Core\Utility\GeneralUtility::makeInstance( \Netzmacher\Caddy\Lib\Session::class );
		$this->_session->setParentObject( $this );
	}

	/**
	 * _initInstancesStockmanager( )
	 *
	 * @return	void
	 * @access private
	 * @internal   #54628
	 * @version    4.0.7
	 * @since      4.0.7
	 */
	private function _initInstancesStockmanager()
	{
		// #i0226, 180312, dwildt, ~
		$this->_stockmanager = \TYPO3\CMS\Core\Utility\GeneralUtility::makeInstance( \Netzmacher\Caddy\Lib\Caddy\Stockmanager::class );
		$this->_stockmanager->setParentObject( $this );
	}

	/**
	 * _initVars( )
	 *
	 * @return	void
	 * @access private
	 * @internal   #54628
	 * @version    4.0.7
	 * @since      4.0.7
	 */
	private function _initVars()
	{
		$this->cObj = $this->pObj->cObj;
		$this->conf = $this->pObj->conf;
		$this->drs = $this->pObj->drs;
		$this->local_cObj = $this->pObj->local_cObj;
	}

	/**
	 * initPidCaddy( )
	 *
	 * @param	integer		$pidCaddy: pid of the page with the current Caddy plugin
	 * @return	void
	 * @access public
	 * @internal   #54628
	 * @version    4.0.3
	 * @since      4.0.3
	 */
	public function initPidCaddy( $pidCaddy = null )
	{
		$this->pidCaddy = ( int ) $pidCaddy;
		if( $pidCaddy === null )
		{
			$this->pidCaddy = ( int ) $GLOBALS[ "TSFE" ]->id;
		}
//var_dump( __METHOD__, __LINE__, $this->pidCaddy );
	}

	/*	 * *********************************************
	 *
	 * Session
	 *
	 * ******************************************** */

	/**
	 * _sessionUpdate( )
	 *
	 * @return	void
	 * @access private
	 * @internal   #i0142
	 * @version    7.3.2
	 * @since      2.0.0
	 */
	private function _sessionUpdate()
	{
		$this->_sessionUpdateDebit();
		$this->_sessionUpdateDeliveryorder();
		$this->_sessionUpdateInvoice();
	}

	/**
	 * _sessionUpdateDebit( )
	 *
	 * @return	void
	 * @access private
	 * @internal   #i0142
	 * @version    7.3.2
	 * @since      7.3.2
	 */
	private function _sessionUpdateDebit()
	{
		// Get the session array
		$sesArray = $GLOBALS[ 'TSFE' ]->fe_user->getKey( 'ses', $this->extKey . '_' . $this->pidCaddy );

		$sesArray[ 'debitBic' ] = $this->_getPmFieldDebitBic();
		$sesArray[ 'debitIban' ] = $this->_getPmFieldDebitIban();
		$sesArray[ 'debitOwner' ] = $this->_getPmFieldDebitOwner();

		$GLOBALS[ 'TSFE' ]->fe_user->setKey( 'ses', $this->extKey . '_' . $this->pidCaddy, $sesArray ); // Generate new session
		$GLOBALS[ 'TSFE' ]->storeSessionData(); // Save session
		// DRS
		if( $this->pObj->drs->drsClean )
		{
			$prompt = 'Debit values are added to the session.';
			\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[INFO/CLEAN] ' . $prompt, $this->pObj->extKey, 0 );
		}
		// DRS
	}

	/**
	 * _sessionUpdateDeliveryorder( )
	 *
	 * @return	void
	 * @access private
	 * @version    2.0.0
	 * @since      2.0.0
	 */
	private function _sessionUpdateDeliveryorder()
	{
		// Get the session array
		// #54634, 131229, dwildt, 1-
		//$sesArray = $GLOBALS['TSFE']->fe_user->getKey( 'ses', $this->extKey . '_' . $GLOBALS["TSFE"]->id );
		// #54634, 131229, dwildt, 1+
		$sesArray = $GLOBALS[ 'TSFE' ]->fe_user->getKey( 'ses', $this->extKey . '_' . $this->pidCaddy );

		$sesArray[ 'deliveryorderAddress' ] = $this->_getPmFieldDeliveryorderAddress();
		$sesArray[ 'deliveryorderCity' ] = $this->_getPmFieldDeliveryorderCity();
		$sesArray[ 'deliveryorderCompany' ] = $this->_getPmFieldDeliveryorderCompany();
		$sesArray[ 'deliveryorderCountry' ] = $this->_getPmFieldDeliveryorderCountry();
		$sesArray[ 'deliveryorderFirstname' ] = $this->_getPmFieldDeliveryorderFirstname();
		$sesArray[ 'deliveryorderLastname' ] = $this->_getPmFieldDeliveryorderLastname();
		$sesArray[ 'deliveryorderZip' ] = $this->_getPmFieldDeliveryorderZip();

		// #54634, 131229, dwildt, 1-
		//$GLOBALS['TSFE']->fe_user->setKey('ses', $this->extKey . '_' . $GLOBALS["TSFE"]->id, $sesArray); // Generate new session
		// #54634, 131229, dwildt, 1+
		$GLOBALS[ 'TSFE' ]->fe_user->setKey( 'ses', $this->extKey . '_' . $this->pidCaddy, $sesArray ); // Generate new session
		$GLOBALS[ 'TSFE' ]->storeSessionData(); // Save session
		// DRS
		if( $this->pObj->drs->drsClean )
		{
			$prompt = 'Delivery address is added to the session.';
			\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[INFO/CLEAN] ' . $prompt, $this->pObj->extKey, 0 );
		}
		// DRS
	}

	/**
	 * _sessionUpdateInvoice( )
	 *
	 * @return	void
	 * @access private
	 * @version    2.0.0
	 * @since      2.0.0
	 */
	private function _sessionUpdateInvoice()
	{
		// Get the session array
		// #54634, 131229, dwildt, 1-
		//$sesArray = $GLOBALS['TSFE']->fe_user->getKey( 'ses', $this->extKey . '_' . $GLOBALS["TSFE"]->id );
		// #54634, 131229, dwildt, 1+
		$sesArray = $GLOBALS[ 'TSFE' ]->fe_user->getKey( 'ses', $this->extKey . '_' . $this->pidCaddy );

		$sesArray[ 'invoiceAddress' ] = $this->_getPmFieldInvoiceAddress();
		$sesArray[ 'invoiceCity' ] = $this->_getPmFieldInvoiceCity();
		$sesArray[ 'invoiceCompany' ] = $this->_getPmFieldInvoiceCompany();
		$sesArray[ 'invoiceCountry' ] = $this->_getPmFieldInvoiceCountry();
		$sesArray[ 'invoiceFirstname' ] = $this->_getPmFieldInvoiceFirstname();
		$sesArray[ 'invoiceLastname' ] = $this->_getPmFieldInvoiceLastname();
		$sesArray[ 'invoiceZip' ] = $this->_getPmFieldInvoiceZip();

		// #54634, 131229, dwildt, 1-
		//$GLOBALS['TSFE']->fe_user->setKey('ses', $this->extKey . '_' . $GLOBALS["TSFE"]->id, $sesArray); // Generate new session
		// #54634, 131229, dwildt, 1+
		$GLOBALS[ 'TSFE' ]->fe_user->setKey( 'ses', $this->extKey . '_' . $this->pidCaddy, $sesArray ); // Generate new session
		$GLOBALS[ 'TSFE' ]->storeSessionData(); // Save session
		// DRS
		if( $this->pObj->drs->drsClean )
		{
			$prompt = 'Invoice address is added to the session.';
			\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[INFO/CLEAN] ' . $prompt, $this->pObj->extKey, 0 );
		}
		// DRS
	}

	/**
	 * _sqlErrorDie( )
	 *
	 * @param string $table
	 * @param string $record
	 * @return	void
	 * @access private
	 * @version    7.3.0
	 * @since      2.0.0
	 */
	private function _sqlErrorDie( $table, $record, $method = __METHOD__, $line = __LINE__ )
	{
		$error = $GLOBALS[ 'TYPO3_DB' ]->sql_error();

		if( empty( $error ) )
		{
			return;
		}

		$query = $GLOBALS[ 'TYPO3_DB' ]->INSERTquery( $table, $record );
		$prompt = '<h1>caddy: SQL-Error</h1>'
						. '<p>' . $error . '</p>'
						. '<p>' . $query . '</p>'
						. '<p>&nbsp</p>'
						. '<p>Method: ' . $method . '</p>'
						. '<p>Line: ' . $line . '</p>'
						. '<p>' . $query . '</p>';
		die( $prompt );
	}

	/**
	 * _sqlInsertItem( )
	 *
	 * @param int $orderUid
	 * @param array $items
	 * @return	int $insertId
	 * @access private
	 * @version    7.3.0
	 * @since      7.3.0
	 */
	private function _sqlInsertItem( $orderUid, $item )
	{
		$table = 'tx_caddy_item';
		$time = time();

		// 170420
		$tax = $item[ 'tax' ];
		if( $tax === NULL )
		{
			$tax = 0;
		}
		$weight = $item[ 'weight' ];
		if( $weight === NULL )
		{
			$weight = 0;
		}

		$record = array(
				'pid' => $this->pObj->pid,
				'tstamp' => $time,
				'crdate' => $time,
				'externalTable' => $this->pObj->conf[ 'db.' ][ 'table' ],
				'externalUid' => $item[ 'uid' ],
				'gross' => $item[ 'gross' ],
				'quantity' => $item[ 'qty' ],
				'sku' => $item[ 'sku' ],
				'tax' => $tax,
				'title' => $item[ 'title' ],
				'tx_caddy_order' => $orderUid,
				'weight' => $weight,
		);

		$GLOBALS[ 'TYPO3_DB' ]->exec_INSERTquery( $table, $record );
		$this->_sqlErrorDie( $table, $record, __METHOD__, __LINE__ );

		$insertId = $GLOBALS[ 'TYPO3_DB' ]->sql_insert_id();

		if( !$this->pObj->drs->drsClean )
		{
			return $insertId;
		}

		$prompt = 'tx_caddy_item (#' . $insertId . ') is inserted into the database.';
		\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[INFO/CLEAN] ' . $prompt, $this->pObj->extKey, 0 );
		return $insertId;
	}

	/**
	 * _sqlInsertItems( )
	 *
	 * @param int $orderUid
	 * @return	int $insertId
	 * @access private
	 * @version    7.3.0
	 * @since      7.3.0
	 */
	private function _sqlInsertItems( $orderUid )
	{
		$sesArray = $GLOBALS[ 'TSFE' ]->fe_user->getKey( 'ses', $this->extKey . '_' . $this->pidCaddy );

		// RETURN : any product, don't increase numbers!
		if( empty( $sesArray[ 'products' ] ) )
		{
			// ERROR
			return;
		}

		$items = $sesArray[ 'products' ];
		foreach( ( array ) $items as $item )
		{
			$this->_sqlInsertItem( $orderUid, $item );
		}
	}

	/**
	 * _sqlInsertOrder( )
	 *
	 * @return	int $insertId
	 * @access private
	 * @version    9.1.1
	 * @since      7.3.0
	 */
	private function _sqlInsertOrder()
	{
		$sesArray = null;
		$table = 'tx_caddy_order';
		$time = time();

		$sesArray = $GLOBALS[ 'TSFE' ]->fe_user->getKey( 'ses', $this->extKey . '_' . $this->pidCaddy );

		$record = array(
				'pid' => $this->pObj->pid,
				'tstamp' => $time,
				'crdate' => $time,
				'customerEmail' => $this->_getPmFieldEmailCustomerEmail(),
				//#i0160, 170919, dwildt, +
				'customerDebitBic' => $this->_getPmFieldDebitBic(),
				'customerDebitIban' => $this->_getPmFieldDebitIban(),
				'customerDebitOwner' => $this->_getPmFieldDebitOwner(),
				'customerDeliveryorderAddress' => $this->_getPmFieldDeliveryorderAddress(),
				'customerDeliveryorderCity' => $this->_getPmFieldDeliveryorderCity(),
				'customerDeliveryorderCompany' => $this->_getPmFieldDeliveryorderCompany(),
				'customerDeliveryorderCountry' => $this->_getPmFieldDeliveryorderCountry(),
				'customerDeliveryorderFirstname' => $this->_getPmFieldDeliveryorderFirstname(),
				'customerDeliveryorderLastname' => $this->_getPmFieldDeliveryorderLastname(),
				'customerDeliveryorderZip' => $this->_getPmFieldDeliveryorderZip(),
				'customerInvoiceAddress' => $this->_getPmFieldInvoiceAddress(),
				'customerInvoiceCity' => $this->_getPmFieldInvoiceCity(),
				'customerInvoiceCompany' => $this->_getPmFieldInvoiceCompany(),
				'customerInvoiceCountry' => $this->_getPmFieldInvoiceCountry(),
				'customerInvoiceFirstname' => $this->_getPmFieldInvoiceFirstname(),
				'customerInvoiceLastname' => $this->_getPmFieldInvoiceLastname(),
				'customerInvoiceZip' => $this->_getPmFieldInvoiceZip(),
				// #t3910, 191208, , dwildt, 1+
				'fe_cruser_id' => $this->_getFeuserUid(),
				'quantity' => count( $sesArray[ 'products' ] ),
				//#i0136, 170311, dwildt, 1+
				'statisticValue01' => $this->_getPmFieldStatisticValue01(),
				'tx_caddy_item' => count( $sesArray[ 'products' ] ),
		);
		$record = $record + $this->_databaseFieldsFiles( $sesArray ) + $this->_databaseFieldsNumbers( $sesArray ) + $this->_databaseFieldsSum( $sesArray )
		;

		$GLOBALS[ 'TYPO3_DB' ]->exec_INSERTquery( $table, $record );
		$this->_sqlErrorDie( $table, $record, __METHOD__, __LINE__ );

		$insertId = $GLOBALS[ 'TYPO3_DB' ]->sql_insert_id();

		if( !$this->pObj->drs->drsClean )
		{
			return $insertId;
		}

		$prompt = 'tx_caddy_order (#' . $insertId . ') is inserted into the database.';
		\TYPO3\CMS\Core\Utility\GeneralUtility::devlog( '[INFO/CLEAN] ' . $prompt, $this->pObj->extKey, 0 );
		return $insertId;
	}

	/*	 * *********************************************
	 *
	 * Stock
	 *
	 * ******************************************** */

	/**
	 * _stock( )
	 *
	 * @return	void
	 * @access private
	 * @internal   #55726
	 * @version    4.0.7
	 * @since      4.0.7
	 */
	private function _stock()
	{
		$this->_stockCheck();
		$this->_stockUpdate();
//var_dump( __METHOD__, __LINE__ );
//die( );
	}

	/**
	 * _stockCheck( )
	 *
	 * @return	void
	 * @access private
	 * @internal   #55726
	 * @version    7.2.13
	 * @since      4.0.7
	 */
	private function _stockCheck()
	{
		$itemsSession = $this->_session->productsGet( $this->pidCaddy );

		foreach( ( array ) $itemsSession as $itemSession )
		{
			$uid = $itemSession[ 'uid' ];
			$qtyByOrder = $itemSession[ 'qty' ];
			$itemRealLive = $this->_session->productGet( $uid );
			$qtyInStock = $this->_stockmanager->getItemQuantity( $itemRealLive );
//var_dump(__METHOD__, __LINE__, $qtyInStock );
			switch( true )
			{
				case( $qtyByOrder > $qtyInStock ):
				case( $qtyByOrder < 1 ):
					$content = null;
					$conf = null;
					$this->_session->sessionDelete( $content, $conf, $this->pidCaddy );
					$title = $itemSession[ 'title' ];
					$this->_stockmanager->dieQtyError( $title, $qtyByOrder, $qtyInStock );
					break;
				default:
					// follow the workflow
					break;
			}
		}
	}

	/**
	 * _stockUpdate( )
	 *
	 * @return	void
	 * @access private
	 * @internal   #55726
	 * @version    4.0.7
	 * @since      4.0.7
	 */
	private function _stockUpdate()
	{
		$items = $this->_session->productsGet( $this->pidCaddy );
		$this->_stockmanager->itemsUpdate( $items );
//var_dump( __METHOD__, __LINE__, $items );
//die( );
	}

}
