<?php

namespace Netzmacher\Caddy\Utility\Debug;

if( !defined( 'TYPO3_MODE' ) )
{
	die( 'Access denied.' );
}

/* * *************************************************************
 *  Copyright notice
 *
 *  (c) 2016-2020 -  Dirk Wildt <http://wildt.at.die-netzmacher.de>
 *  All rights reserved
 *
 *  This script is part of the TYPO3 project. The TYPO3 project is
 *  free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  The GNU General Public License can be found at
 *  http://www.gnu.org/copyleft/gpl.html.
 *
 *  This script is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  This copyright notice MUST APPEAR in all copies of the script!
 * ************************************************************* */

/**
 *
 * @author  Dirk Wildt <http://wildt.at.die-netzmacher.de>
 * @package TYPO3
 * @subpackage caddy
 * @version     8.10.2
 * @since       8.9.6
 */
class Console
{

	/**
	 * log( ): 
	 *
	 * @param	mixed			$data: 
	 * @param	string		$method: 
	 * @param	string		$line: 
	 * @param	boolean		$prompt: 
	 * @return	string		$path : absolute path or false in case of an error
	 * @access public
	 * @since       8.10.2
	 * @version     8.9.6
	 */
	static function log( $data, $method, $line, $prompt = false )
	{
		// RETURN : Don't output debug prompt
		if( !$prompt )
		{
			return;
		}

		SELF::_ExtConfUnserialize();
		if( !SELF::_DebuggingConsole() )
		{
			SELF::_ExtConfSerialize();
			return;
		}

		SELF::_Prompt( $data, $method, $line );
		SELF::_ExtConfSerialize();
	}

	/**
	 * @access private
	 * @return      boolean
	 * @version     8.10.0
	 * @since       8.10.0
	 */
	static private function _DebuggingConsole()
	{
		$_DebuggingConsole = $GLOBALS[ 'TYPO3_CONF_VARS' ][ 'EXT' ][ 'extConf' ][ 'caddy' ][ 'debuggingConsole' ];

		switch( $_DebuggingConsole )
		{
			case( 'Enabled (for debugging only!)' ):
				return TRUE;
			case( 'Disabled' ):
			case( null ):
			default:
				return FALSE;
		}
	}

	/**
	 * @access private
	 * @return      void
	 * @version     8.10.2
	 * @since       8.10.2
	 */
	static private function _ExtConfSerialize()
	{
		if( is_array( $GLOBALS[ 'TYPO3_CONF_VARS' ][ 'EXT' ][ 'extConf' ][ 'caddy' ] ) )
		{
			$GLOBALS[ 'TYPO3_CONF_VARS' ][ 'EXT' ][ 'extConf' ][ 'caddy' ] = serialize( $GLOBALS[ 'TYPO3_CONF_VARS' ][ 'EXT' ][ 'extConf' ][ 'caddy' ] );
		}
	}

	/**
	 * @access private
	 * @return      void
	 * @version     8.10.2
	 * @since       8.10.2
	 */
	static private function _ExtConfUnserialize()
	{
		if( !is_array( $GLOBALS[ 'TYPO3_CONF_VARS' ][ 'EXT' ][ 'extConf' ][ 'caddy' ] ) )
		{
			$GLOBALS[ 'TYPO3_CONF_VARS' ][ 'EXT' ][ 'extConf' ][ 'caddy' ] = unserialize( $GLOBALS[ 'TYPO3_CONF_VARS' ][ 'EXT' ][ 'extConf' ][ 'caddy' ] );
		}
	}

	/**
	 * @param	mixed			$data: 
	 * @param	string		$method: 
	 * @param	string		$line: 
	 * @return      void
	 * @access private
	 * @version     8.10.0
	 * @since       8.10.0
	 */
	static private function _Prompt( $data, $method, $line )
	{
		// 180324, dwildt, 1+
		$method = str_replace( '\\', '\\\\', $method );

		if( empty( $data ) )
		{
			echo "<script>"
			. "console.log( '" . $method . "' + ' #" . $line . "' );"
			. "</script>";
			return;
		}

//		if( is_object( $output ) )
//		{
//			$output = 'Sorry: non printable object';
//		}
//		if( is_array( $output ) )
//		{
//			$output = implode( ',', $output );
//		}

		if( is_string( $data ) )
		{
			echo "<script>" . PHP_EOL
			. "  console.log( '" . $method . "' + ' #" . $line . ": " . $data . "' );" . PHP_EOL
			. "</script>" . PHP_EOL
			;
			return;
		}

		$data = json_encode( $data );
		echo "<script>" . PHP_EOL
		. "  console.log( '" . $method . "' + ' #" . $line . ": " . $data . "' );" . PHP_EOL
		. "</script>" . PHP_EOL
		;
	}

}
