<?php

namespace Netzmacher\Caddy\Utility\Template;

/* * *************************************************************
 *  Copyright notice
 *
 *  (c) 2019-2020 - Dirk Wildt <http://wildt.at.die-netzmacher.de>
 *  All rights reserved
 *
 *  This script is part of the TYPO3 project. The TYPO3 project is
 *  free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  The GNU General Public License can be found at
 *  http://www.gnu.org/copyleft/gpl.html.
 *
 *  This script is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  This copyright notice MUST APPEAR in all copies of the script!
 * ************************************************************* */

/**
 * Class for extending the Caddy classes
 *
 * @package TYPO3
 * @subpackage caddy
 * @author Dirk Wildt <http://wildt.at.die-netzmacher.de>
 * @internal #i0288
 * @version 8.17.0
 * @since 8.17.0
 */
class CleanupTemplateUtility extends \Netzmacher\Caddy\AbstractPlugin
{

	public $extKey = 'caddy';
	private $sessionArray = [];

	/**
	 * caddyEmailSubparts( )  : Remove unused subparts
	 *
	 * @param		string	$tmpl
	 * @param		integer	$pid
	 * @return  string  $tmpl
	 * @access public
	 * @version 8.17.0
	 * @since		8.17.0
	 */
	public function RemoveUnusedSubparts( $tmpl, $pid = null )
	{
		$this->_initSessionArray( $pid );

		$tmpl = $this->_removeInsurance( $tmpl );
		$tmpl = $this->_removePayment( $tmpl );
		$tmpl = $this->_removeRecyclingfee( $tmpl );
		$tmpl = $this->_removeShipping( $tmpl );
		$tmpl = $this->_removeSpecials( $tmpl );
		$tmpl = $this->_removeTaxNormal( $tmpl );
		$tmpl = $this->_removeTaxReduced( $tmpl );

//		var_dump( __METHOD__, __LINE__, $tmpl );
//		die();

		return $tmpl;
	}

	/**
	 * _initSessionArray( )  : Remove unused subparts
	 *
	 * @param		string	$tmpl
	 * @param		integer	$pid
	 * @return  string  $tmpl
	 * @access private
	 * @version 8.17.0
	 * @since		8.17.0
	 */
	public function _initSessionArray( $pid )
	{
		if( empty( $pid ) )
		{
			$pid = $GLOBALS[ "TSFE" ]->id;
		}
		$this->sessionArray = $GLOBALS[ 'TSFE' ]->fe_user->getKey( 'ses', $this->extKey . '_' . $pid );
	}

	/**
	 * _remove( ): Removes a subpart, if sum net is empty
	 *
	 * @param		array		$optionType : 
	 * @param		string	$tmpl
	 * @return  string  $tmpl
	 * @access private
	 * @version 8.17.0
	 * @since		8.17.0
	 */
	private function _remove( $optionType, $tmpl )
	{
		$value = $this->sessionArray[ 'options' ][ $optionType ][ 'sum' ][ 'net' ];
		if( !empty( $value ) )
		{
			return $tmpl;
		}

		$subpart = '###OPTIONS' . strtoupper( $optionType ) . '###';
		$tmpl = $this->MarkerBasedTemplateService->substituteSubpart( $tmpl, $subpart, null );

		return $tmpl;
	}

	/**
	 * _removeInsurance( )  : Removes a subpart, if sum net is empty
	 *
	 * @param		string	$tmpl
	 * @return  string  $tmpl
	 * @access private
	 * @version 8.17.0
	 * @since		8.17.0
	 */
	private function _removeInsurance( $tmpl )
	{
		$tmpl = $this->_remove( 'insurance', $tmpl );
		return $tmpl;
	}

	/**
	 * _removePayment( )  : Removes a subpart, if sum net is empty
	 *
	 * @param		string	$tmpl
	 * @return  string  $tmpl
	 * @access private
	 * @version 8.17.0
	 * @since		8.17.0
	 */
	private function _removePayment( $tmpl )
	{
		$tmpl = $this->_remove( 'payment', $tmpl );
		return $tmpl;
	}

	/**
	 * _removeRecyclingfee( )  : Removes a subpart, if sum net is empty
	 *
	 * @param		string	$tmpl
	 * @return  string  $tmpl
	 * @access private
	 * @version 8.17.0
	 * @since		8.17.0
	 */
	private function _removeRecyclingfee( $tmpl )
	{
		$tmpl = $this->_remove( 'recyclingfee', $tmpl );
		return $tmpl;
	}

	/**
	 * _removeShipping( )  : Removes a subpart, if sum net is empty
	 *
	 * @param		string	$tmpl
	 * @return  string  $tmpl
	 * @access private
	 * @version 8.17.0
	 * @since		8.17.0
	 */
	private function _removeShipping( $tmpl )
	{
		$tmpl = $this->_remove( 'shipping', $tmpl );
		return $tmpl;
	}

	/**
	 * _removeTaxNormal( )  : Removes a subpart, if sum net is empty
	 *
	 * @param		string	$tmpl
	 * @return  string  $tmpl
	 * @access private
	 * @version 8.17.0
	 * @since		8.17.0
	 */
	private function _removeTaxNormal( $tmpl )
	{
		$value = $this->sessionArray[ 'sumsumtaxnormal' ];
		if( !empty( $value ) )
		{
			return $tmpl;
		}

		$subpart = '###TAXNORMAL###';
		$tmpl = $this->MarkerBasedTemplateService->substituteSubpart( $tmpl, $subpart, null );

		return $tmpl;
	}

	/**
	 * _removeTaxReduced( )  : Removes a subpart, if sum net is empty
	 *
	 * @param		string	$tmpl
	 * @return  string  $tmpl
	 * @access private
	 * @version 8.17.0
	 * @since		8.17.0
	 */
	private function _removeTaxReduced( $tmpl )
	{
		$value = $this->sessionArray[ 'sumsumtaxreduced' ];
		if( !empty( $value ) )
		{
			return $tmpl;
		}

		$subpart = '###TAXREDUCED###';
		$tmpl = $this->MarkerBasedTemplateService->substituteSubpart( $tmpl, $subpart, null );

		return $tmpl;
	}

	/**
	 * _removeSpecials( )  : Removes a subpart, if sum net is empty
	 *
	 * @param		string	$tmpl
	 * @return  string  $tmpl
	 * @access private
	 * @version 8.17.0
	 * @since		8.17.0
	 */
	private function _removeSpecials( $tmpl )
	{
		$tmpl = $this->_remove( 'specials', $tmpl );
		return $tmpl;
	}

}
